(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export ProgLog.
Require Import ZArith.
Require Import Addr.
Require Export ListAddrSet.

Set Implicit Arguments. 


Fixpoint array (x:addr) (l:list val) :=
  match l with
    | nil => emp
    | v::l' => x |-> v ** array (x + 4) l'
  end.


Lemma arrayConsE : forall x v l,
  mimpl (array x (v::l))
  (x |-> v ** array (x + 4) l).
Proof.
  intros x v l m Hp; assumption.
Qed.

Lemma arrayConsI : forall (x:addr) v l,
  mimpl (x |-> v ** array (x + 4) l)
        (array x (v::l)).
Proof.
  intros x v l m Hp; assumption.
Qed.

Lemma arrayConsEqv : forall x v l,
  miff (array x (v::l))
       (x |-> v ** array (x + 4) l).
Proof.
  intros x v l m; constructor; intros Hp; auto.
Qed.
Hint Rewrite arrayConsEqv : SLogI.
Hint Rewrite <- arrayConsEqv : SLogE.


Lemma arrayEmptyEqv : forall x,
  miff (array x nil) emp.
Proof.
  simpl; intros _; intros m; tauto.
Qed.
Hint Rewrite arrayEmptyEqv : SLogI.
Hint Rewrite <- arrayEmptyEqv : SLogE.


Lemma arraySnocI : forall l x v,
  mimpl (array x l ** x + 4 * length l |-> v)
  (array x (l ++ v :: nil)).
Proof.
  induction l; intros x v m Hp; simpl in *; slei Hp.
  ssplit.
  srewrite <- IHl. ssplit.
Qed.


Lemma arraySnocE : forall l x v,
  mimpl (array x (l ++ v :: nil))
    (array x l ** x + 4 * length l |-> v).
Proof.
  induction l; intros x v m Hp; simpl in *; slei Hp.
  ssplit.
  srewrite IHl in Hp. ssplit.
Qed.


Lemma arraySnocEqv : forall l x v,
  miff (array x l ** x + 4 * length l |-> v) (array x (l ++ v :: nil)).
Proof.
  intros l x v m; constructor; intros Hp.
  apply arraySnocI; auto.
  apply arraySnocE; auto.
Qed.
(* only need this in one direction *)
Hint Rewrite arraySnocEqv : SLogE.


Lemma arrayAppI : forall x l' l,
  mimpl (array x l ** array (x + 4 * length l) l')
  (array x (l ++ l')).
Proof.
  induction l'; intros l m Hp; sarewriteE in Hp.

  autorewrite with list. autof.

  scar Hp. sarewriteI in Hp.
  replace (l ++ a :: l') with ((l ++ a :: nil) ++ l')
    by (autorewrite with list; simpl; auto).
  apply IHl'; clear IHl'.
  autorewrite with list AddrSimpl.
  replace (x + 4 * length l + 4)
    with (x + 4 * (length l + 1)%word) in Hp by addrEqSolve.
  auto.
Qed.


Lemma arrayAppE : forall x l' l,
  mimpl (array x (l ++ l')) 
  (array x l ** array (x + 4 * length l) l').
Proof.
  induction l'; intros l m Hp; sarewriteI.

  autorewrite with list in Hp; auto.

  gscal. sarewriteE.
  replace (l ++ a :: l') with ((l ++ a :: nil) ++ l') in Hp
    by (autorewrite with list; reflexivity).
  srewrite IHl' in Hp; clear IHl'.
  replace (x + 4 * length (l ++ a :: nil))
    with (x + 4 * length l + 4) in Hp.
  2:autorewrite with list; rewrite natWordAddEq; addrEqSolve.
  auto.
Qed.


Lemma arrayAppEqv : forall x l' l,
  miff (array x (l ++ l')) (array x l ** array (x + 4 * length l) l').
Proof.
  intros x l' l m; constructor; intros Hp.
  apply arrayAppE; auto.
  apply arrayAppI; auto.
Qed.
Hint Rewrite arrayAppEqv : SLogI.
Hint Rewrite <- arrayAppEqv : SLogE.


(* helper lemma for arrayLenBounded *)
Lemma arrayGetCont : forall l x k,
  (k < length l)%nat ->
  mimpl (array x l) (x + 4 * k |->- ** top).
Proof.
  induction l; intros x k LT m Hp; simpl in LT.

  elf; omega.

  sarewriteE in Hp.
  destruct k.
  ssplit.

  srewrite (IHl (x + 4) k) in Hp. 2:omega.
  scl2 Hp; ssplit.

Qed.


(* the basic idea is to show that if the length is any longer, than
   the same element exists twice, which is a contradiction. *)
Lemma arrayLenBounded : forall x l A m,
  (array x l ** A) m ->
  (Z_of_nat (length l) <= indexBound)%Z.
Proof.
  intros x l A m Hp.

  assert (D : (Z_of_nat (length l) <= indexBound \/ 
              indexBound < Z_of_nat (length l))%Z)
    by omega.
  destruct D; auto; elf.

  destruct l.
  discriminate H.
  sarewriteE in Hp.

  assert (LT : (Zabs_nat (indexBound - 1) < length l)%nat).
    replace (length l) with (Zabs_nat (Z_of_nat (length l)))
      by (rewrite nat_to_Z_to_nat; auto).
    autorewrite with list in H.
    womega.

  srewrite (@arrayGetCont l (x + 4) _ LT) in Hp; sle Hp.
  replace (x + 4 + 4 * Zabs_nat (indexBound - 1)) with x in Hp.
  contContSolver Hp.

  clear; addrEqSolve.
  replace (4 + 4 * Zabs_nat (indexBound - 1))%word with 0.
  addrEqSolve. clear. 
  (* eliminate conversion of 2^32 to a nat *)
  rewrite Z_to_nat_to_word by (vm_compute; discriminate).
  womega. reflexivity. 

Qed.
Implicit Arguments arrayLenBounded [x l A m].


Lemma arrayLenWordBounded : forall x l A m,
  (array x l ** A) m ->
  (Z_of_nat (length l) < wordBound)%Z.
Proof.
  intros x l A m Hp.
  assert (indexBound < wordBound)%Z as LT by (compute; auto).
  apply arrayLenBounded in Hp; omega.
Qed.
Implicit Arguments arrayLenWordBounded [x l A m].


Lemma arrayLenOkWordNat : forall x l A m,
  (array x l ** A) m ->
  okWordNat (length l).
Proof.
  unfold okWordNat; intros x l A m Hp; constructor.
  omega.
  apply (arrayLenWordBounded Hp).
Qed.
Implicit Arguments arrayLenOkWordNat [x l A m].


Lemma array_addrsInE : forall (x:addr) y f,
  AS.In x (array_addrs y (length f)) ->
  mimpl (array y f) (x |->- ** top).
Proof.
  clear; intros x y f Ix m Hp.
  setsimpl in Ix; destruct Ix as [k' [E LT]]; subst.
  assert (okWordNat (length f)) as okf.
    hyr Hp (empIR Hp); exact (arrayLenOkWordNat Hp).
  assert (X := fun LT => @arrayGetCont _ _ (word_to_nat k') LT _ Hp);
    simpl in X.
  rewrite word_to_nat_to_word in X; apply X; clear X Hp.
  replace (length f) with (Zabs_nat (Z_of_nat (length f)))
    by (rewrite nat_to_Z_to_nat; auto).
  womega.
Qed.


Lemma array_addrs_cardinalHp : forall x l A m,
  (array x l ** A) m ->
  AS.cardinal (array_addrs x (length l)) = word_to_nat (length l).
Proof.
  intros x l A m Hp.
  apply array_addrs_cardinal.
  assert (X := arrayLenBounded Hp); clear - X.
  womega.
Qed.


Lemma arrayFasSubDom : forall x y l,
  mimpl (array x (y::l)) (fasSubDom (AS.singleton x)).
Proof.
  simpl; unfold array, fasSubDom; intros x y l m Hp.
  apply AFAS.fasSingletonI; sli; smatch.
Qed.


(* a buffer is just an array where we don't care about the contents *)
Definition buffer (x:addr) (k:nat) : mpred :=
  lexists l, !(length l = k) ** array x l.


Lemma bufferLenBounded : forall x l A m,
  (buffer x l ** A) m ->
  (Z_of_nat l <= indexBound)%Z.
Proof.
  unfold buffer; intros x l A m Hp; sle Hp; subst l.
  apply (arrayLenBounded Hp).
Qed.


Lemma bufferConsEqv : forall x l,
  miff (buffer x (S l)) (x |->- ** buffer (x + 4) l).
Proof.
  unfold buffer; intros x l m; constructor; intros Hp; sle Hp.

  destruct x0.
  discriminate H.
  sarewriteE in Hp.
  simpl in H; injector H; sli; eauto.

  subst. sarewriteI in Hp. sli; eauto.
Qed.

(* This isn't a very good form for going backwards *)
Hint Rewrite bufferConsEqv : SLogI.


Lemma bufferEmptyI : forall x,
  mimpl emp (buffer x 0).
Proof.
  unfold buffer; intros x m Hp; exists (@nil val); sli.
Qed.

Lemma bufferEmptyE : forall x,
  mimpl (buffer x 0) emp.
Proof.
  unfold buffer; intros x m Hp; sle Hp.
  destruct x0.
  auto.
  discriminate H.
Qed.


Lemma bufferEmptyEqv : forall x,
  miff (buffer x 0) emp.
Proof.
  intros x m; constructor; intros Hp.
  srewrite bufferEmptyE in Hp; auto.
  srewrite <- bufferEmptyI; auto.
Qed.
Hint Rewrite bufferEmptyEqv : SLogI.
Hint Rewrite <- bufferEmptyEqv : SLogE.


Lemma bufferAppE : forall x k k',
  mimpl (buffer x (k + k')) (buffer x k ** buffer (x + 4 * k) k').
Proof.
  unfold buffer; intros x k k' m Hp; sle Hp.  rename x0 into l.

  assert (exists l1, exists l2, 
    k = length l1 /\ k' = length l2 /\ l1 ++ l2 = l) as E.
  generalize dependent x; generalize dependent m; 
    generalize dependent k; generalize dependent k'.
  induction l; intros.

  exists (@nil val); exists (@nil val); simpl in * .
  destruct k.
  destruct k'. auto. discriminate H.
  discriminate H.

  simpl in Hp, H.
  destruct k.

  simpl in H; subst k'.
  exists (@nil val); exists (a::l); auto.
  
  simpl in H; injector H.
  inversion_clear Hp.
  hyr IHl (IHl _ _ H _ _ H2); clear H.
  destruct IHl as [l1 [l2 [kEq [k'Eq Leq]]]]; subst.
  exists (a::l1); exists l2; auto.
  (* done proving E *)

  destruct E as [l1 [l2 [kEq [k'Eq lEq]]]]; subst.
  sli.
  
  sarewriteE; auto.

Qed.


Lemma bufferAppI : forall x k k',
  mimpl (buffer x k ** buffer (x + 4 * k) k') (buffer x (k + k')).
Proof.
  unfold buffer; intros x k k' m Hp; sle Hp; subst.
  exists (x0 ++ x1); sli.
  sarewriteI.
  autorewrite with list; reflexivity.
Qed.


Lemma bufferAppEqv : forall x k k',
  miff (buffer x (k + k')) (buffer x k ** buffer (x + 4 * k) k').
Proof.
  intros x k k' m; constructor; intros Hp.
  srewrite <- bufferAppE; auto.
  srewrite <- bufferAppI; auto.
Qed.
Hint Rewrite bufferAppEqv : SLogI.
Hint Rewrite <- bufferAppEqv : SLogE.


Lemma arrayAddrsFasDomI : forall (k:nat) x,
  mimpl (buffer x k) (fasDom (array_addrs x k)).
Proof.
  induction k; simpl; intros x m Hp.
  setsimpl.
  sarewriteI; sarewriteE in Hp; auto.
  
  assert (LE := bufferLenBounded (empIR Hp)).
  replace (S k) with (k + 1)%nat in Hp by omega.
  sarewriteE in Hp.
  srewrite IHk in Hp; clear IHk.
  assert (Ix : AS.In (x + 4 * k) (array_addrs x (S k)))
    by (apply array_addrsI; womega).
  simpl in Ix.
  srewrite <- (fasDomI Ix).
  searchMatch.
  rewrite <- array_addrsRemoveEnd. auto.
  womega.
Qed.


Lemma arrayAddrsFasDomE : forall (k:nat),
  (Z_of_nat k <= indexBound)%Z -> forall x,
  mimpl (fasDom (array_addrs x k)) (buffer x k).
Proof.
  unfold buffer; induction k; intros kOk x m Hp.

  exists (@nil val); sarewriteI.
  unfold fasDom in Hp; setsimpl in Hp; apply AFAS.fasEmptyE in Hp; auto.

  assert (Ix : AS.In (x + 4 * k) (array_addrs x (S k)))
    by (apply array_addrsI; womega).

  srewrite (fasDomE Ix) in Hp; clear Ix.
  replace (AS.remove (x + 4 * k) (array_addrs x (S k))) 
    with (array_addrs x k) in Hp.
  srewrite IHk in Hp; clear IHk.
  scl2 Hp; sle Hp.
  subst k; exists (x0 ++ x1::nil); sli.
  sarewriteI.

  autorewrite with list; simpl; omega.

  womega.

  apply array_addrsRemoveEnd.
  womega.

Qed.


Lemma bufferConsI : forall (x:addr) v l,
  mimpl (x |-> v ** buffer (x + 4) l) (buffer x (S l)). 
Proof.
  intros x v l m Hp; sarewriteI; eauto.
Qed.
Hint Rewrite <- bufferConsI : SLogE.


Lemma arrayDomI : forall x l,
  mimpl (array x l) (fasDom (array_addrs x (length l))).
Proof.
  intros x l m Hp; apply arrayAddrsFasDomI; exists l; sli.
Qed.


Lemma bufferSnocEqv : forall x l,
  miff (buffer x (S l)) (buffer x l ** x + 4 * l |->-).
Proof.
  unfold buffer; intros x0 k m; constructor; intros Hp.
  sle Hp.

  assert (E : exists l, exists y, x = l ++ y :: nil).
    clear - H.
    generalize dependent k; induction x.
    discriminate 1.
    intros; simpl in H; injector H.
    destruct x; simpl.
      exists (@nil val); econstructor; simpl; reflexivity.
      simpl in IHx; destruct (IHx (length x)); clear IHx. auto.
      destruct H; rewrite H.
      exists (a :: x0); exists x1; reflexivity.
  destruct E as [l [y E]]; subst x.
  sarewriteE in Hp.
  assert (Keq : k = length l).
    autorewrite with list in H; omega.
  subst k; sli; searchMatch.
  sle Hp; subst k.
  sarewriteI in Hp.
  sli; eauto.
  autorewrite with list; omega.
Qed.
Hint Rewrite bufferSnocEqv : SLogI.


Lemma bufferFrame : forall (sp:block) k,
  (k < windexBound)%word ->
  miff (buffer sp (word_to_nat k)) (stackFrame sp (4 * k)).
Proof.
  intros sp.

  cut (forall k,
    (Z_of_nat k < indexBound)%Z ->
    miff (buffer sp k) (stackFrame sp (4 * (nat_to_word k)))).
  intros X k LE.
  hyr X (X (word_to_nat k)).
  rewrite word_to_nat_to_word in X; apply X.
  rewrite word_to_nat_to_Z. womega.

  unfold stackFrame.
  induction k.

  (* base case *)
  replace (4 * 0) with 0 by ring.
  intros _ m.
  constructor; intros Hp; sle Hp.
  sarewriteE in Hp. setsimpl; sli.
  setsimpl in Hp; sle Hp. sarewriteI.

  (* inductive case *)
  intros kOk. unSucc.
  replace (4 * (k + 1))%word with (4 * k + 4)%word by ring.
  rewrite zeroWordSetSplitEnd. rewrite wordAlignEq.
  intros m; constructor; intros Hp.

  apply fasI.
  sarewriteE in Hp; scl2 Hp. srewrite IHk in Hp. smatch. 
  clear - kOk; womega.
  clear; setsimpl; rewrite wordAlignEq; intros [_ LT]; womega.

  srewrite (@fasE (4 * k)) in Hp. 2:auto. scl2 Hp; sle Hp.
  sarewriteI. smatch.
  srewrite IHk. 2:clear -kOk; womega.
  replace (WS.remove (4 * k) (WS.add (4 * k) (zeroWordSet (4 * k))))
    with (zeroWordSet (4 * k)) in Hp. auto.
  clear.
  apply WSE.set_ext; intros x; setsimpl; rewrite wordAlignEq; constructor;
    intros X.
  destruct X as [X Y]; constructor. intros E; subst x; womega. auto. tauto.
  clear - kOk. womega.
Qed.



