(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export Word.
Require Export ListWordSet.
Require Export List.


Set Implicit Arguments.



(* based on Leroy-Blazy's paper on a memory model
http://pauillac.inria.fr/~xleroy/publi/memory-model-journal.pdf

Extended to better support separation reasoning.
  - each block had a set of valid subaddresses, instead of a range
  - defines a union of memories
  - defines unary memory
  - intrinsically canonical

*)


Module Type CoreMemorySig.
  
  Parameter mem : Type.
  Parameter block : Set.

  Declare Module BlockOrdered : UsualOrderedType with Definition t := block.

  Definition addr := prod block word.
  Definition blockOf (a:addr) : block :=
    let (b, _) := a in b.
  Definition offsetOf (a:addr) : word :=
    let (_, i) := a in i.

  Inductive val : Set :=
    | Vundef : val
    | Vword : word -> val
    | Vptr : addr -> val.


  (* predicates *)
  
  Parameter validBlock : mem -> block -> Prop.
  Parameter blockOffs : mem -> block -> WS.t.
  Definition validPtr (m:mem) (p:addr) : Prop :=
      validBlock m (blockOf p) /\ WS.In (offsetOf p) (blockOffs m (blockOf p)).


  (* constructors *)
  Parameter empty : mem.
  Parameter sing_mem : addr -> val -> mem.
  Parameter block_mem : block -> word -> (list val) -> mem.
  Parameter union : mem -> mem -> mem.


  (* operations *)
  Parameter load : mem -> addr -> option val.
  Parameter store : mem -> addr -> val -> option mem.
  Parameter free : mem -> block -> option mem.


  (* decidability of predicates *)
  Parameter eq_block_dec : forall (b b':block), {b = b'} + {b <> b'}.

  Parameter validBlock_dec : forall m b,
    {validBlock m b} + {~validBlock m b}.

  Parameter validPtr_dec : forall m p,
    {validPtr m p} + {~validPtr m p}.


  (* basic operation properties *)

  Parameter sing_memLoadEq : forall p v,
    load (sing_mem p v) p = Some v.

  Parameter sing_memStoreEq : forall p v v',
    store (sing_mem p v) p v' = Some (sing_mem p v').

  Parameter storeLoadEq : forall m p v m',
    store m p v = Some m' ->
    load m' p = Some v.
  
  Parameter storeLoadNeq : forall m m' p v,
    store m p v = Some m' -> forall p',
      p <> p' ->
      load m' p' = load m p'.

  Parameter freeLoadNeq : forall m m' b,
    free m b = Some m' -> forall p,
      blockOf p <> b ->
      load m' p = load m p.


  (* validity properties *)
  
  Parameter emptyValidE : forall b, 
    validBlock empty b -> 
    False.

  Parameter sing_memValidI : forall p v,
    validBlock (sing_mem p v) (blockOf p).

  Parameter sing_memValidE : forall p b v,
    validBlock (sing_mem p v) b ->
    b = blockOf p.

  Parameter block_memValidI : forall b i vv,
    0%word <> wordAlign i ->
    validBlock (block_mem b i vv) b.

  Parameter block_memValidE : forall b b' i vv,
    validBlock (block_mem b i vv) b' ->
    b = b' /\ 0%word <> wordAlign i.

  Parameter unionValidI : forall m b,
    validBlock m b -> forall m',
    validBlock (union m m') b.

  Parameter unionValidE : forall m m' b,
    validBlock (union m m') b ->
    validBlock m b \/ validBlock m' b.

  Parameter storeValidI : forall m m' p v,
    store m p v = Some m' -> forall b,
      validBlock m' b <-> validBlock m b.

  Parameter storeValidI1 : forall m m' p v,
    store m p v = Some m' -> forall b,
      validBlock m' b -> validBlock m b.
  
  Parameter storeValidI2 : forall m m' p v,
    store m p v = Some m' -> forall b,
      validBlock m b -> validBlock m' b.
  
  Parameter freeValidI1 : forall m m' b,
    free m b = Some m' -> forall b',
      validBlock m' b' -> 
      b' <> b /\ validBlock m b'.

  Parameter freeValidI2 : forall m m' b,
    free m b = Some m' -> forall b',
      b' <> b -> 
      validBlock m b' -> 
      validBlock m' b'.

  Parameter freeValidE : forall m b,
    validBlock m b ->
    exists m', free m b = Some m'.

  Parameter freeInvalidE : forall m b,
    ~validBlock m b ->
    free m b = None.

  Parameter invalidBlockEmptyI : forall m b,
    ~validBlock m b ->
    blockOffs m b = WS.empty.
  
  Parameter invalidBlockEmptyE : forall m b,
    blockOffs m b = WS.empty ->
    validBlock m b ->
    False.


  (** blockOffs properties *)

  Parameter sing_memAddrsEq : forall p v,
    blockOffs (sing_mem p v) (blockOf p) = WS.add (offsetOf p) WS.empty.

  Parameter block_memAddrsEq : forall b i vv,
    blockOffs (block_mem b i vv) b = zeroWordSet i.

  Parameter block_memAddrsNeq : forall b b',
    b <> b' -> forall len vv,
      blockOffs (block_mem b len vv) b' = WS.empty.  
  
  Parameter unionBlockAddrsEq : forall m m' b,
    blockOffs (union m m') b = blockOffs m b |_| blockOffs m' b.

  Parameter storeAddrsEq : forall m m' p v,
    store m p v = Some m' -> forall b,
      blockOffs m' b = blockOffs m b.

  Parameter freeAddrsEq : forall m m' b,
    free m b = Some m' -> forall b',
      b' <> b -> 
      blockOffs m' b' = blockOffs m b'.



  (* properties of valid pointers *)

  Parameter emptyPtrE : forall p,
    validPtr empty p ->
    False.

  Parameter sing_memPtrI : forall p v,
    validPtr (sing_mem p v) p.

  Parameter sing_memPtrE : forall p p' v,
    validPtr (sing_mem p v) p' ->
    p' = p.

  Parameter block_memPtrI : forall p i,
    WS.In (offsetOf p) (zeroWordSet i) -> forall vv,
    validPtr (block_mem (blockOf p) i vv) p.

  Parameter block_memPtrE : forall b i p vv,
    validPtr (block_mem b i vv) p ->
    blockOf p = b /\ WS.In (offsetOf p) (zeroWordSet i).

  Parameter unionPtrI : forall m p,
    validPtr m p -> forall m',
    validPtr (union m m') p.

  Parameter unionPtrE : forall m m' p,
    validPtr (union m m') p ->
    validPtr m p \/ validPtr m' p .

  Parameter storePtrI1 : forall m m' p v,
    store m p v = Some m' -> forall p',
      validPtr m' p' -> validPtr m p'.
  
  Parameter storePtrI2 : forall m m' p v,
    store m p v = Some m' -> forall p',
      validPtr m p' -> validPtr m' p'.
  
  Parameter loadPtrI : forall m p v,
    load m p = Some v ->
    validPtr m p.
  
  Parameter storePtrE : forall m p,
    validPtr m p -> forall v,
      exists m', store m p v = Some m'.

  Parameter loadPtrE : forall m p,
    validPtr m p ->
    exists v, load m p = Some v.

  Parameter invalidPtrE : forall m p,
    ~validPtr m p ->
    load m p = None.


  (* equalities on unions of memories *)

  Parameter unionEmpty : forall m, 
    union m empty = m.

  Parameter unionComm : forall m m',
    union m m' = union m' m.

  Parameter unionAssoc : forall h h' h'',
    union (union h h') h'' = union h (union h' h'').


  (* domain disjointedness *)

  Definition dom_disj m m' :=
    (forall p, validPtr m p -> validPtr m' p -> False).

  Parameter block_memEmptyEq : forall b len vv,
    wordAlign len = 0%word ->
    block_mem b len vv = empty.

  Parameter loadUnionEq : forall m p v,
    load m p = Some v -> forall m',
      dom_disj m m' ->
      load (union m m') p = Some v.

  Parameter storeUnionEq : forall m m' p v,
    store m p v = Some m' -> forall m'',
      dom_disj m m'' ->
      store (union m m'') p v = Some (union m' m'').

  Parameter freeUnionEq : forall m m' b len vv,
    free (union m (block_mem b len vv)) b = Some m' ->
    ~validBlock m b ->
    m = m'.


  (* you can split off the end of a non-empty block *)
  Parameter block_memSplit : forall w,
    (4 <= w + 4)%word -> forall b vv,
      block_mem b (w + 4)%word vv = 
      union (sing_mem (b, wordAlign w) (hd Vundef vv))
        (block_mem b w (tail vv)).


End CoreMemorySig.
