(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export FSetList.
Require Import FSetProperties.
Require Import CoreLog.
Require Import MyTactics.
Require Import OrderedTypeEx.
Require Import FSetIffSpecSig.

Set Implicit Arguments.


(*

In any FSetList built with an ordered type with the 'usual equality',
the Equal equality implies 'usual equality', because the set is kept in
an ordered list, and we have proof irrelevance.

*)



(* aliasing so it will be exported without exporting all of OrderedTypeEx *)
Module Type UsualOrderedType := UsualOrderedType.


Module Type FSetEq.
  Declare Module M : S.
  (* this looks pretty terrible to me *)
  Declare Module MiffFn : SetIffType .
  Module Miff := MiffFn M.

  Import M.

  Parameter elt_eq : forall x y,
    E.eq x y ->
    x = y.

  Parameter eq_elt_dec : forall (x y:elt),
    x = y \/ x <> y.

  Parameter set_ext : forall x y,
    Equal x y ->
    x = y.

  Parameter Disj : M.t -> M.t -> Prop.

  Parameter DisjI : forall s s',
    (forall x, In x s -> In x s' -> False) ->
    Disj s s'.

  Parameter DisjE : forall s s',
    Disj s s' -> forall x,
      In x s -> In x s' ->
      False.

  Parameter empty_union_eqv1 : forall S,
      union empty S = S.

  Parameter empty_union_eqv2 : forall S,
      union S empty = S.

End FSetEq.
  

Module Make (X : UsualOrderedType) <: FSetEq.

  Module Import M := FSetList.Make (X).
  Module MiffFn := Properties.
  Module Import Miff := MiffFn (M).

  Lemma elt_eq : forall x y,
    M.E.eq x y ->
    x = y.
  Proof.
    unfold M.E.eq, X.eq; auto.
  Qed.

  Lemma ltSorto : forall l a,
    lelistA X.lt a l ->
    sort X.lt l ->
    forall b, List.In b l -> X.lt a b.
  Proof.
    induction l; simpl; intros x LTl S b Ib.
    
    elf; auto.
    
    assert (X.lt x a) as LT by (inversion_clear LTl; auto); clear LTl.
    inversion_clear S.
    
    assert (a = b \/ X.lt a b) as D.
    destruct Ib; auto.
    destruct D; subst; auto.
    eapply X.lt_trans; eauto.

  Qed.


  Lemma eq_elt_dec : forall (a b:X.t),
    a = b \/ a <> b.
  Proof.
    intros a b.
    destruct (X.compare a b); auto; right.
  Qed.


  Lemma sortEq : forall (x y:list X.t),
    (forall a, List.In a x <-> List.In a y) ->
    sort X.lt x ->
    sort X.lt y ->
    x = y.
  Proof.
    induction x; destruct y; intros E S1 S2; simpl in E.
    auto.
    elf; rewrite (E t0); auto.
    elf; rewrite <- (E a); auto.
    
    inversion_clear S1.
    inversion_clear S2.
    rename x into l; rename y into l'; rename a into x; rename t0 into x'.
    
    assert (forall y, List.In y l -> X.lt x y) as LT.
    intros y Iy; eapply ltSorto; eauto.
      assert (forall y, List.In y l' -> X.lt x' y) as LT'.
      intros y Iy; eapply ltSorto; eauto.
      
      assert (x = x' \/ (x <> x' /\ List.In x l' /\ List.In x' l)) as D.
      destruct (eq_elt_dec x x').

      auto.

      right.
      constructor.
      auto.
      destruct (E x) as [E' _].
      destruct E'; auto. subst; elf; auto.
      destruct (E x') as [_ E'].
      destruct E'; auto. subst; elf; auto.

      destruct D as [E' | [NE [Ix Ix']]].
    
      subst x'; f_equal; apply IHx; auto.
      intros a; destruct (E a).
      constructor; intros Ia.
      destruct H3; auto; subst a; hyr LT (LT _ Ia); elf.
      apply (@X.lt_not_eq _ _ LT); reflexivity.
      destruct H4; auto; subst a; hyr LT' (LT' _ Ia); elf.
      apply (@X.lt_not_eq _ _ LT'); reflexivity.
      
      hyr LT (LT _ Ix').
      hyr LT' (LT' _ Ix).
      elf.
      assert (X.lt x x) as XLT by (eapply X.lt_trans; eauto).
      apply (@X.lt_not_eq _ _ XLT); reflexivity.

    Qed.


    Lemma InEqv : forall a x,
      InA X.eq a x <-> List.In a x.
    Proof.
      induction x; simpl.
      
      constructor; simpl; intros Ia.
      inversion Ia.
      elf; auto.
      
      destruct (eq_elt_dec a0 a) as [E | NE].
      subst a0; clear IHx; constructor; intros Ia.
      auto.
      constructor; reflexivity.
    
      destruct IHx as [IH1 IH2].
      constructor; intros Ia.
      inversion_clear Ia.
      elf; auto.
      right; auto.
      destruct Ia.
      subst; elf; auto.
      constructor 2; auto.
  
    Qed.
  

    Lemma thisBuildEq : forall x (xOk : sort X.lt x),
      this (Build_slist xOk) = x.
    Proof.
      reflexivity.
    Qed.

    Lemma set_ext : forall x y,
      Equal x y ->
      x = y.
    Proof.
      unfold M.Equal, M.In, E.eq.
      intros [x xOk] [y yOk] E.
      
      assert (E' : x = y).
      apply sortEq; auto.
      rewrite thisBuildEq in E; rewrite thisBuildEq in E.
      clear - E; intros a; destruct (E a) as [E1 E2]; clear E.
      constructor; intros Ia.
      destruct (InEqv a y) as [X _].
      destruct (InEqv a x) as [_ Y].
      auto.
      destruct (InEqv a y) as [_ X].
      destruct (InEqv a x) as [Y _].
      auto.
    
      subst y; f_equal; apply proof_irrel.
  
    Qed.


    (* these eqv things duplicate existing stuff from Facts, so
       at some point I should replace these with those, change all of the hints,
       and make sure everything still goes through. *)
    Lemma empty_eqv : forall x,
      In x empty <-> False.
    Proof.
      intros x; constructor; intro I.
      apply (empty_1 I).
      contradiction.
    Qed.


    Lemma add_eqv : forall x y S,
      In x (add y S) <-> x = y \/ In x S.
    Proof.
      intros x y S; constructor; intro I.
      destruct (eq_elt_dec x y).
      auto.
      right; eapply add_3; eauto.
      destruct I.
      subst; apply add_1; reflexivity.
      apply add_2; auto.
    Qed.

    
    Lemma remove_eqv : forall x y S,
      In x (remove y S) <-> x <> y /\ In x S.
    Proof.
      intros x y S; constructor; intros I.
      destruct (eq_elt_dec y x).
      elf; apply (remove_1 H I).
      constructor; auto; eapply remove_3; eauto.
      destruct I; apply remove_2; auto.
    Qed.

    
    Infix "|_|" := union (at level 60, right associativity).


    Lemma union_eqv : forall x S S',
      In x (S |_| S') <-> In x S \/ In x S'.
    Proof.
      intros x S S'; constructor; intro I.
      apply union_1; auto.
      destruct I; [apply union_2 | apply union_3]; auto.
    Qed.
  
    Lemma diff_eqv : forall x S S',
      In x (diff S S') <-> In x S /\ ~In x S'.
    Proof.
      intros x S S'; constructor; intro I.
      constructor; [eapply diff_1 | eapply diff_2]; eauto.
      destruct I; apply diff_3; auto.
    Qed.

    Lemma inter_eqv : forall x S S',
      In x (inter S S') <-> In x S /\ In x S'.
    Proof.
      intros x S S'; constructor; intro I.
      constructor; [eapply inter_1 | eapply inter_2]; eauto.
      destruct I; apply inter_3; auto.
    Qed.
    
    Lemma empty_union_eqv2 : forall S,
      union S empty = S.
    Proof.
      intros S; apply set_ext; apply Miff.empty_union_2; apply empty_1.
    Qed.
    
    Lemma empty_union_eqv1 : forall S,
      union empty S = S.
    Proof.
      intros S; apply set_ext; apply Miff.empty_union_1; apply empty_1.
    Qed.

    Lemma empty_diff_1' : forall s,
      diff empty s = empty.
    Proof.
      intros s; apply set_ext.
      apply empty_is_empty_1; apply empty_diff_1; auto.
    Qed.

    Lemma empty_diff_2' : forall s,
      diff s empty = s.
    Proof.
      intros s; apply set_ext; apply empty_diff_2; auto.
    Qed.

    Lemma remove_diff_singleton' : forall x s,
      diff s (singleton x) = remove x s.
    Proof.
      intros x s; rewrite <- (set_ext (remove_diff_singleton s x)); auto.
    Qed.

    Lemma empty_inter2' : forall s,
      inter s empty = empty.
    Proof.
      intros s; apply set_ext.
      apply empty_is_empty_1; apply empty_inter_2; auto.
    Qed.


    Lemma cardinal_emptyEq : 
      cardinal empty = 0%nat.
    Proof.
      rewrite cardinal_1; auto.
    Qed.


    Lemma inSingEq : forall y x,
      In y (singleton x) <-> y = x.
    Proof.
      clear; intros y x; constructor; intros Iy.
      symmetry; apply singleton_1; auto.
      subst x; apply singleton_2; reflexivity.
    Qed.
    

    Lemma inter_empty1 : forall s,
      inter empty s = empty.
    Proof.
      intros s; apply set_ext; intros x; constructor; intros Ix.
      apply inter_1 in Ix; auto.
      elf; apply (empty_1 Ix).
    Qed.

    Lemma inter_empty2 : forall s,
      inter s empty = empty.
    Proof.
      intros s; apply set_ext; intros x; constructor; intros Ix.
      apply inter_2 in Ix; auto.
      elf; apply (empty_1 Ix).
    Qed.


    Lemma singleton_choose : forall x,
      choose (singleton x) = Some x.
    Proof.
      clear; intros x.
      assert (H1 := @choose_1 (singleton x)).
      assert (H2 := @choose_2 (singleton x)).
      destruct (choose (singleton x)).
      clear H2; assert (Ix : In e (singleton x)).
      auto.
      rewrite inSingEq in Ix; subst; auto.
      assert (E : Empty (singleton x)) by auto.
      assert (E' := @Miff.empty_is_empty_1 (singleton x) E).
      apply set_ext in E'.
      assert (Ix : In x M.empty).
      rewrite <- E'; rewrite inSingEq; auto.
      rewrite empty_eqv in Ix; contradiction.
    Qed.


    Hint Rewrite add_eqv empty_eqv remove_eqv union_eqv diff_eqv inter_eqv 
      empty_union_eqv1 empty_union_eqv2 singleton_choose
      empty_diff_1' empty_diff_2' remove_diff_singleton' empty_inter2'
      cardinal_emptyEq inSingEq inter_empty1 inter_empty2
      : SetEq.
    (* it looks like this doesn't do anything, but it is still nice to have 
       here to enable cutting and pasting *)

    Ltac setsimpl := autorewrite with SetEq.

    Lemma addRemoveEqv : forall x s,
      In x s ->
      add x (remove x s) = s.
    Proof.
      intros x s Ix.
      apply set_ext.
      intros a; setsimpl; constructor; intros Ia.
      destruct Ia.
      subst; auto.
      destruct H; auto.
      destruct (eq_elt_dec a x); [left | right]; auto.
    Qed.

    Lemma removeAddEqv : forall x s,
      ~In x s ->
      remove x (add x s) = s.
    Proof.
      intros x s NI; apply set_ext; intros a; setsimpl; constructor; 
        intros Ia.
      destruct Ia as [NE D]; destruct D; auto; elf; auto.
      constructor; auto; intro; subst; auto.
    Qed.

    Lemma singleton_inter_1 : forall x s,
      In x s ->
      inter s (singleton x) = singleton x.
    Proof.
      intros x s Is; apply set_ext; intros y; setsimpl; constructor; 
        intros Iy.
      destruct Iy; auto.
      subst x; auto.
    Qed.

    Lemma singleton_inter_2 : forall s x,
      ~In x s ->
      inter s (singleton x) = empty.
    Proof.
      clear; intros s x NI; apply set_ext; intros y; setsimpl;
        constructor; intros Iy.
      destruct Iy; subst; auto.
      contradiction.
    Qed.

    Lemma singletonAdd : forall x s,
      singleton x |_| s = add x s.
    Proof.
      intros x s; apply set_ext; intros a; setsimpl; tauto.
    Qed.


    (* for some reason, there's no built in notion of disjointedness *)
    Definition Disj (s s':t) := diff s s' = s.

    Lemma DisjI : forall s s',
      (forall x, In x s -> In x s' -> False) ->
      Disj s s'.
    Proof.
      unfold Disj; intros s s' F.
      apply set_ext; intros a.
      rewrite Miff.FM.diff_iff.
      constructor; intros Ia.
      destruct Ia; auto.
      constructor; eauto.
    Qed.


    Lemma DisjE : forall s s',
      Disj s s' -> forall x,
	In x s -> In x s' ->
	False.
    Proof.
      unfold Disj; intros s s' D a Ix Ix'.
      rewrite <- D in Ix.
      rewrite Miff.FM.diff_iff in Ix; destruct Ix; auto.
    Qed.


    Lemma inter_Disj : forall s s',
      Disj s s' ->
      inter s s' = empty.
    Proof.
      intros s s' D.
      apply set_ext; intros x; constructor; intros Ix.
      elf; apply (@DisjE _ _ D x).
      apply inter_1 in Ix; auto.
      apply inter_2 in Ix; auto.
      elf; apply (empty_1 Ix).
    Qed.

    (* shim for compat with 8.2 *)
    Lemma cardinal_induction :
      forall P : t -> Type,
        (forall s, Empty s -> P s) ->
        (forall s s', P s -> forall x, ~In x s -> Add x s s' -> P s') ->
        forall n s, cardinal s = n -> P s.
    Proof.
      intros P Pemp Padd n s.
      generalize dependent n.
      refine (@set_induction (fun s => forall n, cardinal s = n -> P s) _ _ s).
      auto.
      eauto.
    Qed.



    Lemma cardinalUnion : forall s s',
      Disj s s' ->
      cardinal (s |_| s') = (cardinal s + cardinal s')%nat.
    Proof.
      intros s s'.
      refine (@cardinal_induction
        (fun s => Disj s s' -> cardinal (s |_| s') = 
          cardinal s + cardinal s') _ _ (cardinal s) s _); auto.
      (* empty *)
      clear.
      intros s sEmpty D.
      apply union_cardinal.
      intros x [X1 X2].
      apply (DisjE D X1 X2).

      clear.
      unfold Add.
      intros s s'' IH x NI AH D.
      replace s'' with (add x s).
      replace (add x s |_| s') with (add x (s |_| s')).
      rewrite add_cardinal_2.
      rewrite add_cardinal_2.
      simpl.
      rewrite IH; auto.
      apply DisjI; intros z Iz Iz'; apply (@DisjE _ _ D z); auto.
      rewrite AH; auto.
      auto.
      intro Ix.
      apply (@DisjE _ _ D x).
      rewrite AH; auto.
      rewrite union_eqv in Ix; destruct Ix; auto; elf; auto.
      symmetry; apply set_ext; apply union_add.
      apply set_ext; intros z; constructor; intros Iz.
      rewrite add_eqv in Iz; rewrite AH; destruct Iz; auto.
      rewrite add_eqv; rewrite AH in Iz; destruct Iz; auto.
    Qed.


    Lemma DisjE_unionRL : forall s s' s'',
      Disj s (s' |_| s'') ->
      Disj s s'.
    Proof.
      intros s s' s'' D; apply DisjI; intros a Ia Ia'.
      apply (DisjE D) with a; auto.
      setsimpl; auto.
    Qed.

    Lemma DisjE_unionRR : forall s s' s'',
      Disj s (s' |_| s'') ->
      Disj s s'.
    Proof.
      intros s s' s'' D; apply DisjI; intros a Ia Ia'.
      apply (DisjE D) with a; auto.
      setsimpl; auto.
    Qed.
      
    Lemma Disj_subsetL : forall s s',
      Disj s s' -> forall s'',
        Subset s'' s ->
        Disj s'' s'.
    Proof.
      intros s s' D s'' Sub; apply DisjI; intros a Ia Ia'.
      apply (DisjE D) with a; auto.
    Qed.


    Lemma Disj_comm : forall s s',
      Disj s s' ->
      Disj s' s.
    Proof.
      intros s s' D; apply DisjI; intros a Ia Ia';
        apply (DisjE D Ia' Ia).
    Qed.


    Section MultiMap.

      Variable fmap : elt -> t.

      Definition multi_map (s:t) : t :=
        fold (fun a => union (fmap a)) s empty.


      Lemma fold_leftUnPresI : forall l x s,
        In x s ->
        In x (fold_left (fun a e => fmap e |_| a) l s).
      Proof.
        induction l; intros x s Ix.
        assumption.
        simpl; apply IHl; apply union_3; auto.
      Qed.
    

      Lemma multi_mapI : forall x s,
        In x s ->
        Subset (fmap x) (multi_map s).
      Proof.
        unfold multi_map; intros x s Ix y Iy.
        rewrite M.fold_1.
        apply M.elements_1 in Ix.
        generalize dependent (elements s); clear s; intros l Il.
        generalize empty.
        induction l.
        inversion Il.
        intros l0; inversion_clear Il; simpl.
        unfold E.eq, X.eq in H; subst a.
        clear IHl; apply fold_leftUnPresI; apply union_2; auto.
        apply IHl; auto.
      Qed.

      Lemma multi_mapE : forall y s,
        In y (multi_map s) ->
        exists x, In x s /\ In y (fmap x).
      Proof.
        assert (forall l s y,
          In y (fold_left (fun a e => fmap e |_| a) l s) ->
          ~In y s ->
          exists x : elt, InA E.eq x l /\ In y (fmap x)) as H.
        induction l; simpl; intros s y Iy NIy.
        elf; auto.
        hyr IHl (IHl _ _ Iy); clear Iy.
        destruct (In_dec y (fmap a)) as [I | NI].
        exists a; constructor; [constructor |]; auto.
        destruct IHl as [x [Ix Iy]].
        intro I; apply NI; apply union_1 in I; destruct I; auto; elf; auto.
        exists x; constructor; [constructor 2 | ]; auto.

        unfold multi_map; intros y s; rewrite M.fold_1; intros Iy.
        destruct (H (elements s) empty y) as [x [Ix Iy']].
        assumption.
        intro I; apply (empty_1 I).
        exists x; constructor; [| auto].
        apply elements_2; auto.
      Qed.

      
      Lemma multi_mapEqv : forall x s,
        In x (multi_map s) <-> exists y, In y s /\ In x (fmap y).
      Proof.
        intros x s; constructor; intros Ix.
        apply multi_mapE; auto.
        destruct Ix as [y [Iy Ix]]; apply (multi_mapI Iy); auto.
      Qed.


      Lemma multi_mapUnionDistrib : forall s s',
        multi_map (s |_| s') = multi_map s |_| multi_map s'.
      Proof.
        intros s s'; apply set_ext; intros x; constructor; intros Ix.

        rewrite multi_mapEqv in Ix.
        rewrite union_eqv; rewrite multi_mapEqv; rewrite multi_mapEqv.
        destruct Ix as [y [Iy Ix]]; rewrite union_eqv in Iy; destruct Iy;
          [left | right]; exists y; constructor; auto.
        
        rewrite union_eqv in Ix; rewrite multi_mapEqv.
        rewrite multi_mapEqv in Ix; rewrite multi_mapEqv in Ix.
        destruct Ix; destruct H as [y [Iy ix]]; exists y; constructor; auto;
          rewrite union_eqv; auto.

      Qed.


    End MultiMap.


    Hint Rewrite multi_mapEqv : SetEq.

    Lemma multi_map_func_inj : forall s f f',
      (forall x, In x s -> f x = f' x) ->
      multi_map f s = multi_map f' s.
    Proof.
      intros s f f' E; apply set_ext; intros a; setsimpl; 
        constructor; intros Ia;
          destruct Ia as [y [Iy Iy']]; exists y; constructor; auto.
      rewrite <- E; auto.
      rewrite E; auto.
    Qed.

    Lemma eq_set_dec : forall (s s':t),
      s = s' \/ s <> s'.
    Proof.
      intros s s'.
      destruct (compare s s').

      right; intro; subst.
      assert (X := lt_not_eq l).
      apply X; apply eq_refl.
      
      left; apply set_ext; auto.
      
      right; intro; subst.
      assert (X := lt_not_eq l).
      apply X; apply eq_refl.
    Qed.


End Make.


Ltac setsimpl := autorewrite with SetEq.

Tactic Notation "setsimpl" "in" hyp(H) := autorewrite with SetEq in H.

(* the tactic set_iff in FSetFacts actually does something like this,
   though only in the goal, so it is of limited use.  but at least we don't 
   need to prove the theorems given above *)

Ltac subsetSolver :=
  let x := fresh "x" in
    let Ix := fresh "Ix" in
      intros x Ix; setsimpl in Ix; setsimpl; intuition.



(* mapping from sets of one type to another *)

Module FSetMap (S1E : FSetEq) (S2E : FSetEq).
  Module S1 := S1E.M.
  Module S2 := S2E.M.

  Section Map.

    Variable f : S1.elt -> S2.elt.

    Definition map (s:S1.t) : S2.t :=
      S1.fold (fun a => S2.add (f a)) s S2.empty.


    Lemma eqDec2 : forall (a b:S2.E.t),
      S2.E.eq a b \/ ~S2.E.eq a b.
    Proof.
      intros a b.
      destruct (S2.E.compare a b); auto.
      right; intro; subst; apply S2.E.lt_not_eq in l; auto.
    Qed.


    Section MapIntro.

      Lemma f_subst : forall x y,
        S1.E.eq x y ->
        S2.E.eq (f x) (f y).
      Proof.
        intros x y E.
        assert (E' := S1E.elt_eq E); subst y; auto.
      Qed.

      (* continuing the fold doesn't remove anything from the set *)
      Lemma fold_leftPres : forall l y s,
        S2.In y s ->
        S2.In y (fold_left (fun a x => S2.add (f x) a) l s).
      Proof.
        induction l; intros y s I.
        assumption.
        simpl.
        apply IHl.
        apply S2.add_2; auto.
      Qed.

      (* if x is in the list we're folding over,
         f x is in the result of the fold *)
      Lemma fold_leftOk1 : forall x l,
        InA S1.E.eq x l -> forall s',
          S2.In (f x) (fold_left (fun a e => S2.add (f e) a) l s').
      Proof.
        induction l; simpl; intros.
        inversion H.
        inversion_clear H.
        apply fold_leftPres.
        assert (E := S1E.elt_eq H0); subst a.
        apply S2.add_1; auto.
        auto.
      Qed.

      Lemma mapI : forall x s,
        S1.In x s ->
        S2.In (f x) (map s).
      Proof.
        unfold map; intros x s I.
        rewrite S1.fold_1; apply fold_leftOk1.
        apply S1.elements_1; auto.
      Qed.


    End MapIntro.


    Lemma mapE0' : forall x l s,
      ~S2.In x s ->
      S2.In x (fold_left (fun a e => S2.add (f e) a) l s) ->
      exists y, InA S1.E.eq y l /\ S2.E.eq x (f y).
    Proof.
      induction l; intros s NI I.
      elf; auto.

      destruct (eqDec2 (f a) x) as [E | NE].
      exists a; constructor; [constructor | ]; auto.

      simpl in I.  hyr IHl (fun NI => IHl _ NI I).
      destruct IHl as [y [Iy E]].
      intro Ifx; apply S2.add_3 in Ifx; auto.

      exists y; auto.

    Qed.
      

    Lemma mapE0 : forall x s,
      S2.In x (map s) ->
      exists y, S1.In y s /\ S2.E.eq x (f y).
    Proof.
      unfold map; intros x s I.
      rewrite S1.fold_1 in I.
      hyr I (fun NI => mapE0' _ NI I).
      destruct I as [y [Iy E]].
      intro I; apply (S2.empty_1 I).
      exists y; auto with *.
    Qed.


    (* if f is injective, we can prove the expected map elimination lemma *)
    Lemma mapE : 
      (forall x y, S2.E.eq (f x) (f y) -> S1.E.eq x y) -> forall x s,
      S2.In (f x) (map s) ->
      S1.In x s.
    Proof.
      intros f_inj x s Ix; destruct (mapE0 Ix) as [y [Iy Feq]].
      assert (E : S1.E.eq y x).
      auto.
      apply (S1.In_1 E Iy).
    Qed.


    (* basically same as ltSorto before *)
    Lemma ltSortoAgainBasically : forall l a,
      lelistA S1.E.lt a l ->
      sort S1.E.lt l -> forall b,
        InA S1.E.eq b l ->
        S1.E.lt a b.
    Proof.
      induction l; simpl; intros x LTl S b Ib.
      
      inversion Ib.
      
      assert (S1.E.lt x a) as LT by (inversion_clear LTl; auto); clear LTl.
      inversion_clear S.
      
      assert (S1.E.eq a b \/ S1.E.lt a b) as D.
      inversion_clear Ib; auto.
      destruct D.
      destruct (S1.E.compare x b).
      auto.
      assert (S1.E.eq x a) as E.
      eapply S1.E.eq_trans; eauto.
      elf; exact (S1.E.lt_not_eq LT E).
      elf.
      assert (S1.E.lt b a) as LT'.
      apply S1.E.lt_trans with x; auto.
      exact (S1.E.lt_not_eq LT' (S1.E.eq_sym H1)).
      eapply S1.E.lt_trans; eauto.
    
    Qed.



    Definition fl_inj (l:list S1.elt) := forall x y,
      InA S1.E.eq x l ->
      InA S1.E.eq y l ->
      S2.E.eq (f x) (f y) ->
      S1.E.eq x y.

    Lemma fl_injNeq : forall l,
      fl_inj l -> forall x y,
        InA S1.E.eq x l ->
        InA S1.E.eq y l ->
        ~S1.E.eq x y ->
        ~S2.E.eq (f x) (f y).
    Proof.
      intros l FL x y Ix Iy NI E; auto.
    Qed.


    Lemma fl_injTail : forall a l,
      fl_inj (a::l) ->
      fl_inj l.
    Proof.
      unfold fl_inj; intros a l FL x y Ix Iy E.
      apply (FL x y); auto.
    Qed.


    Lemma map_cardinal' : forall (l:list S1.elt) (s:S2.t),
      fl_inj l ->
      (forall x, InA S1.E.eq x l -> S2.In (f x) s -> False) ->
      sort S1.E.lt l ->
      length (S2.elements (fold_left (fun a e => S2.add (f e) a) l s)) =
      length l + length (S2.elements s).
    Proof.
      induction l; simpl; intros s FL D So.
      auto.

      hyr IHl (IHl (S2.add (f a) s)).
      replace (S (length l + length (S2.elements s))) with
        (length l + length (S2.elements (S2.add (f a) s))).
      rewrite <- IHl; clear IHl.
      auto.
      apply (fl_injTail FL); auto.
      intros x Ix Ifx; apply (D x); clear D.
      constructor 2; auto.
      cut (~S2.E.eq (f a) (f x)).
      intros NE.
      hyr Ifx (S2.add_3 NE Ifx); auto.
      apply (fl_injNeq FL); auto.
      inversion_clear So.
      clear - H H0 Ix; intro E.
      assert (S1.E.lt a x) as LT.
      eapply ltSortoAgainBasically; eauto.
      exact (S1.E.lt_not_eq LT E).
      
      inversion_clear So; auto.

      clear IHl.
      replace (length (S2.elements (S2.add (f a) s))) with
        (S (length (S2.elements s))).
      auto.
      assert (NI : ~S2.In (f a) s).
      intros I; apply (D a); auto; constructor.
      clear - NI.
      rewrite <- S2.cardinal_1; rewrite <- S2.cardinal_1.
      generalize dependent NI.
      intros NI.
      rewrite S2E.Miff.add_cardinal_2; auto.
    Qed.


    Definition fs_inj (s:S1.t) := forall x y,
      S1.In x s ->
      S1.In y s ->
      S2.E.eq (f x) (f y) ->
      S1.E.eq x y.


    Lemma fl_injI : forall s,
      fs_inj s ->
      fl_inj (S1.elements s).
    Proof.
      unfold fs_inj, fl_inj; intros s FS x y Ix Iy E.
      apply (FS x y); auto with *.
    Qed.


    Lemma map_cardinal : forall s,
      fs_inj s ->
      S2.cardinal (map s) = S1.cardinal s.
    Proof.
      unfold map; intros s FS.
      rewrite S2.cardinal_1; rewrite S1.cardinal_1.
      rewrite S1.fold_1; rewrite map_cardinal'; auto.
      rewrite <- S2.cardinal_1; rewrite S2E.Miff.cardinal_1; auto with *.
      apply fl_injI; auto with *.
      intros x Ix Ifx.
      apply (S2.empty_1 Ifx).
      auto with *.
    Qed.


    Lemma singleton_map : forall x,
      map (S1.singleton x) = S2.singleton (f x).
    Proof.
      intros x; apply S2E.set_ext; intros y; constructor; intros Iy.

      apply S2.singleton_2.
      apply mapE0 in Iy.
      destruct Iy as [z [Iz E]].
      apply S1.singleton_1 in Iz.
      assert (E' := S1E.elt_eq Iz); subst z.
      assert (E' := S2E.elt_eq E); subst y; auto.

      apply S2.singleton_1 in Iy.
      apply S2E.elt_eq in Iy; subst.
      apply mapI.
      apply S1.singleton_2; auto.
    Qed.

    Lemma map_empty : map S1.empty = S2.empty.
    Proof.
      apply S2E.set_ext; intros y; constructor; intros Iy.

      elf; apply mapE0 in Iy; destruct Iy as [z [Iz E]].
      apply (S1.empty_1 Iz).

      elf; apply (S2.empty_1 Iy).

    Qed.

  End Map.


End FSetMap.





