(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import CoreLog.
Require Export MyTactics.


Set Implicit Arguments.


Ltac elf := elimtype False.

Ltac hyr H H' :=
  let H'' := fresh in
    (assert (H'' := H'); clear H; rename H'' into H).


Module Type MapSig.
  Section Map.
    Variable A : Type.
    Variable B : Type.
    Variable Aeq : forall (x y:A), {x = y} + {x <> y}.
    
    Parameter map : Type -> Type -> Type.

    Parameter map_get : map A B -> A -> B.
    Parameter map_set : (forall (x y:A), {x = y} + {x <> y}) ->
      map A B -> A -> B -> map A B.
    Parameter map_union : (B -> B -> B) -> map A B -> map A B -> map A B.

    Parameter const_map : B -> map A B.


    Parameter map_ext : forall m m',
      (forall x, map_get m x = map_get m' x) ->
      m = m'.

    Parameter map_setEq : forall m a b,
      map_get (map_set Aeq m a b) a = b.

    Parameter map_setNeq : forall a a',
      a <> a' -> forall m b,
	map_get (map_set Aeq m a b) a' = map_get m a'.

    Parameter const_mapEq : forall a b,
      map_get (const_map b) a = b.

    Parameter map_unionEq : forall f m m' x,
      map_get (map_union f m m') x = f (map_get m x) (map_get m' x).

  End Map.


  Section MapMap.
    Variables A A' B : Type.

    Parameter map_map_dom : (A' -> A) -> map A B -> map A' B.

    Parameter map_map_domEq : forall f m x,
      map_get (map_map_dom f m) x = map_get m (f x).

  End MapMap.


End MapSig.


Module Map : MapSig.
  Section Map.

    Variables A A' B : Type.
    Variable Aeq : forall (x y:A), {x = y} + {x <> y}.

    Definition map := A -> B.

    Definition map_get (m:map) (a:A) : B := m a.
    Definition map_set (m:map) (a:A) (b:B) : map :=
      fun a' => if Aeq a a' then b else map_get m a'.
    Definition map_union (f:B -> B -> B) (m m':map) :=
      fun a => f (m a) (m' a).
    Definition map_map_dom (f:A' -> A) (m:map) (a:A') : B :=
      map_get m (f a).

    Definition const_map (b:B) : map := fun _ => b.


    Lemma map_ext : forall m m',
      (forall x, map_get m x = map_get m' x) ->
      m = m'.
    Proof.
      unfold map_get; intros m m' E; apply (fn_ext _ _ E).
    Qed.


    Lemma map_setEq : forall m a b,
      map_get (map_set m a b) a = b.
    Proof.
      unfold map_set, map_get; intros m a b.
      destruct Aeq; auto.
      elf; auto.
    Qed.

    Lemma map_setNeq : forall a a',
      a <> a' -> forall m b,
	map_get (map_set m a b) a' = map_get m a'.
    Proof.
      unfold map_set, map_get; intros a a' NE m b.
      destruct Aeq.
      elf; auto.
      reflexivity.
    Qed.

    Lemma const_mapEq : forall a b,
      map_get (const_map b) a = b.
    Proof.
      reflexivity.
    Qed.

    Lemma map_unionEq : forall f m m' x,
      map_get (map_union f m m') x = f (map_get m x) (map_get m' x).
    Proof.
      reflexivity.
    Qed.

    Lemma map_map_domEq : forall f m x,
      map_map_dom f m x = m (f x).
    Proof.
      reflexivity.
    Qed.

  End Map.

End Map.


Export Map.


Hint Rewrite map_setEq map_setNeq const_mapEq map_unionEq using 
  (auto; fail) : FnMap.

Ltac msimpl := autorewrite with FnMap in * |- * ; auto.

