(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import FnMap.


Set Implicit Arguments.


(* partial maps *)

Module Type PMapSig.
  Section PMap.
    Variable A : Type.
    Variable B : Type.
    Variable Aeq : forall (x y:A), {x = y} + {x <> y}.

    Parameter pmap : Type -> Type -> Type.


    (* operations *)

    Parameter empty_pmap : pmap A B.

    Parameter const_pmap : B -> pmap A B.

    Parameter pmap_get : pmap A B -> A -> option B.

    Parameter pmap_set : (forall (x y:A), {x = y} + {x <> y}) ->
      pmap A B -> A -> B -> pmap A B.

    Parameter pmap_update : 
      (forall (x y:A), {x = y} + {x <> y}) ->
      pmap A B -> A -> B -> option (pmap A B).

    Parameter pmap_union : (option B -> option B -> option B) -> 
      pmap A B -> pmap A B -> pmap A B.

    Parameter pmap_diff : pmap A B -> pmap A B -> pmap A B.

    Parameter pmap_remove : (forall (x y:A), {x = y} + {x <> y}) ->
      pmap A B -> A -> pmap A B.


    (* properties *)

    Parameter pmap_ext : forall m m',
      (forall x, pmap_get m x = pmap_get m' x) ->
      m = m'.

    Parameter empty_pmapEq : forall a,
      pmap_get empty_pmap a = None.

    Parameter pmap_setEq : forall m a b,
      pmap_get (pmap_set Aeq m a b) a = Some b.

    Parameter pmap_setNeq : forall a a',
      a <> a' -> forall m b,
	pmap_get (pmap_set Aeq m a b) a' = pmap_get m a'.

    Parameter pmap_updateSome : forall m a,
      pmap_get m a <> None -> forall b,
      pmap_update Aeq m a b = Some (pmap_set Aeq m a b).

    Parameter pmap_updateNone : forall m a,
      pmap_get m a = None -> forall b,
	pmap_update Aeq m a b = None.

    Parameter pmap_unionEq : forall f m m' x,
      pmap_get (pmap_union f m m') x = f (pmap_get m x) (pmap_get m' x).

    Parameter pmap_diffEq : forall m' x,
      pmap_get m' x <> None -> forall m,
      pmap_get (pmap_diff m m') x = None.

    Parameter pmap_diffNeq : forall m' x,
      pmap_get m' x = None -> forall m,
	pmap_get (pmap_diff m m') x = pmap_get m x.

    Parameter pmap_removeEq : forall m x,
      pmap_get (pmap_remove Aeq m x) x = None.

    Parameter pmap_removeNeq : forall x y,
      x <> y -> forall m,
      pmap_get (pmap_remove Aeq m x) y = pmap_get m y.

    Definition in_pmap_dom (x:A) (m:pmap A B) : Prop :=
      pmap_get m x <> None.

  End PMap.

End PMapSig.


Module PMap : PMapSig.

  Section PMap.
    Variable A : Type.
    Variable B : Type.
    Variable Aeq : forall (x y:A), {x = y} + {x <> y}.

    Definition pmap := map A (option B).
    
    Definition empty_pmap : pmap := const_map _ None.
    
    Definition const_pmap (b:B) : pmap := const_map _ (Some b).
    
    Definition pmap_get : pmap -> A -> option B := 
      map_get (A := A) (B := option B).
    
    Definition pmap_set (m:pmap) (a:A) (b:B) := map_set Aeq m a (Some b).

    Definition pmap_update (m:pmap) (a:A) (b:B) : option pmap :=
      match pmap_get m a with
	| Some _ => Some (pmap_set m a b)
	| None => None
      end.

    Definition pmap_diff (m m':pmap) : pmap :=
      map_union (fun b b' =>
	match b' with
	  | Some _ => None
	  | None => b
	end) m m'.

    Definition pmap_union (f:option B -> option B -> option B)
      (m m':pmap) : pmap := map_union f m m'.

    Definition pmap_remove (m:pmap) (a:A) : pmap := map_set Aeq m a None.


    (* properties *)

    Lemma pmap_ext : forall m m',
      (forall x, pmap_get m x = pmap_get m' x) ->
      m = m'.
    Proof.
      unfold pmap_get; intros m m' E; apply (map_ext E); auto.
    Qed.

    Lemma empty_pmapEq : forall a,
      pmap_get empty_pmap a = None.
    Proof.
      intros a; unfold pmap_get; unfold empty_pmap; msimpl.
    Qed.


    Lemma pmap_setEq : forall m a b,
      pmap_get (pmap_set m a b) a = Some b.
    Proof.
      unfold pmap_set, pmap_get; intros m a b; msimpl.
    Qed.

    Lemma pmap_setNeq : forall a a',
      a <> a' -> forall m b,
	pmap_get (pmap_set m a b) a' = pmap_get m a'.
    Proof.
      unfold pmap_set, pmap_get; intros a a' NE m b; msimpl.
    Qed.

    Lemma pmap_updateSome : forall m a,
      pmap_get m a <> None -> forall b,
	pmap_update m a b = Some (pmap_set m a b).
    Proof.
      unfold pmap_update; intros m a NE b.
      destruct (pmap_get m a); intuition.
    Qed.

    Lemma pmap_updateNone : forall m a,
      pmap_get m a = None -> forall b,
	pmap_update m a b = None.
    Proof.
      unfold pmap_update; intros m a E b; rewrite E; auto.
    Qed.

    Lemma pmap_unionEq : forall f m m' x,
      pmap_get (pmap_union f m m') x = f (pmap_get m x) (pmap_get m' x).
    Proof.
      intros f m m' x; unfold pmap_get, pmap_union; msimpl.
    Qed.

    Lemma pmap_diffEq : forall m' x,
      pmap_get m' x <> None -> forall m,
      pmap_get (pmap_diff m m') x = None.
    Proof.
      unfold pmap_get, pmap_diff; intros m' x m NE; msimpl.
      destruct (map_get m' x); intuition.
    Qed.

    Lemma pmap_diffNeq : forall m' x,
      pmap_get m' x = None -> forall m,
	pmap_get (pmap_diff m m') x = pmap_get m x.
    Proof.
      unfold pmap_get, pmap_diff; intros m' x E m; msimpl.
      rewrite E; auto.
    Qed.

    
    Lemma pmap_removeEq : forall m x,
      pmap_get (pmap_remove m x) x = None.
    Proof.
      unfold pmap_get, pmap_remove; intros m x; msimpl.
    Qed.

    Lemma pmap_removeNeq : forall x y,
      x <> y -> forall m,
      pmap_get (pmap_remove m x) y = pmap_get m y.
    Proof.
      unfold pmap_get, pmap_remove; intros x y NE m; msimpl.
    Qed.

    Definition in_pmap_dom (a:A) (p:pmap) : Prop :=
      pmap_get p a <> None.


  End PMap.


End PMap.


Export PMap.


Hint Rewrite empty_pmapEq pmap_setEq pmap_setNeq pmap_updateSome
  pmap_updateNone pmap_unionEq pmap_diffEq pmap_diffNeq
  pmap_removeEq pmap_removeNeq using (auto; fail) : FnMap.


Section listToPmap.

  Require Import List.

  Variables (A B : Type).
  Variable Aeq : forall (x y:A), {x = y} + {x <> y}.

  Fixpoint list_to_pmap (l:list (prod A B)) : pmap A B :=
    match l with
      | nil => empty_pmap _ _
      | (w, v)::l' => pmap_set Aeq (list_to_pmap l') w v
    end.

  Lemma list_to_pmapEq : forall x v l,
    In (x, v) l ->
    (forall v', In (x, v') l -> v = v') ->
    pmap_get (list_to_pmap l) x = Some v.
  Proof.
    induction l; simpl; intros I Ev.
    tauto.
    destruct a as [x' v'].
    destruct (Aeq x x') as [E | NE]; subst; msimpl.
    rewrite (Ev v'); auto.
    destruct I as [E | I]; auto; elf; injector E; auto.
  Qed.


  Lemma list_to_pmapNeq : forall x l,
    ~(exists v, In (x, v) l) ->
    pmap_get (list_to_pmap l) x = None.
  Proof.
    induction l; simpl; intros NI.
    msimpl.
    destruct a as [x' v'].
    assert (x' <> x).
    intro; subst; apply NI; exists v'; auto.
    msimpl; apply IHl; intros [v D]; apply NI; exists v; auto.
  Qed.


End listToPmap.



