(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import ZArith.
Require Export Memory.
Require Export FSetListEq.
Require Import ListWordSet.
Require Export Addr.
Require Val.


Set Implicit Arguments.


Module AddrOrderedType <: UsualOrderedType.

  Definition t := addr.
  Definition eq := @eq addr.

  Definition lt (x y:addr) :=
    let (b1, k1) := x in
    let (b2, k2) := y in
      BlockOrdered.lt b1 b2 \/ (b1 = b2 /\ (k1 < k2)%word).

  Definition eq_refl := @refl_equal t.
  Definition eq_sym := @sym_eq t.
  Definition eq_trans := @trans_eq t.

  Lemma lt_trans : forall x y z : addr,
    lt x y -> lt y z -> lt x z.
  Proof.
    unfold lt; intros [b1 k1] [b2 k2] [b3 k3] LT1 LT2.
    destruct LT1.
      destruct LT2.
        left; eapply BlockOrdered.lt_trans; eassumption.

        destruct H0 as [E LT]; subst; constructor; auto.

      destruct H as [E LT]; subst.
      destruct LT2; [left | right].
        auto.

        destruct H; subst; constructor; auto.
        womega.

  Qed.


 Lemma lt_not_eq : forall x y : addr, lt x y -> ~ eq x y.
 Proof.
   unfold lt, eq; intros [b k] [b' k'] LT E; injector E.
   destruct LT.
   apply (BlockOrdered.lt_not_eq _ _ H); apply BlockOrdered.eq_refl.
   destruct H as [_ H]; womega.
 Qed.

 Definition compare : forall x y : t, Compare lt eq x y.
 Proof.
   unfold lt, eq; intros [b k] [b' k'].
   destruct (BlockOrdered.compare b b').

   constructor 1; constructor; auto.

   unfold BlockOrdered.eq in e; subst b'.
   destruct (eq_word_dec k k') as [E | NE].
   subst; constructor 2; auto.
   destruct (lt_word_dec k k') as [LT | NLT].
   constructor 1; right; constructor; auto.
   constructor 3; right; constructor; auto.
   womega.

   constructor 3; constructor; auto.
   
 Qed.

 Definition eq_dec := eq_addr_dec.
 
End AddrOrderedType.



Module Import ASE := FSetListEq.Make (AddrOrderedType).
Module AS := ASE.M.

Infix "|_|" := AS.union (at level 60, right associativity).

Hint Rewrite add_eqv empty_eqv remove_eqv union_eqv diff_eqv inter_eqv 
  empty_diff_1' empty_diff_2' remove_diff_singleton' empty_inter2'
  singleton_choose
  multi_mapEqv empty_union_eqv1 empty_union_eqv2 : SetEq.
Hint Rewrite cardinal_emptyEq inSingEq inter_empty1 inter_empty2 : SetEq.
Hint Rewrite removeAddEqv addRemoveEqv 
  singleton_inter_1 singleton_inter_2 using assumption : SetEq.
Hint Rewrite singletonAdd : SetEq.
Hint Resolve DisjE_unionRR DisjE_unionRL : ListSet.


Module WASetMap := FSetMap WSE ASE.
Module AASetMap := FSetMap ASE ASE.

Hint Rewrite WASetMap.map_empty WASetMap.singleton_map : SetEq.
Hint Rewrite AASetMap.map_empty AASetMap.singleton_map : SetEq.


(* A nicer induction principle for sets, based on the existing one *)
Lemma aset_induction : forall P : AS.t -> Type,
  P AS.empty ->
  (forall s, P s -> forall x, ~AS.In x s -> P (AS.add x s)) ->
  forall (s:AS.t), P s.
Proof.
  intros P Pempty Padd.
  assert (forall (s:ASE.M.t), P s). 2:auto.
  induction s using ASE.Miff.set_induction.

  assert (s = AS.empty)
    by (apply ASE.set_ext; apply ASE.Miff.empty_is_empty_1; auto).
  subst; auto.

  rename H into NI; rename H0 into xAdd.
  assert (s2 = AS.add x s1) as s2Eq.
    unfold ASE.Miff.Add in xAdd; apply ASE.set_ext.
    intros z; constructor; intros Iz.
    rewrite xAdd in Iz; destruct Iz; [apply AS.add_1 | apply AS.add_2]; auto.
    rewrite xAdd; destruct (eq_addr_dec x z) as [E | NE].
    subst; auto.
    right; eapply AS.add_3; eauto.
  subst s2; auto.

Qed.


(* array_addrs x len is the addresses of the array starting at x with length len
   the addresses are word-aligned, the length is the number of words *)
Definition array_addrs (a:addr) (len:word) : AS.t :=
  WASetMap.map (fun k => a + (4 * k)%word)%addr (wordZeroRangeSet len).

Lemma array_addrsI : forall i len,
  (i < len)%word -> forall x,
  AS.In (x + (4 * i)%word)%addr (array_addrs x len).
Proof.
  unfold array_addrs; intros i len LT x.
  apply (WASetMap.mapI (fun k => x + (4 * k)%word)%addr); setsimpl; auto.
Qed.


Lemma array_addrsE : forall x y k,
  impl (AS.In x (array_addrs y k))
  (exists k', x = y + 4 * k' /\ (k' < k)%word).
Proof.
  unfold array_addrs; clear; intros x y k Ix.
  apply WASetMap.mapE0 in Ix; destruct Ix as [z [Iz E]].
  setsimpl in Iz.
  exists z; auto.
Qed.

Hint Rewrite array_addrsE : SetEq.
(*Hint Rewrite <- array_addrsE : SetEq.
Having rewrite hints in both directions seems like a bad idea.
*)


Definition ptr_In (x:val) (S:AS.t) :=
  match x with 
    | Vptr p => AS.In p S
    | _ => False
  end.

Lemma addrPtr_In : forall (a:addr) s,
  ptr_In a s = AS.In a s.
Proof.
  destruct a; auto.
Qed.


Lemma notPtrInNull : forall S,
  ptr_In Val.null_ptr S = False.
Proof.
  auto.
Qed.



Hint Rewrite notPtrInNull addrPtr_In : AddrSimpl.


Definition singVal (x:val) : AS.t :=
  match Val.addrOf x with
    | Some a => AS.singleton a
    | None => AS.empty
  end.

Lemma singValNull : singVal Val.null_ptr = AS.empty.
Proof.
  reflexivity.
Qed.

Lemma singValVptr : forall x,
  singVal (Vptr x) = AS.singleton x.
Proof.
  reflexivity.
Qed.


Hint Rewrite singValNull singValVptr : AddrSimpl.



Lemma array_addrs_cardinal : forall k,
  (k <= Z_to_word indexBound)%word -> forall x,
  AS.cardinal (array_addrs x k) = word_to_nat k.
Proof.
  clear.
  unfold array_addrs; intros k LEk x.
  rewrite WASetMap.map_cardinal.
  apply wordZeroRangeSetCardinal.
  (* show that the function is injective over the set *)
  unfold WASetMap.fs_inj.
  unfold WASetMap.S2.E.eq, WS.E.eq, WordOrderedType.eq,
    ASE.M.E.eq, AddrOrderedType.eq.
  destruct x as [b x]; simpl; intros y z Iy Iz E; apply addrInjHelper in E.
  apply wordAddInj in E; clear - E Iy Iz LEk.
  setsimpl in Iy; setsimpl in Iz.
  apply wordIndexMultInj; auto; clear E; womega.
Qed.


(* working around set ugliness *)
Lemma emptyCardEq :
  AS.cardinal AS.empty = ASE.M.cardinal ASE.M.empty.
Proof.
  reflexivity.
Qed.
Hint Rewrite emptyCardEq : SetEq.


Lemma array_addrs_empty : forall x,
  array_addrs x 0 = AS.empty.
Proof.
  unfold array_addrs; clear; intros x; simpl.
  apply ASE.set_ext; intros y; setsimpl; tauto.
Qed.
Hint Rewrite array_addrs_empty : SetEq.



(** Set Contiguity *)


(* a set of addresses is contiguous if it is equal to a set of array addrs *)
Definition contiguous x s :=
  s = array_addrs x (AS.cardinal s).


Module AWSetMap := FSetMap ASE WSE.
Hint Rewrite AWSetMap.map_empty AWSetMap.singleton_map : SetEq.


(* move to word *)
Lemma wordAlignEq : forall x,
  wordAlign (4 * x) = 4 * x.
Proof.
  intros x; womega.
Qed.

Lemma multSmaller : forall x,
  (4 * x <= Z_to_word (wordBound - 4))%word.
Proof.
  intros x; womega.
  computet (wval (Z_to_word (4294967296 - 4)))%Z.
  assert (x mod indexBound <= indexBound - 1)%Z.
    assert (x mod indexBound < indexBound)%Z as LT
      by (destruct (Z_mod_lt x indexBound); wauto).
    omega.
  womega.
Qed.


(* move to word *)
Lemma wordLeE : forall x y,
  (x <= y)%word ->
  (x = y \/ x < y)%word.
Proof.
  intros x y LE; womega.
  assert (x = y \/ x < y)%Z as D by omega; destruct D; [left | right].
  womega. auto.
Qed.


(* workaround for bug in 8.2beta3, suggested by M. Sozeau.  See Bug #1905. *)
(*
Hopefully this is fixed now, in 8.2 release.

Instance id_impl_impl : Morphisms.Morphism (impl ==> impl) id.
Proof.
  intros A B; unfold impl, id; auto.
Qed.*)


Lemma array_addrsBounded : forall x k,
  (Z_of_nat (AS.cardinal (array_addrs x k)) <= indexBound)%Z.
Proof.
  intros x k.
  set (s := array_addrs x k).

  (* first show that the set of offsets from x has the same cardinality
     as the original set *)
  set (s' := AWSetMap.map (fun a:addr => snd a - snd x)%word s).
  assert (CE1 : WS.cardinal s' = AS.cardinal s).
    subst s'; apply AWSetMap.map_cardinal.
    intros a b Ia Ib E; subst s; setsimpl in Ia; setsimpl in Ib.
    destruct Ia as [ka [Ea LTa]]; destruct Ib as [kb [Eb LTb]].
    destruct x as [bx x]; destruct a as [ba a]; destruct b as [bb b]; 
      simpl in E.
    assert (a - x = b - x)%word as E' by auto.
    clear E; rename E' into E.
    replace (a - x)%word with ((-x) + a)%word in E by ring.
    replace (b - x)%word with ((-x) + b)%word in E by ring.
    apply wordAddInj in E.
    simpl in Ea; injection Ea; simpl in Eb; injection Eb.
    intros; subst a bb ba b; reflexivity.
    
  rewrite <- CE1; clear CE1; subst s; rename s' into s.

  (* next, show that the cardinality of the offsets is smaller than
     the set of all offsets *)
  set (lastAddr := Z_to_word (wordBound - 4)%Z).
  set (s' := WS.add lastAddr (zeroWordSet lastAddr)).
  assert (Z_of_nat (WS.cardinal s) <= Z_of_nat (WS.cardinal s'))%Z as LT.
    apply inj_le.
    (* some kind of name aliasing causes Coq to freeze, so let's work around it *)
    assert (R := WSE.Miff.subset_cardinal).
    replace WSE.M.cardinal with WS.cardinal in R by reflexivity; 
      apply R; clear R.
    (* done with work around *)
    subst s s'; intros a Ia.
    (* again, setoid rewriting with WSE.add_eqv causes Coq to hang here *)
    destruct (WSE.add_eqv a lastAddr (zeroWordSet lastAddr)) as [_ X].
    replace WSE.M.add with WS.add in X by reflexivity; apply X; clear X.
    replace WSE.M.In with WS.In by reflexivity.
    setsimpl.
    (* okay done with that *)
    apply AWSetMap.mapE0 in Ia; destruct Ia as [b [Ib Ea]].
    assert (a = snd b - snd x)%word by auto; clear Ea; subst a.
    setsimpl in Ib; destruct Ib as [k' [Ek Ik]].
    destruct x as [bx x]; destruct b as [bb b]; simpl in * .
    injection Ek; clear Ek; intros; subst bx b.
    replace (x + 4 * k' - x)%word with (4 * k')%word by ring.
    rewrite wordAlignEq.
    replace (wordAlign lastAddr) with lastAddr
      by (apply wordInj; vm_compute; reflexivity).
    assert (4 * k' = lastAddr \/ (4 * k' < lastAddr)%word) as D.
    2:destruct D; auto.
    apply wordLeE; apply multSmaller.

  (* finally, show that the card of the set of all offsets is <= 2^30 *)
  assert (Z_of_nat (WS.cardinal s') <= indexBound)%Z as LT'. 2:omega.
  clear; subst s'.
  (* again, the names are causing Coq to hang *)
  assert (R := WSE.Miff.add_cardinal_2).
  replace WSE.M.add with WS.add in R by reflexivity.
  replace WSE.M.In with WS.In in R by reflexivity.
  replace WSE.M.cardinal with WS.cardinal in R by reflexivity.
  rewrite R; clear R.
  (* done with work around *)
  rewrite inj_S; rewrite zeroWordSetBounded.
  subst lastAddr; vm_compute; discriminate.
  
  setsimpl; intros [E LT].
  rewrite <- E in LT.
  apply wordLtInj in LT; omega.

Qed.


Lemma contiguousBounded : forall x s,
  contiguous x s ->
  (Z_of_nat (AS.cardinal s) <= indexBound)%Z.
Proof.
  intros x s Cx; rewrite Cx; apply array_addrsBounded.
Qed.


Lemma array_addrsUnion : forall a,
  (a <= windexBound)%word -> forall b,
    (b <= windexBound)%word -> forall x,
      array_addrs x a |_| array_addrs (x + 4 * a) b =
        array_addrs x (a + b)%word.
Proof.
  clear. intros a aLT b bLT x.
  apply ASE.set_ext; intros y.
  (* 8.2beta3 problem: see bug 1911 *) rewrite union_eqv. (* was setsimpl *)
  constructor; intros Ia.
  
  destruct Ia as [Ia | Ia].
  
  rewrite array_addrsE in Ia; destruct Ia as [c [E LT]]; subst.
  apply array_addrsI; womega.

  apply array_addrsE in Ia; destruct Ia as [c [E LT]]; subst.
  replace (x + 4 * a + 4 * c) with (x + 4 * (a + c)%word)
    by (destruct x; simpl; f_equal; ring).
  apply array_addrsI; womega.

  setsimpl in Ia; destruct Ia as [c [cEq cLT]]; subst y.

  assert (wval c < wval a \/ wval c >= wval a)%Z as D by omega; 
    destruct D as [LT' | GE]; [left | right].

  apply array_addrsI; womega.

  replace (x + 4 * c) with (x + 4 * a + 4 * (c - a)%word)
    by (destruct x; simpl; f_equal; ring).
  apply array_addrsI; womega.

Qed.


Lemma merge_contiguous : forall x s,
  contiguous x s -> forall s',
    contiguous (x + 4 * AS.cardinal s) s' ->
    (Z_of_nat (AS.cardinal s + AS.cardinal s') <= indexBound)%Z ->
    contiguous x (s |_| s').
Proof.
  unfold contiguous; intros x s Cs s' Cs' ss'Ok.
  assert (sOk := contiguousBounded Cs).
  assert (s'Ok := contiguousBounded Cs').

  rewrite ASE.cardinalUnion.
  replace M.cardinal with AS.cardinal by reflexivity.
  pattern s at 1; rewrite Cs; pattern s' at 1; rewrite Cs'; clear Cs Cs'.

  apply set_ext; intros a. 
  (* 8.2beta3 problem: see bug 1911 *) rewrite union_eqv. (* was setsimpl *)
  constructor; intros Ia.
  destruct Ia as [Ia | Ia]; setsimpl in Ia; destruct Ia as [k [Aeq kLt]]; subst a;
    replace M.cardinal with AS.cardinal in * by reflexivity.

  clear - kLt sOk s'Ok ss'Ok.
  apply array_addrsI; womega.

  replace (x + 4 * AS.cardinal s + 4 * k) 
    with (x + 4 * (AS.cardinal s + k)%word) 
      by (destruct x; simpl; f_equal; ring).
  apply array_addrsI; womega.

  setsimpl in Ia.
  destruct Ia as [k [aE Lt]]; subst a.
  clear - sOk s'Ok ss'Ok Lt.
  assert (wval k < Z_of_nat (AS.cardinal s) \/ 
    wval k >= Z_of_nat (AS.cardinal s))%Z as D by omega.
  destruct D; [left | right].
  apply array_addrsI; womega.

  replace (x + 4 * k)
    with ((x + 4 * AS.cardinal s) + 4 * (k - AS.cardinal s)%word)
      by (destruct x; simpl; f_equal; ring).
  apply array_addrsI; womega.

  (* Disj s s' *)
  apply ASE.DisjI; intros a Ia Ia'.
  rewrite Cs in Ia; rewrite Cs' in Ia'; clear Cs Cs'.
  setsimpl in Ia; destruct Ia as [k [AE LT]]; subst.
  setsimpl in Ia'; destruct Ia' as [k' [E LT']].
  destruct x; simpl in E; apply addrInjHelper in E.
  replace (w + 4 * AS.cardinal s + 4 * k')%word
    with (w + (4 * (AS.cardinal s + k')))%word in E by ring.
  apply wordAddInj in E; apply wordIndexMultInj in E.
  womega.
  clear E; womega.
  clear E; womega; womega.

Qed.


Lemma split_contiguous : forall x s s',
  contiguous x (s |_| s') ->
  ASE.Disj s s' ->
  contiguous x s ->
  contiguous (x + 4 * AS.cardinal s) s'.
Proof.
  unfold contiguous; intros x s s' Cx Dss' Cx'.
  assert (Bss' := contiguousBounded Cx).
  rewrite ASE.cardinalUnion in Bss'. 2:auto.
  rewrite inj_plus in Bss'.
  replace M.cardinal with AS.cardinal in Bss' by reflexivity.
  assert (Bs := contiguousBounded Cx').
  
  rewrite ASE.cardinalUnion in Cx. 2:auto.
  rewrite natWordAddEq in Cx; rewrite <- array_addrsUnion in Cx.
  replace M.cardinal with AS.cardinal in * by reflexivity.
  rewrite <- Cx' in Cx.
  set (s'' := array_addrs (x + 4 * AS.cardinal s) (AS.cardinal s')) in * .
  rename Cx into E.
  apply ASE.set_ext; intros a; constructor; intros Ia.
  assert (Ia' : AS.In a (s |_| s')) by (setsimpl; auto).
  rewrite E in Ia'; setsimpl in Ia'; destruct Ia'.
  elf; apply (ASE.DisjE Dss' H); auto. auto.
  assert (Ia' : AS.In a (s |_| s'')) by (setsimpl; auto).
  rewrite <- E in Ia'; setsimpl in Ia'; destruct Ia'.
  elf; subst s''; simpl in Ia.
  rewrite Cx' in H; apply array_addrsE in Ia; apply array_addrsE in H.
  clear E.
  destruct Ia as [k [E LT]]; destruct H as [k' [E' LT']].
  subst a.
  rewrite addrAddAssoc in E'; apply addrAddInj in E'.
  replace (4 * AS.cardinal s + 4 * k)%word
    with (4 * (AS.cardinal s + k))%word in E' by ring.
  apply wordIndexMultInj in E'; [ | clear E' | clear E']; womega.
  auto.
  replace M.cardinal with AS.cardinal by reflexivity; womega.
  replace M.cardinal with AS.cardinal by reflexivity; womega.
Qed.


Lemma split_contiguous_end : forall x s s',
  contiguous x (s |_| s') ->
  ASE.Disj s s' ->
  contiguous (x + 4 * AS.cardinal s) s' ->
  contiguous x s.
Proof.
  unfold contiguous; intros x s s' Cx Dss' Cx'.
  assert (Bss' := contiguousBounded Cx).
  rewrite ASE.cardinalUnion in Bss'. 2:auto.
  rewrite inj_plus in Bss'.
  replace M.cardinal with AS.cardinal in Bss' by reflexivity.
  assert (Bs := contiguousBounded Cx').
(*
  assert (Z_of_nat (AS.cardinal s') <= indexBound)%Z as Bs' by uomega.
  assert (okS : okWordZ (Z_of_nat (AS.cardinal s))) by uomega.
  assert (okS' : okWordZ (Z_of_nat (AS.cardinal s'))) by uomega.
*)
  rewrite ASE.cardinalUnion in Cx. 2:auto.
  rewrite natWordAddEq in Cx; rewrite <- array_addrsUnion in Cx.
  replace M.cardinal with AS.cardinal in * by reflexivity.
  rewrite <- Cx' in Cx.
  set (s'' := array_addrs x (AS.cardinal s)) in *.
  rename Cx into E.
  apply ASE.set_ext; intros a; constructor; intros Ia.
  assert (Ia' : AS.In a (s |_| s')) by (setsimpl; auto).
  rewrite E in Ia'; setsimpl in Ia'; destruct Ia'.
  auto.
  elf; apply (ASE.DisjE Dss' Ia); auto.
  assert (Ia' : AS.In a (s'' |_| s')) by (setsimpl; auto).
  rewrite <- E in Ia'; setsimpl in Ia'; destruct Ia'.
  auto.
  elf; subst s''; simpl in Ia.
  rewrite Cx' in H; apply array_addrsE in Ia; apply array_addrsE in H.
  clear E.
  destruct Ia as [k [E LT]]; destruct H as [k' [E' LT']].
  subst a.
  rewrite addrAddAssoc in E'; apply addrAddInj in E'.
  replace (4 * AS.cardinal s + 4 * k')%word 
    with (4 * (AS.cardinal s + k'))%word in E' by ring.
  apply wordIndexMultInj in E'; womega.
  replace M.cardinal with AS.cardinal by reflexivity; womega.
  replace M.cardinal with AS.cardinal by reflexivity; womega.
Qed.
  
  

Ltac setEqSolver := apply ASE.set_ext; 
  let x := fresh in
    let Ix := fresh in
      intros x; setsimpl; constructor; intros Ix; intuition.


Lemma array_addrsRemoveEnd : forall k,
  (Z_of_nat k < indexBound)%Z -> forall x,
  array_addrs x k = AS.remove (x + 4 * k) (array_addrs x (S k)).
Proof.
  intros k LT x.
  apply ASE.set_ext; intros z; setsimpl; constructor; intros Iz.

  constructor.
  intro; subst; setsimpl in Iz; destruct Iz as [a [E Ia]].
  apply addrAddInj in E; apply wordIndexMultInj in E; womega.

  setsimpl in Iz; destruct Iz as [a [E LTa]]; subst.
  apply array_addrsI.
  womega.

  destruct Iz as [NE0 Iz].
  setsimpl in Iz; destruct Iz as [a [E LTa]]; subst.
  apply array_addrsI.
  assert (a <> k) as NE by (intro E; subst; auto).
  clear - NE LTa LT.
  womega.
Qed.


Lemma cardinalEq : ASE.M.cardinal = AS.cardinal.
Proof.
  reflexivity.
Qed.
Hint Rewrite cardinalEq : SetEq.





(* ordered induction on sets *)

Lemma elementsSnocRemove : forall s l a,
  AS.elements s = l ++ a :: nil ->
  AS.elements (AS.remove a s) = l.
Proof.
  intros s l a E; apply ASE.sortEq.
  
  intros b; rewrite <- ASE.InEqv; rewrite <- ASE.InEqv.
  constructor; intros Ib.
  apply AS.elements_2 in Ib; setsimpl in Ib.
  destruct Ib as [NE Ib].
  apply AS.elements_1 in Ib; rewrite E in Ib; clear E.
  clear -Ib NE; induction l.
  simpl in Ib; inversion_clear Ib. elf; auto. inversion H.
  simpl in Ib; inversion_clear Ib; [constructor 1 | constructor 2]; auto.

  apply AS.elements_1; setsimpl.
  destruct (eq_addr_dec a b) as [aE | NE].

    (* case: a = b.  by contradiction *)
    elf; subst b.
    assert (LT : AddrOrderedType.lt a a).
      assert (X := AS.elements_3 s); rewrite E in X.
      clear - Ib X; induction l. inversion Ib.
      simpl in * ; inversion_clear X.
      destruct (eq_addr_dec a0 a) as [E | NE].
      subst a0; eapply ASE.ltSorto; eauto.
      clear; induction l; simpl; auto.
      inversion_clear Ib. elf; auto.
      apply IHl; auto.
    assert (X := AddrOrderedType.lt_not_eq LT); apply X; reflexivity.
    
    (* case: a <> b. *)
    constructor. auto.
    apply AS.elements_2; rewrite E.
    clear - Ib; induction l.
    inversion Ib.
    simpl; inversion_clear Ib; [constructor | constructor 2]; auto.

  apply AS.elements_3.

  assert (X := AS.elements_3 s).
  rewrite E in X; clear - X.
  induction l. constructor.
  simpl in X; inversion_clear X.
  constructor; auto; clear IHl.
  clear - H0; destruct l. constructor.
  simpl in H0; inversion_clear H0; constructor; auto.

Qed.


Lemma elementsConsRemove : forall s l a,
  AS.elements s = a :: l ->
  AS.elements (AS.remove a s) = l.
Proof.
  intros s l a E.
  apply ASE.sortEq.

  intros b; rewrite <- ASE.InEqv; rewrite <- ASE.InEqv.

  constructor; intros Ib.
  apply AS.elements_2 in Ib; setsimpl in Ib.
  destruct Ib as [NE Ib].
  apply AS.elements_1 in Ib; rewrite E in Ib; clear E.
  inversion_clear Ib; [elf | ]; auto.

  apply AS.elements_1; setsimpl.
  destruct (eq_addr_dec a b) as [aE | NE].

    (* case: a = b.  by contradiction *)
    elf; subst b.
    assert (LT : AddrOrderedType.lt a a).
      assert (X := AS.elements_3 s); rewrite E in X.
      inversion_clear X.
      eapply ASE.ltSorto; eauto.
      rewrite <- ASE.InEqv; auto.
    clear - LT; destruct a; simpl in LT; destruct LT as [LT | LT].
    assert (X := BlockOrdered.lt_not_eq _ _ LT); apply X; reflexivity.
    destruct LT as [_ LT]; apply wordLtInj in LT; omega.
    
    (* case: a <> b. *)
    constructor. auto.
    apply AS.elements_2; rewrite E; constructor 2; auto.

  apply AS.elements_3.

  assert (X := AS.elements_3 s).
  rewrite E in X; inversion_clear X; auto.

Qed.


(*
These two lemmas are fairly useless, because they use the weird 
address <, not the "real" address <.

Lemma set_induction_inc : forall P,
   P AS.empty ->
  (forall s, P s -> forall x,
    (forall y, AS.In y s -> x < y) -> P (AS.add x s)) ->
  forall s, P s.
Proof.
  intros P Pempty Padd.
  assert (forall l s, l = AS.elements s -> P s). 2:eauto.
  induction l; intros s E.

  replace s with AS.empty; [auto | clear - E].
  apply ASE.set_ext; intros x; setsimpl; constructor; intros Ix.

  contradiction.
  assert (X := AS.elements_1 Ix).
  rewrite <- E in X; inversion X.

  clear Pempty.
  assert (Ia : AS.In a s).
    apply AS.elements_2; rewrite <- E; constructor; auto.
  hyr Padd (fun X => Padd (AS.remove a s) X a); setsimpl in Padd.
  apply Padd; clear Padd.
  apply IHl; symmetry; apply elementsConsRemove; auto.
  
  clear IHl.
  intros y Iy; setsimpl in Iy.
  destruct Iy as [NE Iy].
  assert (X := AS.elements_3 s).
  rewrite <- E in X; inversion_clear X.
  eapply ASE.ltSorto; eauto.
  rewrite <- ASE.InEqv.
  replace l with (AS.elements (AS.remove a s)).
  apply AS.elements_1.
  setsimpl; auto.
  apply elementsConsRemove; auto.
Qed.


Lemma set_induction_dec : forall P,
   P AS.empty ->
  (forall s, P s -> forall x, 
    (forall y, AS.In y s -> y < x) -> P (AS.add x s)) ->
  forall s, P s.
Proof.
  intros P Pempty Padd.

  cut (forall l s, rev l = AS.elements s -> P s).
  clear; intros X s; apply (X (rev (AS.elements s))).
  rewrite rev_involutive; auto.

  induction l; intros s E.

  replace s with AS.empty; [auto | clear - E].
  apply ASE.set_ext; intros x; setsimpl; constructor; intros Ix.

  contradiction.
  assert (X := AS.elements_1 Ix).
  rewrite <- E in X; inversion X.

  clear Pempty.
  simpl in E.
  assert (Ia : AS.In a s).
    apply AS.elements_2; rewrite <- E; rewrite ASE.InEqv; clear.
    induction (rev l); simpl; auto.
  hyr Padd (fun X => Padd (AS.remove a s) X a); setsimpl in Padd.
  apply Padd; clear Padd.
  apply IHl; symmetry; apply elementsSnocRemove; auto.
  
  clear IHl Ia P.
  assert (Iok : forall z, AS.In z s -> z = a \/ z < a).
    intros z Iz.
    apply AS.elements_1 in Iz; rewrite <- E in Iz.
    assert (X := AS.elements_3 s).
    rewrite <- E in X; clear E.
    induction (rev l).
    simpl in Iz; inversion_clear Iz; [auto | inversion H].
    simpl in Iz, X; inversion_clear X; inversion_clear Iz.
    assert (z = a0) by auto; subst a0; clear H1.
    right; eapply ASE.ltSorto; eauto.
      clear; induction l0; simpl; auto.
    destruct IHl0; auto.

  intros y Iy; setsimpl in Iy; destruct Iy as [NE Iy]; 
    destruct (Iok y Iy); [elf |]; auto.

Qed.
*)


Lemma empty_contiguous : forall x,
  contiguous x AS.empty.
Proof.
  intros x; unfold contiguous; setsimpl; reflexivity.
Qed.




