(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import OrderedTypeEx.
Require Export FSetListEq.
Require Import NatUtil.


Set Implicit Arguments.


Module Import NSE := FSetListEq.Make (Nat_as_OT).
Module Import NS := NSE.M.

Infix "|_|" := NS.union (at level 60, right associativity).

Hint Rewrite add_eqv empty_eqv remove_eqv union_eqv diff_eqv inter_eqv
  empty_diff_1' empty_diff_2' remove_diff_singleton' empty_inter2'
  singleton_choose
  multi_mapEqv empty_union_eqv1 empty_union_eqv2 : SetEq.
Hint Rewrite cardinal_emptyEq inSingEq inter_empty1 inter_empty2 : SetEq.
Hint Rewrite removeAddEqv addRemoveEqv 
  singleton_inter_1 singleton_inter_2 using assumption : SetEq.
Hint Rewrite singletonAdd : SetEq.
Hint Resolve DisjE_unionRR DisjE_unionRL : ListSet.


(* {k | 0 <= k < x } *)
Fixpoint zeroRangeSet (x:nat) : t :=
  match x with
    | 0 => empty
    | S x' => add x' (zeroRangeSet x')
  end.


Lemma zeroRangeSetI : forall x k,
  k < x ->
  In k (zeroRangeSet x).
Proof.
  induction x; simpl; intros k LT.
  elf; omega.
  compare k x; intro NE.
  apply add_1; auto.
  apply add_2; apply IHx; omega.
Qed.


Lemma zeroRangeSetE : forall x k,
  In k (zeroRangeSet x) ->
  k < x.
Proof.
  induction x; simpl; intros k I; setsimpl in I.
  elf; auto.
  destruct I.
  omega.
  assert (X := IHx _ H); omega.
Qed.
Implicit Arguments zeroRangeSetE [x k].


Lemma zeroRangeSetEqv : forall x y,
  In x (zeroRangeSet y) <-> x < y.
Proof.
  intros x y; constructor; intros Ix.
  apply zeroRangeSetE; auto.
  apply zeroRangeSetI; auto.
Qed.


Hint Rewrite zeroRangeSetEqv : SetEq.


(*
(* {k | x <= k < y } *)
Definition rangeSet (x y:nat) : set :=
  set_diff (zeroRangeSet y) (zeroRangeSet x).

Lemma rangeSetI : forall k x y,
  x <= k < y ->
  set_In k (rangeSet x y).
Proof.
  unfold rangeSet; intros k x y [LE LT].
  apply set_diffI.
  apply zeroRangeSetI; auto.
  intro I; hyr I (zeroRangeSetE I); omega.
Qed.


Lemma rangeSetE : forall k x y,
  set_In k (rangeSet x y) ->
  x <= k < y.
Proof.
  unfold rangeSet; intros k x y I.
  destruct (set_diffE I) as [I' NI]; clear I.
  hyr I' (zeroRangeSetE I').
  assert (~k < x).
  intro LT; apply NI; apply zeroRangeSetI; auto.
  omega.
Qed.
Implicit Arguments rangeSetE [k x y].


Lemma zeroRangeSetEmpty :
  zeroRangeSet 0 = NS.empty.
Proof.
  apply set_ext; intro x; constructor; intro I; elf.
  hyr I (zeroRangeSetE I); destruct x; omega.
  apply (empty_setE I).
Qed.




Lemma finite_subsetE : forall i s,
  set_subset s (zeroRangeSet i) ->
  s = NS.empty \/ exists x, set_In x s.
Proof.
  induction i; intros s Sub.

  rewrite zeroRangeSetEmpty in Sub; left; apply empty_setSubsetEq; auto.

  simpl in Sub.
  destruct (set_In_dec i s) as [I | NI].
  right; econstructor; eauto.
  apply IHi; intros x Ix; hyr Sub (Sub _ Ix).
  destruct (set_addE Sub); auto; subst; elf; auto.
  
Qed.


Lemma zeroRangeSetEqv : forall x k,
  set_In k (zeroRangeSet x) <-> k < x.
Proof.
  intros x k; constructor; intro I.
  apply (zeroRangeSetE I).
  apply zeroRangeSetI; auto.
Qed.


Lemma rangeSetEqv : forall k x y,
  set_In k (rangeSet x y) <-> x <= k < y.
Proof.
  intros k x y; constructor; intro I.
  apply (rangeSetE I).
  apply rangeSetI; auto.
Qed.


Hint Rewrite zeroRangeSetEqv rangeSetEqv zeroRangeSetEmpty : SetEq.
*)


(* {4k + z | 0 <= 4k + z < 4x + z } *)
Fixpoint zeroAlignSet' (numBytes:nat) : NS.t :=
  match numBytes with
    | (0 | 1 | 2 | 3) => NS.empty
    | S (S (S (S numBytes'))) =>
      NS.add numBytes' (zeroAlignSet' numBytes')
  end.


(* {4k | 0 <= 4k + 4 <= x } *)
Definition zeroAlignSet (numBytes:nat) : NS.t :=
  zeroAlignSet' (natWordAlign numBytes).


Lemma zeroAlignSetI' : forall k numWords,
  k < numWords -> forall z,
  NS.In (4 * k + z) (zeroAlignSet' (4 * numWords + z)).
Proof.
  induction numWords; intros LT z.
  elf; omega.
  replace (4 * S numWords + z) with (S (S (S (S (4 * numWords + z))))) by omega.
  assert (D : k = numWords \/ k < numWords).
  omega.
  clear LT; destruct D as [E | LT].
  subst k; simpl; apply NS.add_1; reflexivity.
  simpl; apply NS.add_2; apply (IHnumWords LT).
Qed.


Lemma zeroAlignSetE' : forall x y,
  NS.In x (zeroAlignSet' y) ->
  exists k, exists z, exists numWords,
    x = 4 * k + z /\ y = 4 * numWords + z /\ z < 4 /\ k < numWords.
Proof.
  induction y using natInd'; intros Ix.

  setsimpl in Ix; contradiction.

  destruct y as [|[|[|[| z]]]]; simpl in Ix; setsimpl in Ix; try (elf; auto; fail).
  clear H0; destruct (eq_nat_dec x z) as [E | NE].

  subst z; clear H Ix.
  destruct (natWordAlignEq' x) as [z [E LT]].
  destruct (natWordAlignEq x) as [k E']; rewrite E' in E; subst.
  exists k; exists z; exists (k + 1).
  constructor; auto; constructor; omega.

  destruct Ix as [E | Ix].
  elf; auto.
  destruct (H z) as [k [z' [x' [E [E' [Lt Lt']]]]]]; clear H; try omega; auto.
  subst; exists k; exists z'; exists (x' + 1); intuition.

Qed.


Lemma zeroAlignSetI : forall k x,
  4 * k < natWordAlign x ->
  NS.In (4 * k) (zeroAlignSet x).
Proof.
  unfold zeroAlignSet; intros k x LE.
  destruct (natWordAlignEq x) as [k' E]; rewrite E.
  replace (4 * k) with (4 * k + 0) by auto.
  replace (4 * k') with (4 * k' + 0) by auto.
  apply zeroAlignSetI'.
  assert (4 * k < 4 * k').
  rewrite <- E; auto.
  omega.
Qed.


Lemma zeroAlignSetE : forall x y,
  NS.In x (zeroAlignSet y) ->
  exists k, x = 4 * k /\ x < natWordAlign y.
Proof.
  unfold zeroAlignSet; intros x k Ix.
  hyr Ix (zeroAlignSetE' _ Ix).
  destruct Ix as [k' [z [nw [xE [yE [zLt k'Lt]]]]]].
  exists k'.
  cut (z = 0).
  intro; subst; intuition.
  clear x xE k'Lt.

  assert (X : forall x y k, k < 3 -> 4 * x = 4 * y + (S k) -> False).
  clear k k' z nw zLt yE.
  destruct x; destruct y; intros k LT E; try discriminate;
    destruct k as [| [| [| k]]]; try discriminate; omega.

  destruct z as [| [| [| [| z]]]]; auto; try (elf; omega); clear zLt k';
    destruct (natWordAlignEq k) as [k' E]; rewrite E in yE; clear k E; elf;
      eauto.

Qed.


Lemma zeroAlignSetEqv : forall x y,
  NS.In x (zeroAlignSet y) <-> exists k, x = 4 * k /\ x < natWordAlign y.
Proof.
  intros x y; constructor; intro H.
  apply (zeroAlignSetE _ H).
  destruct H as [k [E LT]]; subst x; apply zeroAlignSetI; auto.
Qed.

Hint Rewrite zeroAlignSetEqv : SetEq.


Lemma zeroRangeSetCardinal : forall k,
  NS.cardinal (zeroRangeSet k) = k.
Proof.
  intros k.
  induction k.
  auto.
  unfold zeroRangeSet; fold zeroRangeSet.
  rewrite Miff.add_cardinal_2.
  f_equal; assumption.
  clear IHk; intro Ik; setsimpl in Ik.
  omega.
Qed.


Lemma zeroAlignSetEq : forall x,
  zeroAlignSet (4 + x) = NS.add (natWordAlign x) (zeroAlignSet x).
Proof.
  intros x; apply NSE.set_ext; intros a; setsimpl; constructor; intros Ia; auto.
Qed.


Lemma zeroAlignSetCard4 : forall x,
  NS.cardinal (zeroAlignSet (4 + x)) = S (NS.cardinal (zeroAlignSet x)).
Proof.
  intros x.
  rewrite zeroAlignSetEq.
  replace (NS.add (natWordAlign x) (zeroAlignSet x))
    with (NS.union (NS.singleton (natWordAlign x)) (zeroAlignSet x))
      by (setsimpl; auto).
  rewrite NSE.cardinalUnion.
  auto.

  apply NSE.DisjI; intros a; setsimpl; intros Ia Ia'; subst a.
  destruct Ia' as [k [_ LT]]; omega.

Qed.



Require Import ZArith.


Lemma zeroAlignCardEq : forall x,
  Z_of_nat (NS.cardinal (zeroAlignSet x)) = (Z_of_nat x / 4)%Z.
Proof.
  induction x using natInd4; try reflexivity.

  replace (Z_of_nat (4 + x)) with (Z_of_nat x + 4)%Z.
  replace ((Z_of_nat x + 4) / 4)%Z with (Z_of_nat x / 4 + 1)%Z.
  rewrite <- IHx.
  rewrite zeroAlignSetCard4; rewrite inj_S; omega.

  clear; generalize (Z_of_nat x); clear; intros x.
  replace (x + 4)%Z with (x + 1 * 4)%Z by omega.
  rewrite Z_div_plus. auto. omega.

  simpl (4 + x); rewrite inj_S; rewrite inj_S; rewrite inj_S; rewrite inj_S.
  omega.

Qed.