(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import OrderedTypeEx.
Require Export FSetListEq.
Require Import ListNatSet.
Require Export Word.


Set Implicit Arguments.


Module WordOrderedType <: UsualOrderedType.

 Definition t := word.
 Definition eq := @Logic.eq t.
 Definition lt x y := (x < y)%word.
 Definition eq_refl := @refl_equal t.
 Definition eq_sym := @sym_eq t.
 Definition eq_trans := @trans_eq t.

 Lemma lt_trans : forall x y z : t, lt x y -> lt y z -> lt x z.
 Proof.
   unfold t, lt; intros; wunfold; omega.
 Qed.

 Lemma lt_not_eq : forall x y : t, lt x y -> ~ eq x y.
 Proof.
   unfold t, lt, eq; intros; wunfold; apply wordNeInj'; omega.
 Qed.

 Definition compare : forall x y : t, Compare lt eq x y.
 Proof.
   unfold lt, eq; intros x y.
   destruct (eq_word_dec x y).
   subst; constructor 2; reflexivity.
   destruct (lt_word_dec x y).
   constructor 1; auto.
   constructor 3; apply wordNeInj in n; wunfold; omega.
 Qed.

 Definition eq_dec := eq_word_dec.
 
End WordOrderedType.


Module Import WSE := FSetListEq.Make (WordOrderedType).
Module WS := WSE.M.

Infix "|_|" := WS.union (at level 60, right associativity).


Hint Rewrite add_eqv empty_eqv remove_eqv union_eqv diff_eqv inter_eqv 
  empty_diff_1' empty_diff_2' remove_diff_singleton' empty_inter2'
  singleton_choose
  empty_union_eqv1 empty_union_eqv2 multi_mapEqv : SetEq.
Hint Rewrite cardinal_emptyEq inSingEq inter_empty1 inter_empty2 : SetEq.
Hint Rewrite removeAddEqv addRemoveEqv 
  singleton_inter_1 singleton_inter_2 using assumption : SetEq.
Hint Rewrite singletonAdd : SetEq.
Hint Resolve DisjE_unionRR DisjE_unionRL : ListSet.


Module NWSetMap := FSetMap NSE WSE.
Hint Rewrite NWSetMap.map_empty NWSetMap.singleton_map : SetEq.


Definition natToWordSet (s:NS.t) : WS.t := NWSetMap.map nat_to_word s.


Hint Unfold NS.E.eq WS.E.eq Nat_as_OT.eq WordOrderedType.eq.


Lemma natToWordSetI : forall x s,
  NS.In x s ->
  WS.In (nat_to_word x) (natToWordSet s).
Proof.
  unfold natToWordSet; intros x s I.
  apply NWSetMap.mapI; auto; intros.
Qed.


Lemma natToWordSetE : forall (x:word) s,
  WS.In x (natToWordSet s) ->
  exists y, nat_to_word y = x /\ NS.In y s.
Proof.
  unfold natToWordSet; intros x s I.
  rewrite <- (word_to_nat_to_word x) in I.
  apply NWSetMap.mapE0 in I. destruct I as [y [Iy E]].
  rewrite word_to_nat_to_word in E; unfold WS.E.eq, WordOrderedType.eq in E; subst.
  exists y; constructor; auto.
Qed.



Definition wordZeroRangeSet (x:word) :=
  natToWordSet (ListNatSet.zeroRangeSet (word_to_nat x)).


Lemma wordZeroRangeSetI : forall x y,
  (x < y)%word ->
  WS.In x (wordZeroRangeSet y).
Proof.
  unfold wordZeroRangeSet; intros x y LT.
  rewrite <- (word_to_nat_to_word x); apply natToWordSetI.
  setsimpl; apply word_to_natLtI; auto.
Qed.


Lemma wordZeroRangeSetE : forall x y,
  WS.In x (wordZeroRangeSet y) ->
  (x < y)%word.
Proof.
  unfold wordZeroRangeSet; intros x y Ix; apply natToWordSetE in Ix.
  destruct Ix as [z [E Iz]]; subst; setsimpl in Iz.
  apply word_to_natLtE; rewrite nat_to_word_to_nat; auto.
  apply okWordNatLtTrans with (word_to_nat y).
  omega.
  apply okWordNatWord.
Qed.


Lemma wordZeroRangeSetEqv : forall x y,
  WS.In x (wordZeroRangeSet y) <-> (x < y)%word.
Proof.
  intros x y; constructor; intro Ix.
  apply wordZeroRangeSetE; auto.
  apply wordZeroRangeSetI; auto.
Qed.


Hint Rewrite wordZeroRangeSetEqv : SetEq.


Definition zeroWordSet (x:word) :=
  natToWordSet (zeroAlignSet (word_to_nat x)).


Lemma zeroWordSetI : forall x,
  x = wordAlign x -> forall y,
    (x < wordAlign y)%word ->
    WS.In x (zeroWordSet y).
Proof.
  unfold zeroWordSet; intros x E y LT.
  rewrite <- (word_to_nat_to_word x).
  apply natToWordSetI.
  rewrite E; rewrite <- word_to_natAlignEq.
  destruct (natWordAlignEq (word_to_nat x)) as [k E'].
  rewrite E'; apply zeroAlignSetI.
  rewrite <- E'; rewrite word_to_natAlignEq; rewrite word_to_natAlignEq.
  apply word_to_natLtI; rewrite <- E; auto.
Qed.


Lemma zeroWordSetE : forall x y,
  WS.In x (zeroWordSet y) ->
  x = wordAlign x /\ (x < wordAlign y)%word.
Proof.
  unfold zeroWordSet; intros x y Ix.
  destruct (natToWordSetE _ Ix) as [z [zE Iz]]; subst; clear Ix.
  destruct (zeroAlignSetE _ Iz) as [k [zE zLt]]; clear Iz; subst.
  constructor.
  apply word_to_natEqE.
  rewrite <- word_to_natAlignEq.
  rewrite nat_to_word_to_natAlign.
  rewrite natWordAlignAlignEq; auto.

  apply word_to_natLtE.
  rewrite <- word_to_natAlignEq.
  assert (X := @nat_to_word_to_natLE (4 * k)).
  omega.

Qed.


Lemma zeroWordSetEqv : forall x y,
  WS.In x (zeroWordSet y) <-> x = wordAlign x /\ (x < wordAlign y)%word.
Proof.
  intros x y; constructor; intros Ix.
  apply zeroWordSetE; auto.
  destruct Ix; apply zeroWordSetI; auto.
Qed.


Hint Rewrite zeroWordSetEqv : SetEq.


Lemma zeroWordSetEmptyEq : forall w,
  wordAlign w = 0%word ->
  zeroWordSet w = WS.empty.
Proof.
  intros w E.
  apply set_ext; intros x; constructor; intros Ix; setsimpl in Ix; elf.
  destruct Ix as [_ LT]; destruct x as [x xOk]; hyr LT (wordLtInj LT).
  unfold wordAlign in E; injector E; wunfold; omega.
  elf; auto.
Qed.


Lemma zeroWordSetEmptyEq0 :
  zeroWordSet 0%word = WS.empty.
Proof.
  rewrite zeroWordSetEmptyEq; [ | apply wordInj]; auto.
Qed.


Hint Rewrite zeroWordSetEmptyEq0 : SetEq.


Lemma finite_wordSubsetE : forall s,
  s = WS.empty \/ exists x, WS.In x s.
Proof.
  intros s.
  assert (X := WS.choose_1 (s := s)).
  assert (Y := WS.choose_2 (s := s)).
  destruct (WS.choose s); [right | left].
  exists e; auto.
  apply set_ext.
  apply Miff.empty_is_empty_1; apply Y; auto.
Qed.


Lemma zeroWordSetSplitEnd : forall w,
  (4 <= w + 4)%word ->
  zeroWordSet (w + 4)%word = WS.add (wordAlign w) (zeroWordSet w).
Proof.
  intros w wOk; apply set_ext; intros x; constructor; intro Ix;
    setsimpl; setsimpl in Ix.

  destruct Ix as [xE LT].
  destruct (eq_word_dec x (wordAlign w)) as [E | NE].
  subst; auto.
  right; constructor; auto.
  rewrite xE in LT; rewrite xE; apply wordAlignPred; auto; rewrite <- xE; auto.

  destruct Ix as [E | [E LT]].

  subst.
  constructor.
  rewrite wordAlignAlignEq; auto.
  apply anotherWordAlignLemma; auto; wunfold; omega.

  constructor; auto.
  assert (wordAlign w < wordAlign (w + 4))%word.
  apply wordAlignLtInj; apply noOverflowLeToLt; auto.
  destruct x; destruct w; apply wordLtInj'; hyr H (wordLtInj H);
    hyr LT (wordLtInj LT); wunfold; omega.

Qed.


Require Import ZArith.


Lemma natToWordSetCardinal : forall s,
  (forall x, NS.In x s -> okWordNat x) ->
  WS.cardinal (natToWordSet s) = NS.cardinal s.
Proof.
  unfold natToWordSet; intros s sOk.
  rewrite NWSetMap.map_cardinal; auto.
  intros x y Ix Iy E.
  apply wordInj' in E; simpl in E.
  rewrite wordModEq in E.
  2:apply (sOk x); auto.
  rewrite wordModEq in E.
  2:apply (sOk y); auto.
  assert (x = y). 2:auto.
  unfold NWSetMap.S1.elt, NWSetMap.S1.E.t, NSE.M.E.t, Nat_as_OT.t in * ; omega.
Qed.


Lemma wordZeroRangeSetCardinal : forall k,
  WS.cardinal (wordZeroRangeSet k) = word_to_nat k.
Proof.
  unfold wordZeroRangeSet; intros k.
  rewrite natToWordSetCardinal.
  apply zeroRangeSetCardinal.
  intros x Ix.
  setsimpl in Ix.
  constructor.
  omega.
  destruct k as [k kOk].
  unfold okWordZ in kOk.
  replace (word_to_nat (wordi kOk)) with (Zabs_nat k) in Ix.
  assert (Z_of_nat x < k)%Z.
  replace k with (Z_of_nat (Zabs_nat k)).
  omega.
  rewrite Z_to_nat_to_Z; auto.
  omega.
  omega.
  reflexivity.
Qed.


Lemma wordZeroRangeSetZero : wordZeroRangeSet 0 = WS.empty.
Proof.
  reflexivity.
Qed.
Hint Rewrite wordZeroRangeSetZero : SetEq.


Lemma zeroWordSetBounded : forall k,
  Z_of_nat (WS.cardinal (zeroWordSet k)) = (word_to_Z k / 4)%Z.
Proof.
  unfold zeroWordSet; intros [k kOk]; unfold okWordZ in kOk; simpl.
  rewrite natToWordSetCardinal.

  rewrite zeroAlignCardEq.
  rewrite word_to_nat_to_Z; auto.

  intros x Ix.
  setsimpl in Ix; destruct Ix as [q [E LT]]; subst.
  unfold word_to_nat in LT; simpl (wval (wordi kOk)) in LT.
  unfold okWordNat.
  constructor. omega.
  replace wordBound with (Z_of_nat (Zabs_nat wordBound))
    by (rewrite Z_to_nat_to_Z; auto; omega).
  apply inj_lt.
  assert (natWordAlign (Zabs_nat k) < Zabs_nat wordBound). 2:omega.
  clear LT q.
  assert (Zabs_nat k < Zabs_nat wordBound). 
  2:destruct (natWordAlignEq' (Zabs_nat k)); omega.
  apply Zabs_nat_lt; auto.

Qed.
  
