(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export OptUtil.
Require Export Memory.
Require Export VFile.

Set Implicit Arguments.


(* This is based on the 07 Cminor paper by Appel and Blazy.  The main
   difference is that our heap is simpler, because we don't have
   floats and don't distinguish read and write permissions. *)


(* defining the machine state *)


Inductive state : Type :=
  statei : mem -> vfile -> block -> state.

Definition memOf (s:state) : mem :=
  let (h, _, _) := s in h.

Definition vfOf (s:state) : vfile :=
  let (_, vf, _) := s in vf.

Definition spOf (s:state) : block :=
  let (_, _, sp) := s in sp.

Definition set_mem (s:state) (h:mem) : state :=
  statei h (vfOf s) (spOf s).

Definition set_vf (s:state) (vf:vfile) : state :=
  statei (memOf s) vf (spOf s).

Definition set_sp (s:state) (sp':block) : state :=
  statei (memOf s) (vfOf s) sp'.


(* defining the instructions *)

Inductive uop : Set :=
  | ldop
  | notop.

Inductive exp : Set :=
  | eval : val -> exp
  | evar : var -> exp
  | esp : exp
  | euop : uop -> exp -> exp
  | ebop : bop -> exp -> exp -> exp
  | eblop : blop -> exp -> exp -> exp.


Inductive stm : Set :=
  | vassign : var -> exp -> stm
  | store : exp -> exp -> stm
  | loop : stm -> stm
  | block_st : stm -> stm
  | exit : nat -> stm
  | call : var -> exp -> list exp -> stm
  | ret : exp -> stm
  | stcomp : stm -> stm -> stm
  | ifte : exp -> stm -> stm -> stm
  | skip : stm.


Notation while := (fun e s => block_st (loop (ifte e s (exit 0)))).
Notation ift := (fun e s => ifte e s skip).

Inductive fdef : Set :=
  | fdefi : list var -> list var -> word -> stm -> fdef.
   (* arg names, local vars, stack space, body *)

Definition frameLen (f:fdef) : word :=
  let (_, _, l, _) := f in l.

Inductive kcontrol : Type :=
  | Kstop : kcontrol
  | Kcomp : stm -> kcontrol -> kcontrol
  | Kblock : kcontrol -> kcontrol
  | Kcall : var -> fdef -> block -> vfile -> kcontrol -> kcontrol.
    (* return variable, function, stack pointer, vars *)

Definition cont := prod state kcontrol.

Definition cdhp := prod block word -> option fdef.


(* operational semantics *)

Definition loadVal (s:state) (v:val) : option val :=
  match v with
    | Vptr p => M.load (memOf s) p
    | _ => None
  end.

Definition uopVal (s:state) (u:uop) (v:val) : option val :=
  match u with
    | ldop => loadVal s v
    | notop => Some (valNot v)
  end.

Fixpoint expVal (s:state) (e:exp) {struct e} : option val :=
    match e with
      | eval v => Some v
      | evar x => VF.get (vfOf s) x
      | esp => Some (Vptr (spOf s, 0%word))
      | eblop b e1 e2 =>
        do v1 <-- expVal s e1;;;
        match valBoolCases v1 with
          | inleft (left _) =>
            match b with
              | andop => do v2 <-- expVal s e2;;; Some (valToBool v2)
              | orop => Some (true:val)
            end
          | inleft (right _) =>
            match b with
              | andop => Some (false:val)
              | orop => do v2 <-- expVal s e2;;; Some (valToBool v2)
            end
          | inright _ => Some Vundef
        end
      | ebop b e1 e2 =>
	do v1 <-- expVal s e1;;;
	do v2 <-- expVal s e2;;;
        Some (bopVal b v1 v2)
      | euop u e =>
	do v <-- expVal s e;;;
        uopVal s u v
    end.

Inductive popToBlock : kcontrol -> kcontrol -> Prop :=
  | blockPop : forall k,
    popToBlock (Kblock k) k
  | compPop : forall s k k',
    popToBlock k k' ->
    popToBlock (Kcomp s k) k'.

Inductive popFunc : kcontrol -> kcontrol -> Prop :=
  | callPopF : forall xl f sp p k,
    popFunc (Kcall xl f sp p k) (Kcall xl f sp p k)
  | compPopF : forall s k k',
    popFunc k k' ->
    popFunc (Kcomp s k) k'
  | blockPopF : forall k k',
    popFunc k k' ->
    popFunc (Kblock k) k'.

(* i'd define this as a fixed point, but that causes coq to loop *)
Inductive expVals (st:state) : list exp -> list val -> Prop :=
  | nil_expVals : expVals st nil nil
  | cons_expVals : forall e v exps vals,
    expVals st exps vals ->
    expVal st e = Some v ->
    expVals st (e::exps) (v::vals).


Definition map_vf (f:vfile -> option vfile) (s:state) : option state :=
  do vf' <-- f (vfOf s);;;
  Some (set_vf s vf').

Definition map_mem (f:mem -> option mem) (s:state) : option state :=
  do mem' <-- f (memOf s);;;
  Some (set_mem s mem').


Inductive step (C:cdhp) : cont -> cont -> Prop := 
  | vassignStep : forall st vf' x e v k,
    expVal st e = Some v ->
    VF.update (vfOf st) x v = Some vf' ->
    step C (st, Kcomp (vassign x e) k) (set_vf st vf', k)
  | storeStep : forall st st' e1 e2 p v2 k,
    expVal st e1 = Some (Vptr p) ->
    expVal st e2 = Some v2 ->
    map_mem (fun h => M.store h p v2) st = Some st' ->
    step C (st, Kcomp (store e1 e2) k) (st', k)
  | loopStep : forall st s k,
    step C (st, Kcomp (loop s) k) (st, Kcomp s (Kcomp (loop s) k))
  | blockStep : forall st s k,
    step C (st, Kcomp (block_st s) k) (st, Kcomp s (Kblock k))
  | exit0Step : forall st k k',
    popToBlock k k' ->
    step C (st, Kcomp (exit 0) k) (st, k')
  | exitSStep : forall st n k k',
    popToBlock k k' ->
    step C (st, Kcomp (exit (S n)) k) (st, Kcomp (exit n) k')
  | callStep : forall (st:state) retx fexp args k f f' argx locx 
      stackSize fbody argVals mem' sp' vf',
    expVal st fexp = Some (Vptr (f, f')) ->
    let fdef := fdefi argx locx stackSize fbody in
    C (f, f') = Some fdef ->
    expVals st args argVals ->
    alloc (memOf st) stackSize = (sp', mem') ->
    vfUpdateVals (base (argx ++ locx)) argx argVals = 
       Some vf' ->
    step C (st, Kcomp (call retx fexp args) k)
      (statei mem' vf' sp',
	Kcomp fbody (Kcall retx fdef (spOf st) (vfOf st) k))
  | retStep : forall st e v k k' x f sp' vf' vf'' m',
    popFunc k (Kcall x f sp' vf' k') ->
    expVal st e = Some v ->
    VF.update vf' x v = Some vf'' ->
    M.free (memOf st) (spOf st) = Some m' ->
    step C (st, Kcomp (ret e) k) (statei m' vf'' sp', k')
  | compStep : forall st s1 s2 k,
    step C (st, Kcomp (stcomp s1 s2) k) (st, Kcomp s1 (Kcomp s2 k))
  | ifteTrueStep : forall st e s1 s2 v k,
    expVal st e = Some v ->
    trueVal v ->
    step C (st, Kcomp (ifte e s1 s2) k) (st, Kcomp s1 k)
  | ifteFalseStep : forall st e s1 s2 k,
    expVal st e = Some (Vword 0) ->
    step C (st, Kcomp (ifte e s1 s2) k) (st, Kcomp s2 k)
  | skipStep : forall st k,
    step C (st, Kcomp skip k) (st, k).



(* statement syntax *)

Delimit Scope CM_scope with CM.

Notation "i <- e" := (vassign i e) (at level 90) : CM_scope.
Notation "[ e1 ] <- e2" := (store e1 e2) (at level 90) : CM_scope.
Notation "I ;; J" := (stcomp I J) (at level 91, right associativity) : CM_scope.
Notation "i <- e @( x , .. , y )" := 
  (call i e (@cons exp x .. (@cons exp y nil) ..)) 
  (at level 90, e at next level) : CM_scope.


(* expression syntax *)
Coercion eval : val >-> exp.
Coercion evar : var >-> exp.
Notation "[ e ]" := (euop ldop e) (at level 45, right associativity) : CM_scope.
Notation "!" := (euop notop) (at level 45) : CM_scope.
Notation "e + f" := (ebop plusop e f) : CM_scope.
Notation "e - f" := (ebop subop e f) : CM_scope.
Notation "e * f" := (ebop multop e f) : CM_scope.
Notation "e < f" := (ebop ltop e f) : CM_scope.
Notation "e > f" := (ebop gtop e f) : CM_scope.
Notation "e <= f" := (ebop lteop e f) : CM_scope.
Notation "e >= f" := (ebop gteop e f) : CM_scope.
Notation "e == f" := (ebop eqop e f) : CM_scope.
Notation "e != f" := (ebop neqop e f) : CM_scope.
Notation "e &&&& f" := (eblop andop e f) (at level 74) : CM_scope.
Notation "e |||| f" := (eblop orop e f) (at level 74) : CM_scope.




