(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export CoreMemorySig.

Set Implicit Arguments.


Module Type MemorySig.
  
  Declare Module CM : CoreMemorySig.

  Parameter mem : Type.
  Notation block := CM.block.
  Notation blockOf := CM.blockOf.
  Notation offsetOf := CM.offsetOf.


  (* predicates *)
  
  Parameter validBlock : mem -> block -> Prop.
  Parameter blockOffs : mem -> block -> WS.t.
  Definition validPtr (m:mem) (p:CM.addr) : Prop :=
    validBlock m (blockOf p) /\ WS.In (offsetOf p) (blockOffs m (blockOf p)).


  (* constructor *)
  Parameter empty : mem.

  (* operations *)
  Parameter alloc : mem -> word -> prod block mem.
  Parameter free : mem -> block -> option mem.
  Parameter load : mem -> CM.addr -> option CM.val.
  Parameter store : mem -> CM.addr -> CM.val -> option mem.
  Parameter core : mem -> CM.mem.

  (* decidability of predicates *)
  Parameter eq_block_dec : forall (b b':block), {b = b'} + {b <> b'}.

  Parameter validBlock_dec : forall m b,
    {validBlock m b} + {~validBlock m b}.

  Parameter validPtr_dec : forall m p,
    {validPtr m p} + {~validPtr m p}.


  (* basic operation properties *)
  
  Parameter allocLoadNeq : forall m m' l b,
    alloc m l = (b, m') -> forall p,
      blockOf p <> b ->
	load m' p = load m p.
  
  Parameter freeLoadNeq : forall m m' b,
    free m b = Some m' -> forall p,
      blockOf p <> b ->
	load m' p = load m p.
  
  Parameter storeLoadEq : forall m p v m',
    store m p v = Some m' ->
    load m' p = Some v.
  
  Parameter storeLoadNeq : forall m m' p v,
    store m p v = Some m' -> forall p',
      p <> p' ->
      load m' p' = load m p'.


  (* validity properties *)
  
  Parameter emptyValidE : forall b, 
    validBlock empty b -> 
    False.

  Parameter coreValidE : forall m b,
    CM.validBlock (core m) b ->
    validBlock m b.

  Parameter allocFresh : forall m l b m',
    alloc m l = (b, m') ->
    ~validBlock m b.
  
  Parameter allocValidI : forall m m' l b,
    alloc m l = (b, m') -> forall b',
      validBlock m' b' <-> b' = b \/ validBlock m b'.

  Parameter allocValidI1 : forall m m' l b,
    alloc m l = (b, m') -> forall b',
      validBlock m' b' -> 
      b' = b \/ validBlock m b'.
  
  Parameter allocValidI2 : forall m m' l b,
    alloc m l = (b, m') -> forall b',
      b' = b \/ validBlock m b' ->
      validBlock m' b'.

  Parameter storeValidI : forall m m' p v,
    store m p v = Some m' -> forall b,
      validBlock m' b <-> validBlock m b.

  Parameter storeValidI1 : forall m m' p v,
    store m p v = Some m' -> forall b,
      validBlock m' b -> validBlock m b.
  
  Parameter storeValidI2 : forall m m' p v,
    store m p v = Some m' -> forall b,
      validBlock m b -> validBlock m' b.
  
  Parameter freeValidI : forall m m' b,
    free m b = Some m' -> forall b',
      (validBlock m' b' <-> (b' <> b /\ validBlock m b')).

  Parameter freeValidI1 : forall m m' b,
    free m b = Some m' -> forall b',
      validBlock m' b' -> 
      b' <> b /\ validBlock m b'.

  Parameter freeValidI2 : forall m m' b,
    free m b = Some m' -> forall b',
      b' <> b -> 
      validBlock m b' -> validBlock m' b'.

  Parameter freeValidE : forall m b,
    validBlock m b ->
    exists m', free m b = Some m'.
  
  
  (** blockOffs properties *)

  Parameter coreBlockAddrsEq : forall m b,
    blockOffs m b = CM.blockOffs (core m) b.

  Parameter allocAddrsEq : forall m m' l b,
    alloc m l = (b, m') ->
    blockOffs m' b = zeroWordSet l.
  
  Parameter allocAddrsNeq : forall m m' l b,
    alloc m l = (b, m') -> forall b',
      b' <> b ->
      blockOffs m' b' = blockOffs m b'.
  
  Parameter storeAddrsEq : forall m m' p v,
    store m p v = Some m' -> forall b,
      blockOffs m' b = blockOffs m b.
  
  Parameter freeAddrsEq : forall m m' b,
    free m b = Some m' -> forall b',
      b' <> b ->
      blockOffs m' b' = blockOffs m b'.
  

  (* properties of valid pointers *)

  Parameter emptyPtrE : forall p,
    validPtr empty p ->
    False.

  Parameter corePtrE : forall m p,
    CM.validPtr (core m) p ->
    validPtr m p.

  Parameter allocPtrEqI : forall m m' l b,
    alloc m l = (b, m') -> forall i,
      (i < wordAlign l)%word ->
      validPtr m' (b, wordAlign i).

  Parameter allocPtrNeqI : forall m m' l b,
    alloc m l = (b, m') -> forall p,
      blockOf p <> b ->
      validPtr m p ->
      validPtr m' p.
  
  Parameter storePtrI1 : forall m m' p v,
    store m p v = Some m' -> forall p',
      validPtr m' p' -> validPtr m p'.
  
  Parameter storePtrI2 : forall m m' p v,
    store m p v = Some m' -> forall p',
      validPtr m p' -> validPtr m' p'.

  Parameter freePtrI1 : forall m m' b,
    free m b = Some m' -> forall p,
      blockOf p <> b ->
      validPtr m' p -> validPtr m p.
  
  Parameter freePtrI2 : forall m m' b,
    free m b = Some m' -> forall p,
      blockOf p <> b ->
	validPtr m p -> validPtr m' p.

  Parameter loadPtrI : forall m p v,
    load m p = Some v ->
    validPtr m p.
  
  Parameter loadPtrE : forall m p,
    validPtr m p ->
    exists v, load m p = Some v.

  Parameter storePtrE : forall m p,
    validPtr m p -> forall v,
      exists m', store m p v = Some m'.


  (* properties of core memory *)

  Parameter coreLoadEq : forall m p,
    CM.load (core m) p = load m p.

  Parameter coreStoreEq : forall m p v cm',
    CM.store (core m) p v = Some cm' ->
    exists m', store m p v = Some m' /\ core m' = cm'.

  Parameter coreAllocEq : forall m m' i sp',
    alloc m i = (sp', m') ->
    exists vv, core m' = CM.union (core m) (CM.block_mem sp' i vv).

  Parameter freeCoreEq : forall m m' b,
    free m b = Some m' -> forall len cm' vv,
      core m = CM.union (CM.block_mem b len vv) cm' ->
      ~CM.validBlock cm' b ->
      core m' = cm'.


End MemorySig.