(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export List.
Require Export SepLog.
Require Export Machine.


Set Implicit Arguments.

(* Program logic for the Cminor machine. *)


(* type abstractions *)
Definition tlist := list Type.

(* this is a little obnoxious *)
Definition tslist := list Set.
Fixpoint tslistToTlist (l:tslist) : tlist :=
  match l with 
    | nil => nil
    | A::l' => (A:Type) :: tslistToTlist l'
  end.
Coercion tslistToTlist : tslist >-> tlist.

Fixpoint tabs (t:tlist) : Type -> Type :=
  fun A =>
    match t with 
      | nil => A
      | B::t' => B -> tabs t' A
    end.

Fixpoint auxBindFA (A B:Type) (P:A -> B -> Prop) (ty:tlist) {struct ty} : 
  tabs ty A -> tabs ty B -> Prop :=
  match ty return tabs ty A -> tabs ty B -> Prop with
    | nil => fun p q => P p q
    | B::ty' => fun p q => forall b, auxBindFA P ty' (p b) (q b)
  end.

Fixpoint auxBindEX (A B:Type) (P:A -> B -> Prop) (ty:tlist) {struct ty} : 
  tabs ty A -> tabs ty B -> Prop :=
  match ty return tabs ty A -> tabs ty B -> Prop with
    | nil => fun p q => P p q
    | B::ty' => fun p q => exists b, auxBindEX P ty' (p b) (q b)
  end.

Fixpoint tabsMap (A:Type) (f:A -> A) (t:tlist) {struct t} :
  tabs t A -> tabs t A :=
  match t return tabs t A -> tabs t A with
    | nil => fun a => f a
    | B::t' => fun p b => tabsMap f t' (p b)
  end.

Definition tabsCons (A B:Type) (f:B -> A -> A) t (a:tabs t A) : 
  tabs (B::t) A :=
  fun b => tabsMap (f b) t a.


(* core state *)

Inductive cstate : Type :=
  | cstatei : CM.mem -> vfile -> block -> cstate.

Definition cmemOf (c:cstate) : CM.mem :=
  let (m, _, _) := c in m.

Definition cvfOf (c:cstate) : vfile :=
  let (_, vf, _) := c in vf.

Definition cspOf (c:cstate) : block :=
  let (_, _, sp) := c in sp.

Definition cset_mem (c:cstate) (m:CM.mem) : cstate :=
  let (_, vf, sp) := c in
    cstatei m vf sp.

Definition cset_vf (c:cstate) (vf:vfile) : cstate :=
  let (m, _, sp) := c in
    cstatei m vf sp.

Definition cset_sp (c:cstate) (sp:block) : cstate :=
  let (m, vf, _) := c in
    cstatei m vf sp.


(* definition of specs etc *)
Definition argTys := nat.
Fixpoint argTysToList (l:argTys) : tlist :=
  match l with
    | 0 => nil
    | S l' => (val:Type)::argTysToList l'
  end.

Definition spred := cstate -> Prop.

Definition cpred aty := tabs (argTysToList aty) mpred.

Definition rpred aty := tabs (argTysToList aty) (val -> mpred).

Inductive spec : Prop :=
  | speci : forall ty aty,
    tabs ty (cpred aty) ->
    tabs ty (rpred aty) ->
    spec.

Definition cdhpty := prod block word -> option spec.

Definition rspred := val -> spred.

Definition vcty := cdhpty -> spred -> (nat -> spred) -> rspred -> spred.
(* The cdhpty is the code heap type.
   The spred is the condition that must hold to fall off the end of the 
   current statement safely.
   The (nat -> spred) is the set of conditions to do an exit.
   The rspred is the condition to do a return, before the stack frame is freed
   The state is the state in which the statement will be executed *)


(* separating conjunction for a state predicates *)

Definition spSconj (p:spred) (A:mpred) : spred :=
  fun s => ((fun m => p (cset_mem s m)) ** A) (cmemOf s).
Infix "**s" := spSconj (at level 60).

Definition rspSconj (q:rspred) (A:mpred) : rspred :=
  fun v => spSconj (q v) A.
Infix "**r" := rspSconj (at level 60).


Lemma spSconjI : forall m m',
  dom_disj m m' -> forall (A:spred) vf sp,
    A (cstatei m vf sp) -> forall (B:mpred),
    B m' ->
    (A **s B) (cstatei (union m m') vf sp).
Proof.
  intros m m' D A vf sp Aok B Bok; unfold spSconj; simpl.
  econstructor; auto.
Qed.


Lemma rspSconjI : forall m m',
  dom_disj m m' -> forall (A:rspred) v vf sp,
    A v (cstatei m vf sp) -> forall (B:mpred),
    B m' ->
    (A **r B) v (cstatei (union m m') vf sp).
Proof.
  intros m m' D A v vf sp Aok B Bok; unfold rspSconj, spSconj; simpl.
  econstructor; auto.
Qed.



(* various auxiliary definitions used in VCs *)

Notation addExit := (fun b q n =>
  match n with
    | 0 => q
    | S n' => b n'
  end) (only parsing).

Definition someBind (A:Type) (o:option A) (P:A -> Prop) : Prop :=
  match o with
    | Some a => P a
    | None => False
  end.

Notation "'letsome' x <-- o ;;; f" :=
  (someBind o (fun x => f))
  (at level 201, x ident, right associativity).

Fixpoint addArgs A (vv:list val) (aty:argTys) {struct aty} :
  tabs (argTysToList aty) A -> option A :=
  match vv, aty return tabs (argTysToList aty) A -> option A with
    | nil, 0 => fun x => Some x
    | v::vv', S aty' => fun x => addArgs A vv' aty' (x v)
    | _, _ => fun _ => None
  end.


(* This gets around a sort-elimination problem I was having with using
   lets to destruct specs, though I don't understand why.  *)
Inductive callOkay (vv:list val) (q':spred) (retVar:var) :
  option spec -> cstate -> Prop :=
  | callOkayi : forall (ty:tlist) aty (p:tabs ty _) (q:tabs ty _) s,
    (exists A,
      auxBindEX (fun p0 (q0:rpred _) =>
        letsome p0' <-- addArgs _ vv aty p0;;;
        letsome q0' <-- addArgs _ vv aty q0;;;
        (p0' ** A) (cmemOf s) /\
        (forall v' m', (q0' v' ** A) m' ->
          (letsome vf' <-- VF.update (cvfOf s) retVar v';;;
            q' (cstatei m' vf' (cspOf s))))) ty p q) ->
    callOkay vv q' retVar (Some (speci ty aty p q)) s.

Definition storeVal s v v' := 
  match storeValM (cmemOf s) v v' with
    | Some m' => Some (cset_mem s m')
    | None => None
  end.

Definition loadValC (s:cstate) (v:val) : option val :=
  match v with
    | Vptr p => CM.load (cmemOf s) p
    | _ => None
  end.

Definition uopValC (s:cstate) (u:uop) (v:val) : option val :=
  match u with
    | ldop => loadValC s v
    | notop => Some (valNot v)
  end.

(* expression evaluation in core states instead of full states *)
Fixpoint expValC (s:cstate) (e:exp) {struct e} : option val :=
    match e with
      | eval v => Some v
      | evar x => VF.get (cvfOf s) x
      | esp => Some (Vptr (cspOf s, 0%word))
      | ebop b e1 e2 =>
	do v1 <-- expValC s e1;;;
	do v2 <-- expValC s e2;;;
        Some (bopVal b v1 v2)
      | eblop b e1 e2 =>
        do v1 <-- expValC s e1;;;
        match valBoolCases v1 with
          | inleft (left _) =>
            match b with
              | andop => do v2 <-- expValC s e2;;; Some (valToBool v2)
              | orop => Some (true:val)
            end
          | inleft (right _) =>
            match b with
              | andop => Some (false:val)
              | orop => do v2 <-- expValC s e2;;; Some (valToBool v2)
            end
          | inright _ => Some Vundef
        end
      | euop u e =>
	do v <-- expValC s e;;;
        uopValC s u v
    end.

Inductive expValsC (st:cstate) : list exp -> list val -> Prop :=
  | nil_expVals : expValsC st nil nil
  | cons_expVals : forall e v exps vals,
    expValsC st exps vals ->
    expValC st e = Some v ->
    expValsC st (e::exps) (v::vals).

(* Verification condition for statements *)


(* @stmVC is needed due to Coq bug 1804 *)
(* make letsome less binding than match? *)
Fixpoint stmVC st P q bt g s {struct st} : Prop :=
  match st with
    | vassign x e =>
      (letsome v <-- expValC s e;;;
        letsome vf' <-- VF.update (cvfOf s) x v;;;
        q (cset_vf s vf'))
    | store e1 e2 =>
      (letsome v1 <-- expValC s e1;;;
        letsome v2 <-- expValC s e2;;;
        letsome s' <-- storeVal s v1 v2;;;
        q s')
    | loop st' =>
      exists I : spred, I s /\ 
        forall s', I s' -> @stmVC st' P I bt g s'
    | block_st st' => 
      @stmVC st' P (fun _ => False) (addExit bt q) g s
    | exit k => bt k s
    | call retVar fe ee =>
      (letsome v <-- expValC s fe;;;
        letsome f <-- addrOf v;;;
        exists vv, expValsC s ee vv /\ callOkay vv q retVar (P f) s)
    | ret e =>
      (letsome v <-- expValC s e;;;
        g v s)
    | stcomp st1 st2 => @stmVC st1 P (@stmVC st2 P q bt g) bt g s
    | ifte e st1 st2 =>
      (letsome v <-- expValC s e;;;
        exists qo,
          let q' := match qo with Some qq => qq | None => q end in
	    match valBoolCases v with
	      | inleft (left _) => @stmVC st1 P q' bt g s
	      | inleft (right _) => @stmVC st2 P q' bt g s
	      | inright _ => False
	    end /\
            (forall s', q' s' -> q s'))
    | skip => q s
  end.


Lemma allocStackFrameOk : forall (A:mpred) m,
  A (core m) -> forall len sp' m',
    alloc m len = (sp', m') ->
    (A ** stackFrame sp' len) (core m').
Proof.
  intros A m Hp len sp' m' AE.
  destruct (coreAllocEq AE) as [vv E]; rewrite E.
  apply addFrameOk; auto.
  intro V; apply (allocFresh AE); apply coreValidE; auto.
Qed.
Implicit Arguments allocStackFrameOk [A m len sp' m'].


Lemma freeStackFrameOk : forall len sp A m,
  (stackFrame sp len ** A) (core m) -> forall m',
    free m sp = Some m' ->
    zeroWordSet len = blockOffs m sp ->
    A (core m').
Proof.
  intros len sp A m Hp m' E E'.
  destruct (dropFrameOk Hp) as [vv [cm' [Me [disj Aok]]]].
  replace (core m') with cm'; auto; clear Aok A Hp.
  symmetry; apply (freeCoreEq E Me); auto.

  intro V.
  rewrite coreBlockAddrsEq in E'; rewrite Me in E'; 
    rewrite CM.unionBlockAddrsEq in E'; rewrite CM.block_memAddrsEq in E'.
  clear E Me.

  assert (Sub : WS.Subset (CM.blockOffs cm' sp) (zeroWordSet len)).
  intros z Iz; rewrite E'; setsimpl; right; auto.

  clear E'.
  destruct (finite_wordSubsetE (CM.blockOffs cm' sp)) as [E | [x Ix]].
  eapply invalidBlockEmptyE; eauto.
  apply (disj (sp, x)); constructor; auto.
  apply block_memValidI.
  intro E; rewrite zeroWordSetEmptyEq in Sub; auto;
    assert (X := Sub _ Ix); setsimpl in X; auto.
  rewrite CM.block_memAddrsEq; auto.

Qed.


(* takes a predicate q describing the memory after the stack pointer
   is freed and turns it into a predicate describing the state before
   the stack pointer is freed *)
Notation specBeforeFree :=
  (fun A l s => (A ** stackFrame (cspOf s) l) (cmemOf s)) (only parsing).


Fixpoint argValList' aty : list val -> tabs (argTysToList aty) (list val) :=
  match aty return list val -> tabs (argTysToList aty) (list val) with
    | 0 => fun l => l
    | S aty' => fun l v => argValList' aty' (l ++ v :: nil) (*(v::l)*)
  end.
Definition argValList aty := argValList' aty nil.


Fixpoint auxBindFA3 (A B C:Type) (P:A -> B -> C -> Prop) (ty:tlist) {struct ty}:
  tabs ty A -> tabs ty B -> tabs ty C -> Prop :=
  match ty return tabs ty A -> tabs ty B -> tabs ty C -> Prop with
    | nil => fun p q r => P p q r
    | B::ty' => fun p q r => forall b, auxBindFA3 P ty' (p b) (q b) (r b)
  end.

Inductive fdefOk (P:cdhpty) (ty:tlist) aty (p:tabs _ (tabs _ _))
  : fdef -> tabs _ (tabs _ _) -> Prop :=
  | fdefOki : forall q argx locx stk body,
    aty = length argx ->
    auxBindFA (fun p0 (q0:rpred _) => 
      auxBindFA3 (fun p0' q0' vv =>
        forall m sp,
	  (p0' ** stackFrame sp stk) m ->
	  forall vf, vfEqv (argx ++ locx) (zip argx vv) vf ->
	    stmVC body P (fun _ => False) (fun _ _ => False)
          (fun v' => specBeforeFree (q0' v') stk)
	  (cstatei m vf sp)) (argTysToList aty) p0 q0 (argValList aty))
      ty p q ->
    @fdefOk P ty aty p (fdefi argx locx stk body) q.

Definition cdhpOk (P0:cdhpty) (C:cdhp) (P:cdhpty) : Prop :=
  forall f ty aty p q, P f = Some (speci ty aty p q) ->
    exists fd, C f = Some fd /\ fdefOk P ty aty p fd q.


(* Now the various judgements for reasoning about the machinery *)

(* this holds if there are no duplicated stack pointers, including
   the argument pointer, and all contained in the kcontrol are valid. *)
Inductive spFresh (m:mem) : block -> kcontrol -> Prop :=
  | stopFresh : forall sp,
    spFresh m sp Kstop
  | compFresh : forall sp s k,
    spFresh m sp k ->
    spFresh m sp (Kcomp s k)
  | blockFresh : forall sp k,
    spFresh m sp k ->
    spFresh m sp (Kblock k)
  | callFresh : forall rr fd sp sp' vf k,
    spFresh m sp k ->
    spFresh m sp' k ->
    sp <> sp' ->
    validBlock m sp' ->
    spFresh m sp (Kcall rr fd sp' vf k).


Inductive kcontrolOk (P:cdhpty) : kcontrol ->
  spred -> (nat -> spred) -> rspred -> Prop :=
  | kstopOk : forall q,
    kcontrolOk P Kstop q (fun _ _ => False) (fun _ _ => False)
  | kcompOk : forall st k (p:spred) q bt g,
    kcontrolOk P k q bt g ->
    (forall s, p s -> stmVC st P q bt g s) ->
    kcontrolOk P (Kcomp st k) p bt g
  | kblockOk : forall k (p:spred) q bt g,
    kcontrolOk P k q bt g ->
    kcontrolOk P (Kblock k) (fun _ => False) (addExit bt q) g
  | kcallOk : forall retVar fd sp' vf k (g:rspred) q' bt' g',
    kcontrolOk P k q' bt' g' ->
    (forall v s, g v s ->
      exists vf', VF.update vf retVar v = Some vf' /\
      specBeforeFree (fun m' =>
	q' (cstatei m' vf' sp')) (frameLen fd) s) ->
    kcontrolOk P (Kcall retVar fd sp' vf k)
      (fun _ => False) (fun _ _ => False) g.


(* every stack frame length in the control stack matches the function def *)
Inductive stackFramesOk (m:mem) : kcontrol -> block -> Prop :=
  | kstopFramesOk : forall sp,
    stackFramesOk m Kstop sp
  | kcompFramesOk : forall st k sp,
    stackFramesOk m k sp ->
    stackFramesOk m (Kcomp st k) sp
  | kblockFramesOk : forall k sp,
    stackFramesOk m k sp ->
    stackFramesOk m (Kblock k) sp
  | kcallFramesOk : forall rv fd sp sp' vf k,
    stackFramesOk m k sp' ->
    zeroWordSet (frameLen fd) = blockOffs m sp ->
    stackFramesOk m (Kcall rv fd sp' vf k) sp.


Definition cstateOf (s:state) : cstate :=
  cstatei (core (memOf s)) (vfOf s) (spOf s).

Inductive contOk (P:cdhpty) : cont -> Prop :=
  contOki : forall st k p bt g,
    kcontrolOk P k p bt g ->
    p (cstateOf st) ->
    validBlock (memOf st) (spOf st) ->
    spFresh (memOf st) (spOf st) k ->
    stackFramesOk (memOf st) k (spOf st) ->
    contOk P (st, k).


(* This holds if the continuation is stopped. *)
Inductive stopCont : cont -> Prop :=
  stopConti : forall st, stopCont (st, Kstop).


(* free preserves stack freshness *)
Lemma freeSpFresh : forall h sp k,
  spFresh h sp k -> forall sp',
    spFresh h sp' k -> forall h',
    free h sp = Some h' ->
    spFresh h' sp' k.
Proof.
  induction 1; inversion_clear 1; intros h' E; constructor; auto.
  apply (freeValidI2 E); auto.
Qed.


(* store preserves stack freshness *)
Lemma storeSpFresh : forall h sp k,
  spFresh h sp k -> forall p v h',
    M.store h p v = Some h' ->
    spFresh h' sp k.
Proof.
  induction 1; intros p v h' E; try (constructor; eauto).
  apply (storeValidI2 E); auto.
Qed.


(* alloc preserves stack freshness *)
Lemma allocSpFresh : forall h sp k,
  spFresh h sp k -> forall i sp' h',
    alloc h i = (sp', h') ->
    spFresh h' sp k.
Proof.
  induction 1; intros i sp'' h' HE; constructor; eauto.
  apply (allocValidI2 HE); auto.
Qed.


(* calling preserves stack freshness *)
Lemma callSpFresh : forall h sp k,
  spFresh h sp k -> 
  validBlock h sp -> forall i sp' h',
    alloc h i = (sp', h') -> forall rv fd vf,
      spFresh h' sp' (Kcall rv fd sp vf k).
Proof.
  intros h sp k Fr spOk i sp' h' HE rv fd vf.
  constructor.

  clear rv fd vf spOk; generalize dependent h; induction 1; intro HE;
    constructor; auto.
  eapply allocSpFresh; eauto.
  intro; subst sp'0; apply (allocFresh HE H0).
  apply (allocValidI2 HE); eauto.

  eapply allocSpFresh; eauto.

  intro; subst sp'; apply (allocFresh HE spOk).

  apply (allocValidI2 HE); eauto.

Qed.


(* freeing preserves stack frame lengths, for all but freed frame *)
Lemma freeStackFramesOk : forall h sp h',
  free h sp = Some h' -> forall k sp',
  stackFramesOk h k sp' ->
  spFresh h sp k ->
  sp <> sp' ->
  stackFramesOk h' k sp'.
Proof.
  intros h sp h' HE; induction 1; intros FR NE; inversion FR; clear FR; subst;
    constructor; eauto.
  rewrite H0; clear IHstackFramesOk H0.
  rewrite (freeAddrsEq HE); auto.
Qed.


(* storing preserves stack frame lengths *)
Lemma storeStackFramesOk : forall h h' p v,
  M.store h p v = Some h' -> forall k sp,
    stackFramesOk h k sp ->
    stackFramesOk h' k sp.
Proof.
  intros h h' p v HE; induction 1; constructor; auto.
  rewrite H0; rewrite (storeAddrsEq HE); auto.
Qed.


Lemma allocStackFramesOk : forall m m' len sp',
  alloc m len = (sp', m') -> forall k sp,
  stackFramesOk m k sp ->
  spFresh m sp k ->
  validBlock m sp ->
  stackFramesOk m' k sp.
Proof.
  intros m m' len sp' AE; induction 1; inversion_clear 1; constructor; auto.
  rewrite H0; clear H0 IHstackFramesOk.
  rewrite (allocAddrsNeq AE); auto.
  intro E; subst sp'; apply (allocFresh AE); auto.
Qed.


Lemma exitPopOk : forall P k q bt g,
  kcontrolOk P k q bt g -> forall s n,
    bt n (cstateOf s) ->
    validBlock (memOf s) (spOf s) ->
    spFresh (memOf s) (spOf s) k -> 
    stackFramesOk (memOf s) k (spOf s) -> forall C,
      exists c', step C (s, Kcomp (exit n) k) c' /\ contOk P c'.
Proof.
  induction 1; intros s n btOk spOk spFr CL C; try (elf; auto; fail).

  inversion_clear spFr; rename H1 into spFr'; inversion_clear CL.
  destruct (IHkcontrolOk _ _ btOk spOk spFr' H1 C) as [c' [ST CO]].
  exists c'; constructor; auto;
    inversion ST; subst; clear ST; econstructor; constructor; auto.

  clear IHkcontrolOk.
  inversion_clear spFr; inversion_clear CL.
  destruct n as [ |n]; [exists (s, k) | exists (s, Kcomp (exit n) k)];
    constructor; econstructor; eauto; try (econstructor; eauto).
Qed.


Lemma retPopOk : forall P k q bt g,
  kcontrolOk P k q bt g -> forall v s,
    g v (cstateOf s) -> forall e,
    expVal s e = Some v ->
    validBlock (memOf s) (spOf s) -> 
    spFresh (memOf s) (spOf s) k -> 
    stackFramesOk (memOf s) k (spOf s) -> forall C,
      exists c', step C (s, Kcomp (ret e) k) c' /\ contOk P c'.
Proof.
  induction 1; intros v s gOk e vOk bOk spFr CL C.
  
  destruct (freeValidE bOk); elf; eauto.

  inversion_clear spFr; rename H1 into spFr'; inversion_clear CL.
  destruct (IHkcontrolOk _ _ gOk _ vOk bOk spFr' H1 C)
    as [c' [stepOk c'Ok]];
    clear IHkcontrolOk gOk vOk bOk.
  inversion stepOk; clear stepOk; subst.
  exists (statei m' vf'' sp', k').
  constructor.
  econstructor; eauto; constructor; eauto.
  assumption.

  inversion_clear spFr; rename H0 into spFr'; inversion_clear CL.
  destruct (IHkcontrolOk _ _ gOk _ vOk bOk spFr' H0 C) 
    as [c' [stepOk c'Ok]];
    clear IHkcontrolOk gOk vOk bOk.
  inversion stepOk; clear stepOk; subst.
  exists (statei m' vf'' sp', k').
  constructor.
  econstructor; eauto; constructor; eauto.
  assumption.

  clear IHkcontrolOk; simpl in H0.
  destruct (freeValidE bOk) as [m' FE].
  destruct (H0 _ _ gOk) as [vf' [vEq qOk]]; clear H0 gOk.
  sconjLift2of2 qOk; hyr qOk (freeStackFrameOk qOk FE).
  inversion CL; clear CL; subst; rename H6 into CL; rename H7 into EE.
  hyr qOk (qOk EE); simpl in qOk; clear EE.
  inversion_clear spFr.
  exists (statei m' vf' sp', k).
  constructor; econstructor; eauto.
  constructor.
  apply (freeValidI2 FE); auto.
  apply (freeSpFresh H0 H1 FE).
  
  apply (freeStackFramesOk FE); auto.

Qed.


Lemma bsEq : forall A (x:A) (f:A -> Prop),
  someBind (Some x) f = f x.
Proof.
  reflexivity.
Qed.


Definition wordOfFdef (fd:fdef) :=
  let (_, _, w, _) := fd in w.


Lemma uopValCE : forall st u v v',
  uopValC (cstateOf st) u v = Some v' ->
  uopVal st u v = Some v'.
Proof.
  intros st u v v' E; destruct u; auto.
  simpl in * ; destruct v; try destruct p; 
    simpl in * ; auto; rewrite <- coreLoadEq; auto.
Qed.


Lemma expValCE : forall st e v,
  expValC (cstateOf st) e = Some v -> 
  expVal st e = Some v.
Proof.
  induction e; simpl; intros v' E; auto.

  destruct (expValC (cstateOf st) e).
  rewrite (IHe v); auto; apply uopValCE; auto.
  discriminate E.
  
  destruct (expValC (cstateOf st) e1). 2:destruct b; discriminate E.
  rewrite (IHe1 v). 2:auto.
  destruct (expValC (cstateOf st) e2); [ | discriminate E];
    injector E; rewrite (IHe2 v0); auto; fail.

  destruct (expValC (cstateOf st) e1). 2:discriminate E.
  rewrite (IHe1 v). 2:auto.
  destruct (valBoolCases v) as [[T | F] | U].

  destruct b.
  destruct (expValC (cstateOf st) e2). 2:discriminate E.
  injector E; rewrite (IHe2 v0); auto. auto.

  destruct b. auto.
  destruct (expValC (cstateOf st) e2). 2:discriminate E.
  injector E; rewrite (IHe2 v0); auto. 

  auto.

Qed.
Implicit Arguments expValCE [st e v].


Lemma expValsCE : forall st ee vv,
  expValsC (cstateOf st) ee vv ->
  expVals st ee vv.
Proof.
  induction ee; intros vv EE.
  inversion_clear EE; constructor.
  inversion_clear EE; constructor; auto.
  apply expValCE; auto.
Qed.


(* you can weaken all of the arguments to a well-formed statement VC *)
Theorem vcArgWeak : forall body P fallOk exitOk q s,
    stmVC body P fallOk exitOk q s -> 
    forall (fallOk' : spred) (exitOk' : nat -> spred) (q':rspred),
      (forall s, fallOk s -> fallOk' s) ->
      (forall k s, exitOk k s -> exitOk' k s) ->
      (forall v s, q v s -> q' v s) ->
      stmVC body P fallOk' exitOk' q' s.
Proof.
  induction body; simpl;
    intros until s; intros VC fallOk' exitOk' q' fallEq exitEq qEq.
  
  destruct (expValC s e); simpl in * ; auto.
  destruct (update (cvfOf s) v v0); simpl in * ; auto.

  destruct (expValC s e); simpl in *; auto.
  destruct (expValC s e0); simpl in *; auto.
  destruct (storeVal s v v0); simpl in *; auto.

  destruct VC as [I [Iok Imp]]; exists I; constructor; auto.
  clear s Iok; intros s Iok.
  apply (IHbody _ _ _ _ _ (Imp _ Iok)); auto; tauto.

  apply (IHbody _ _ _ _ _ VC); auto.
  clear s VC q' qEq; intros [|k]; simpl; auto.
  auto.

  destruct (expValC s e); simpl in *; auto.
  destruct (addrOf v0); simpl in *; auto.
  destruct VC as [vv [vvOk callOk]]; exists vv; constructor; auto.
  inversion callOk; clear callOk; subst; constructor.
  clear H vvOk.
  destruct H0 as [A H0]; exists A.
  induction ty; simpl in * .
  destruct (addArgs mpred vv aty p); simpl in * . 2:assumption.
  destruct (addArgs (val -> mpred) vv aty q0); simpl in *. 2:assumption.
  destruct H0 as [pOk X]; constructor; auto.
  intros v' m' qOk; hyr X (X _ _ qOk); unfold someBind in X |- * .
  destruct (update (cvfOf s) v v'); auto.

  destruct H0 as [b' H0]; exists b'.
  apply IHty; clear IHty; auto.
  
  destruct (expValC s e); simpl in VC |- * ; auto.

  apply (IHbody1 _ _ _ _ _ VC); clear IHbody1 VC s body1; auto.
  intros s X; apply (IHbody2 _ _ _ _ _ X); auto.

  destruct (expValC s e); simpl in *; auto.
  destruct VC as [q0 [VC qImp]].  exists q0.
  constructor.  2:destruct q0; autof.
  destruct (valBoolCases v) as [[ | ] | ]; auto.
  eapply IHbody1; eauto. destruct q0; autof.
  eapply IHbody2; eauto. destruct q0; autof.

  autof.
  
Qed.
Implicit Arguments vcArgWeak [body P fallOk exitOk q s].



(*** Now we work towards proving the frame rule *)

(* frame rule for expression evaluation *)
Lemma expValFrame : forall vf sp m e v,
  expValC (cstatei m vf sp) e = Some v -> forall m',
    dom_disj m m' ->
    expValC (cstatei (union m m') vf sp) e = Some v.
Proof.
  induction e; intros ve E m' D; simpl in * ; auto.

  (* case: unary operator *)
  assert (E' : exists v, expValC (cstatei m vf sp) e = Some v).
  destruct (expValC (cstatei m vf sp) e).
  econstructor; auto.
  discriminate.

  destruct E' as [v E'].
  rewrite E' in E.
  rewrite (IHe _ E'); auto.
  clear IHe E'.
  destruct u; simpl in * ; auto.
  destruct v; try destruct p; simpl in *; auto; apply loadUnionEq; auto.
  
  (* case: binary operator *)
  assert (E1 : exists v, expValC (cstatei m vf sp) e1 = Some v).
  destruct (expValC (cstatei m vf sp) e1).
  econstructor; auto.
  discriminate.

  destruct E1 as [v1 E1]; rewrite E1 in E.

  assert (E2 : exists v, expValC (cstatei m vf sp) e2 = Some v).
  destruct (expValC (cstatei m vf sp) e2).
  econstructor; auto.
  discriminate.

  destruct E2 as [v2 E2]; rewrite E2 in E.
  injector E.
  rewrite (IHe1 _ E1); auto.
  rewrite (IHe2 _ E2); auto.

  (* case: binary logical operator *)
  assert (E1 : exists v, expValC (cstatei m vf sp) e1 = Some v).
    destruct (expValC (cstatei m vf sp) e1).
    econstructor; auto.
    discriminate.
  destruct E1 as [v1 E1]; rewrite E1 in E.
  rewrite (IHe1 _ E1); auto.
  destruct (valBoolCases v1) as [[T | F] | U]. 3:auto.

  destruct b.
  assert (E2 : exists v, expValC (cstatei m vf sp) e2 = Some v).
    destruct (expValC (cstatei m vf sp) e2).
    econstructor; auto. discriminate.
  destruct E2 as [v2 E2]; rewrite E2 in E; injector E.
  rewrite (IHe2 _ E2); auto.
  auto.

  destruct b. auto.
  assert (E2 : exists v, expValC (cstatei m vf sp) e2 = Some v).
    destruct (expValC (cstatei m vf sp) e2).
    econstructor; auto. discriminate.
  destruct E2 as [v2 E2]; rewrite E2 in E; injector E.
  rewrite (IHe2 _ E2); auto.

Qed.
Implicit Arguments expValFrame [vf sp m e v m'].


(* frame rule for expressions evaluation *)
Lemma expValsFrame : forall vf sp m ee vv,
  expValsC (cstatei m vf sp) ee vv -> forall m',
    dom_disj m m' ->
    expValsC (cstatei (union m m') vf sp) ee vv.
Proof.
  induction 1; intros m' D; constructor; auto.
  apply (expValFrame H0 D).
Qed.
Implicit Arguments expValsFrame [vf sp m ee vv m'].


(* frame rule for callOkay *)
Lemma callOkayFrame : forall vv q n p A s,
  (callOkay vv q n p **s A) s ->
  callOkay vv (q **s A) n p s.
Proof.
  intros until A; intros [m0 vf sp]; unfold spSconj at 1; simpl;
    intros [m m' D Cok Aok].
  inversion Cok; clear Cok; subst; destruct H as [B H]; apply callOkayi.
  exists (B ** A).
  induction ty; simpl in * .

  destruct (addArgs mpred vv aty p0) as [Z | ]. 2:assumption.
  destruct (addArgs (val -> mpred) vv aty q0) as [Z' | ]. 2:assumption.
  simpl in *.
  destruct H as [pOk qImp]; constructor.
  apply sconjAssocL; econstructor; eauto.
  clear pOk; intros v' m'' Mok.
  hyr Mok (sconjAssocR Mok); destruct Mok as [m1 m2 D' Mok1 Mok2].
  hyr qImp (qImp _ _ Mok1); unfold someBind in qImp |- * .
  destruct (update vf n v'); auto.
  apply spSconjI; auto.

  destruct H as [b H]; exists b; eauto.

Qed.


(* frame rule for statements *)
Theorem stmFrame : forall body P fallOk exitOk q A s,
  (stmVC body P fallOk exitOk q **s A) s ->
  stmVC body P (fallOk **s A) (fun k => exitOk k **s A) (q **r A) s.
Proof.
  (* this form of the theorem is a little easier to induct on *)
  cut (forall body P q sp fallOk exitOk vf m,
      stmVC body P fallOk exitOk q (cstatei m vf sp) ->
      forall m' (A:mpred),
        A m' ->
        dom_disj m m' ->
        stmVC body P (fallOk **s A) (fun k => exitOk k **s A)
        (q **r A) (cstatei (union m m') vf sp)).
  intros X body P fallOk exitOk q A [m0 vf sp] VC; unfold spSconj in VC;
    simpl in VC; destruct VC as [m m' D VC Aok]; auto.
  
  induction body; simpl;
    intros P q sp fallOk exitOk vf m VC m' A Aok D.

  (* case: assign *)
  assert (E : exists v, expValC (cstatei m vf sp) e = Some v).
  destruct (expValC (cstatei m vf sp) e); [econstructor | elf]; auto.
  destruct E as [va E]; rewrite E in VC.
  rewrite (expValFrame E D); simpl in * .
  clear E; assert (E : exists vf', update vf v va = Some vf').
  destruct (update vf v va); [econstructor | elf]; auto.
  destruct E as [vf' E]; rewrite E in VC; rewrite E; clear E; simpl in * .
  apply spSconjI; auto.

  (* case: store *)
  assert (E : exists v, expValC (cstatei m vf sp) e = Some v).
  destruct (expValC (cstatei m vf sp) e); [econstructor | elf]; auto.
  destruct E as [v E]; rewrite E in VC.
  rewrite (expValFrame E D); clear E; simpl in * .
  assert (E : exists v, expValC (cstatei m vf sp) e0 = Some v).
  destruct (expValC (cstatei m vf sp) e0); [econstructor | elf]; auto.
  destruct E as [v' E]; rewrite E in VC.
  rewrite (expValFrame E D); clear E; 
    unfold storeVal, storeValM in *; simpl in * .
  destruct v; try destruct p; simpl in *; try (elf; auto; fail).
  assert (E : exists m', CM.store m a v' = Some m').
  destruct (CM.store m a v'); [econstructor | elf]; auto.
  destruct E as [m'' E]; rewrite E in VC.
  assert (D' : dom_disj m'' m').
  apply (dom_disjStore E); auto.
  rewrite (storeUnionEq E D); clear E; simpl in * .
  apply spSconjI; auto.

  (* case: loop *)
  destruct VC as [Inv [InvOk InvImp]]; exists (Inv **s A); constructor.
  apply spSconjI; auto.
  clear InvOk Aok D m m' vf sp; intros [m vf sp] InvOk.
  unfold spSconj in InvOk; simpl in InvOk.
  destruct InvOk as [m1 m2 D' Mok1 Mok2].
  assert (VC := InvImp _ Mok1); clear InvImp Mok1.
  exact (IHbody _ _ _ _ _ _ _ VC _ _ Mok2 D').

  (* case: block *)
  fold (stmVC body).
  hyr IHbody (IHbody _ _ _ _ _ _ _ VC _ _ Aok D).
  apply (vcArgWeak IHbody).
  intros s Mok; destruct Mok; auto.
  intros [ | k]; simpl; auto.
  auto.

  (* case: exit *)
  apply spSconjI; auto.

  (* case: call *)
  assert (E : exists v, expValC (cstatei m vf sp) e = Some v).
  destruct (expValC (cstatei m vf sp) e); [econstructor | elf]; auto.
  destruct E as [va E]; rewrite E in VC.
  rewrite (expValFrame E D); clear E; simpl in * .
  destruct (addrOf va); simpl in *; auto.
  destruct VC as [vv [vvOk VC]]; exists vv; constructor.
  apply (expValsFrame vvOk D).
  apply callOkayFrame; unfold spSconj; simpl; econstructor; auto.

  (* return *)
  assert (E : exists v, expValC (cstatei m vf sp) e = Some v).
  destruct (expValC (cstatei m vf sp) e); [econstructor | elf]; auto.
  destruct E as [va E]; rewrite E in VC.
  rewrite (expValFrame E D); clear E; simpl in * .
  apply rspSconjI; auto.

  (* composition *)
  hyr IHbody1 (IHbody1 _ _ _ _ _ _ _ VC _ _ Aok D); clear VC.
  apply (vcArgWeak IHbody1); clear IHbody1; [ | tauto | tauto].
  clear Aok m m' vf sp D; intros [m vf sp] H.
  unfold spSconj in H; simpl in H.
  destruct H as [m1 m2 D Mok1 Mok2].
  exact (IHbody2 _ _ _ _ _ _ _ Mok1 _ _ Mok2 D).

  (* branch *)
  assert (E : exists v, expValC (cstatei m vf sp) e = Some v).
  destruct (expValC (cstatei m vf sp) e); [econstructor | elf]; auto.
  destruct E as [v E]; rewrite E in VC.
  rewrite (expValFrame E D); clear E; simpl in * .
  destruct VC as [q' [VC qImp]].
  exists (match q' with 
            | Some qq => Some (qq **s A)
            | None => None
          end); constructor.
  destruct (valBoolCases v) as [[|]|]; simpl in *; destruct q'; 
    auto; intros VC'.
  unfold spSconj; destruct q'; intros; smatch.

  (* skip *)
  apply spSconjI; auto.

Qed.


(* applying stmFrame to the special case of the body of a function *)
Lemma frameFuncBody : forall body P p q sp stk argx locx vv,
  (forall m,
    (p vv ** stackFrame sp stk) m -> forall vf,
      vfEqv (argx ++ locx) (zip argx vv) vf ->
      stmVC body P (fun _ => False) (fun _ _ => False)
      (fun v' => specBeforeFree (q v') stk) (cstatei m vf sp)) ->
  forall m vf A,
    (p vv ** A ** stackFrame sp stk) m ->
    vfEqv (argx ++ locx) (zip argx vv) vf ->
    stmVC body P (fun _ => False) (fun _ _ => False)
      (fun v' => specBeforeFree (q v' ** A) stk) 
      (cstatei m vf sp).
Proof.
  intros body P p q sp stk argx loc vv Imp m vf A Mok Veq.
  scl3 Mok; scl2 Mok; apply sconjAssocR in Mok.
  destruct Mok as [m1 m2 D Mok1 Mok2].
  hyr Mok1 (Imp _ Mok1 _ Veq); clear Imp Veq.
  assert (Mok := spSconjI D _ _ _ Mok1 _ Mok2); clear D Mok1 Mok2.
  apply (vcArgWeak (stmFrame _ _ _ _ _ Mok)); clear Mok vf sp.
  intros s Mok; destruct Mok; auto.
  intros _ s Mok; destruct Mok; auto.
  unfold rspSconj, spSconj.
  intros v [m vf sp] Mok; simpl in * .
  sli; clear m1 m2 loc argx body P p.
  destruct Mok as [m1 m2 D Mok1 Mok2].
  gscl3; gscl2; apply sconjAssocL; econstructor; auto.
Qed.


Lemma frameFuncBody' : forall body P p q sp stk locx,
  (forall m,
    (p ** stackFrame sp stk) m -> forall vf,
      vfEqv locx nil vf ->
      stmVC body P (fun _ => False) (fun _ _ => False)
      (fun v' => specBeforeFree (q v') stk) (cstatei m vf sp)) ->
  forall m vf A,
    (p ** A ** stackFrame sp stk) m ->
    vfEqv locx nil vf ->
    stmVC body P (fun _ => False) (fun _ _ => False)
      (fun v' => specBeforeFree (q v' ** A) stk) 
      (cstatei m vf sp).
Proof.
  intros body P p q sp stk loc Imp m vf A Mok Veq.
  scl3 Mok; scl2 Mok; scar Mok.
  destruct Mok as [m1 m2 D Mok1 Mok2].
  hyr Mok1 (Imp _ Mok1 _ Veq); clear Imp Veq.
  assert (Mok := spSconjI D _ _ _ Mok1 _ Mok2); clear D Mok1 Mok2.
  apply (vcArgWeak (stmFrame _ _ _ _ _ Mok)); clear Mok vf sp.
  intros s Mok; destruct Mok; auto.
  intros _ s Mok; destruct Mok; auto.
  unfold rspSconj, spSconj.
  intros v [m vf sp] Mok; simpl in * .
  sli; clear m1 m2 loc body P p.
  destruct Mok as [m1 m2 D Mok1 Mok2].
  gscl3; gscl2; apply sconjAssocL; econstructor; auto.
Qed.


(* add an auxiliary variable for the frame rule *)

Definition specFramePre t t' (p:tabs t (tabs t' mpred)) :=
  tabsCons (fun A p => tabsCons (fun A p => p ** A) t' p A) t p.
Implicit Arguments specFramePre [t t'].

Definition specFramePost t t' (p:tabs t (tabs t' (val -> mpred))) :=
  tabsCons (fun A p => tabsCons (fun A p v => p v ** A) t' p A) t p.
Implicit Arguments specFramePost [t t'].


Fixpoint tabsInst A (vv:list val) {struct vv} : 
  (tabs (argTysToList (length vv)) A) -> A :=
  match vv return (tabs (argTysToList (length vv)) A) -> A with
    | nil => fun x => x
    | v::vv' => fun f => tabsInst A vv' (f v)
  end.


Definition auxBindInst : forall vv A B P TA TB,
  auxBindFA P (argTysToList (length vv)) TA TB ->
  P (tabsInst A vv TA) (tabsInst B _ TB).
Proof.
  induction vv; simpl; intros.
  exact H.
  exact (IHvv _ _ _ _ _ (H a)).
Qed.


Definition auxBindInst3 : forall vv A B C P TA TB TC,
  auxBindFA3 P (argTysToList (length vv)) TA TB TC ->
  P (tabsInst A vv TA) (tabsInst B _ TB) (tabsInst C _ TC).
Proof.
  induction vv; simpl; intros.
  exact H.
  exact (IHvv _ _ _ _ _ _ _ (H a)).
Qed.


Lemma consInstPush : forall (vv:list val) f
  (p : tabs (argTysToList (length vv)) mpred) A,
  mimpl (f A (tabsInst _ vv p))
  (tabsInst _ vv (tabsCons (B := mpred) f (argTysToList (length vv)) p A)).
Proof.
  induction vv; simpl; intros f p A m Hp.
  assumption.
  firstorder.
Qed.


Lemma consInstPushPost : forall (vv:list val) f
  (q : tabs (argTysToList (length vv)) (val -> mpred)) v A,
  mimpl (tabsInst _ vv 
    (tabsCons (B := mpred) f (argTysToList (length vv)) q A) v)
  (f A (tabsInst _ vv q) v).
Proof.
  induction vv; simpl; intros f q v A m Hp.
  assumption.
  apply IHvv; assumption.
Qed.


(* frame rule for function definitions *)
Lemma fdefFrame : forall P ty aty p q fd,
  fdefOk P ty aty p fd q ->
  fdefOk P (mpred::ty) aty (specFramePre p) fd (specFramePost q).
Proof.
  unfold specFramePre, specFramePost; intros P ty aty p q fd FD.
  inversion FD; clear FD; subst.
  apply fdefOki; auto.
  unfold tabsCons; simpl; intro A.
  induction ty; simpl in * .

  (* base case for ty.  induct on aty. *)
  rename H0 into FA.
  unfold argValList in *; generalize dependent (@nil val).
  generalize dependent (length argx).

  induction n; simpl; intros.

    (* base case for n *)
    sle H.
    eapply frameFuncBody. 3:eauto.
    instantiate (1 := fun _ => p). autof. autof.

    (* inductive case for n *)
    hyr FA (FA b); hyr IHn (IHn (p b)); autof.

  (* inductive case for ty *)
  intros b; hyr H0 (H0 b); hyr IHty (IHty (p b) (q b)); autof.

Qed.


Lemma addArgsImpl : forall vv p' p'',
  addArgs mpred vv (length vv) p' = Some p'' ->
  mimpl p'' (tabsInst mpred vv p').
Proof.
  induction vv; simpl; intros p' p'' E m Hp.
  injector E; auto.
  hyr IHvv (IHvv (p' a) p'' E).
  srewrite <- IHvv; auto.
Qed.
Implicit Arguments addArgsImpl [vv p' p''].


Lemma addArgsImplPost : forall vv q q',
  addArgs (val -> mpred) vv (length vv) q = Some q' -> forall v,
  mimpl (tabsInst _ vv q v) (q' v).
Proof.
  induction vv; simpl; intros q q' E v m Hp.
  injector E; auto.
  hyr IHvv (IHvv (q a) q' E v).
  srewrite <- IHvv; auto.
Qed.


Lemma tabsInstListEq : forall vv,
  tabsInst (list val) vv (argValList (length vv)) = vv.
Proof.
  cut (forall vv vv',
    tabsInst (list val) vv (argValList' (length vv) vv') = vv' ++ vv).
  intros H vv.
  unfold argValList; rewrite H.
  reflexivity.

  induction vv; simpl; intros. autorewrite with list; auto.
  rewrite IHvv; autorewrite with list; reflexivity.

Qed.


Theorem safety : forall P C c,
  contOk P c ->
  cdhpOk P C P ->
  (exists c', step C c c' /\ contOk P c') \/ stopCont c.
Proof.
  intros P C [st k]; inversion_clear 1;
    rename H0 into kOk; rename H1 into pOk; intro Cok.
  destruct kOk; [right; constructor | left | left | left].

  assert (Imp := H _ pOk); clear H pOk; rename H3 into spFr; 
    rename H2 into spOk; rename H4 into SF.
  rename st0 into s; destruct s; simpl in Imp.

  (* assign *)
  rename v into x.
  assert (E : exists v, expValC (cstateOf st) e = Some v).
  destruct (expValC (cstateOf st) e); [econstructor | elf]; auto.
  destruct E as [v E]; rewrite E in Imp; rewrite bsEq in Imp.
  hyr E (expValCE E).
  assert (E' : exists vf', update (cvfOf (cstateOf st)) x v = Some vf').
  simpl; destruct (update (vfOf st) x v); [econstructor | elf]; auto.
  destruct E' as [vf' E']; simpl in E'; rewrite E' in Imp; rewrite bsEq in Imp.
  exists (set_vf st vf', k); constructor; econstructor; eauto.
  inversion_clear spFr; auto.
  inversion_clear SF; auto.

  (* store *)
  rename e into e1; rename e0 into e2.
  assert (E1 : exists b, exists i, 
    expValC (cstateOf st) e1 = Some (Vptr (b, i))).
  destruct (expValC (cstateOf st) e1); [idtac | elf; auto].
  destruct (expValC (cstateOf st) e2); [idtac | elf; auto].
  unfold storeVal in Imp; 
    destruct v; try (elf; auto; fail); destruct v0; destruct a;
      econstructor; econstructor; reflexivity.
  destruct E1 as [b [i E1]]; rewrite E1 in Imp; rewrite bsEq in Imp.
  hyr E1 (expValCE E1).
  assert (E2 : exists v2, expValC (cstateOf st) e2 = Some v2).
  destruct (expValC (cstateOf st) e2); [econstructor | elf]; auto.
  destruct E2 as [v2 E2]; rewrite E2 in Imp; rewrite bsEq in Imp.
  hyr E2 (expValCE E2).
  assert (E3 : exists st', storeVal (cstateOf st) (Vptr (b, i)) v2 = Some st').
  destruct (storeVal (cstateOf st) (Vptr (b, i)) v2);
    [econstructor | elf]; auto.
  destruct E3 as [st' Meq]; rewrite Meq in Imp; rewrite bsEq in Imp.
  assert (Seq : cspOf st' = spOf st).
  unfold storeVal in Meq; unfold storeValM in Meq.
  simpl in Meq.
  destruct (CM.store (core (memOf st)) (b, i) v2).
  injection Meq; intro; subst; auto.
  discriminate Meq.
  unfold storeVal, storeValM in Meq; simpl in Meq.
  assert (Meq' : exists m', CM.store (core (memOf st)) (b, i) v2 = Some m').
  destruct (CM.store (core (memOf st)) (b, i) v2);
    [econstructor; auto | elf]; discriminate Meq.
  destruct Meq' as [m' Meq']; rewrite Meq' in Meq; injection Meq; intro; subst.
  clear Meq; destruct (coreStoreEq Meq') as [m [Meq coreEq]]; subst m'.
  exists (set_mem st m, k);
    constructor; econstructor; eauto; simpl.
  unfold map_mem; rewrite Meq; auto.
  apply (storeValidI2 Meq); auto.
  inversion_clear spFr; eapply storeSpFresh; eauto.
  apply (storeStackFramesOk Meq); auto; inversion_clear SF; auto.

  (* loop *)
  destruct Imp as [I [Iok vcOk]].
  exists (st, Kcomp s (Kcomp (loop s) k)).
  constructor; econstructor; eauto; econstructor; eauto;
    econstructor; eauto.
  intros s' Iok'; exists I; constructor; auto.

  (* block *)
  exists (st, Kcomp s (Kblock k)).
  constructor; econstructor; eauto; econstructor; eauto; econstructor; eauto.
  inversion_clear spFr; auto.
  inversion_clear SF; auto.

  (* exit *)
  inversion_clear spFr; inversion_clear SF; eapply exitPopOk; eauto.

  (* call *)
  Focus 1.
  rename e into fe.
  assert (FE : exists f, exists f',
    expValC (cstateOf st) fe = Some (Vptr (f, f'))).
    destruct (expValC (cstateOf st) fe); [| elf; autof].
    unfold addrOf in Imp; destruct v; simpl in Imp; try (elf; autof).
    destruct v0; try (elf; autof).
    destruct a; econstructor; econstructor; reflexivity.
  destruct FE as [f [f' FE]]; rewrite FE in Imp; unfold addrOf in Imp.
  rewrite bsEq in Imp; hyr FE (expValCE FE); simpl in Imp.
  destruct Imp as [vv [eeExp callOk]].
  hyr eeExp (expValsCE eeExp).

  inversion callOk as [ty args p' q' st' Imp Peq' E]; clear callOk; subst st'.
  assert (Peq : P (f, f') = Some (speci ty args p' q')) by auto. clear Peq'.
  destruct (Cok _ _ _ _ _ Peq) as [[argx locx stk body] [Ceq X]].
  assert (X' : exists fd, fdefOk P ty args p' fd q' /\ 
    fd = fdefi argx locx stk body).
    econstructor; constructor; eauto.
  clear X; destruct X' as [fd [X E]]; hyr X (fdefFrame X).
  inversion X; subst fd q0.
  injection H1; clear H1; intros; subst body0 stk0 locx0 argx0.
  rename H0 into Imp'; clear Cok Peq X.
  simpl in Imp, Imp'.

  assert (E' : exists X, alloc (memOf st) stk = X).
    econstructor; reflexivity.
  destruct E' as [[sp' m'] E'].
  assert (vfOk := fun E => @argInitOk argx vv E locx).
  cut (length argx = length vv).
  intros Evv; destruct vfOk as [vf vfEq]. auto.
  rewrite Evv in H; subst args.

  exists (statei m' vf sp',
    Kcomp body (Kcall v (fdefi argx locx stk body) (spOf st) (vfOf st) k)).
  constructor. econstructor; eassumption.

  destruct Imp as [A Imp]; hyr Imp' (Imp' A).
  induction ty; simpl in Imp, Imp', p', q'.
  assert (H : exists p'', addArgs mpred vv (length vv) p' = Some p'').
    destruct (addArgs mpred vv (length vv) p'). econstructor; auto. 
    elf; exact Imp.
  destruct H as [p'' p''E]; rewrite p''E in Imp; simpl in Imp.
  assert (H' : exists q'', addArgs _ vv (length vv) q' = Some q'').
    destruct (addArgs _ vv (length vv) q'). econstructor; auto. 
    elf; exact Imp.
  destruct H' as [q'' q''E]; rewrite q''E in Imp; simpl in Imp.
  destruct Imp as [p''Ok Imp].
  hyr p''Ok (allocStackFrameOk p''Ok E'); sle p''Ok.
  hyr Imp' (auxBindInst3 _ _ _ _ _ Imp'); cbv beta in Imp'.
  unfold specFramePre, specFramePost in Imp'.
  unfold tabsCons at 1 3 in Imp'; simpl in Imp'.
  econstructor. econstructor. econstructor. eauto.
  instantiate
    (1 := fun v' s => ((q'' v' ** A) ** stackFrame (cspOf s) stk) (cmemOf s)).
  simpl.
  intros v' s Hp.
  assert (VFE : exists vf', update (vfOf st) v v' = Some vf').
    clear - Imp Hp.
    destruct Hp.
    hyr Imp (Imp _ _ H0).
    destruct (update (vfOf st) v v'). econstructor; auto. elf; auto.
  destruct VFE as [vf' VFE]; exists vf'; constructor. assumption. smatch.
  clear - Imp VFE Hp.
  hyr Imp (Imp _ _ Hp).
  rewrite VFE in Imp; assumption.
  instantiate (1 := (fun s => s = cstatei (core m') vf sp')).
  cbv beta; intros; subst.
  apply (fun a b X c Y => vcArgWeak (Imp' a b X c Y)); clear Imp'.
  3:autof. 3:autof.
  srewrite <- consInstPush; sli; smatch.
  clear - p''Ok p''E.
  srewrite (addArgsImpl p''E) in p''Ok; auto.
  apply vfEqvI.
  clear - vfEq.
  rewrite tabsInstListEq; auto.
  clear - q''E; intros v s Hp.
  smatch. hyr Hp (consInstPushPost _ _ _ _ _ _ Hp); simpl in Hp.
  srewrite (addArgsImplPost _ _ q''E) in Hp; auto.
  cbv beta; reflexivity.

  simpl; apply (allocValidI2 E'); auto.
  inversion_clear spFr; constructor; eapply callSpFresh; eauto.
  simpl; constructor; constructor; simpl; [ | symmetry; apply (allocAddrsEq E')].
  inversion_clear spFr; inversion_clear SF; apply (allocStackFramesOk E'); auto.

  destruct Imp as [b Imp].
  apply (IHty _ _ Imp (Imp' b)).

  subst args; clear - Imp.
  destruct Imp as [A Imp].
  induction ty; simpl in *.
  generalize dependent vv.
  induction argx; destruct vv; intros. reflexivity. elf; autof. elf; autof.
  simpl in *; rewrite (IHargx _ _ _ Imp); auto.
  destruct Imp as [b Imp]; apply (IHty _ _ Imp).

  (* return *)
  assert (E : exists v, expValC (cstateOf st) e = Some v).
    destruct (expValC (cstateOf st) e); [econstructor | elf]; auto.
  destruct E as [v E]; rewrite E in Imp; rewrite bsEq in Imp.
  hyr E (expValCE E).
  inversion_clear spFr; eapply retPopOk; eauto.
  inversion_clear SF; auto.

  (* compose *)
  econstructor; constructor; econstructor; eauto;
    econstructor; eauto; econstructor; eauto.
  inversion_clear spFr; auto.
  inversion_clear SF; auto.

  (* ifte *)
  assert (E : exists v, expValC (cstateOf st) e = Some v).
  destruct (expValC (cstateOf st) e); [econstructor | elf]; auto.
  destruct E as [v E]; rewrite E in Imp; rewrite bsEq in Imp.
  apply expValCE in E.
  destruct Imp as [q' [Imp Qimp]].
  inversion_clear spFr; inversion_clear SF; destruct (valBoolCases v).
  destruct s; [exists (st, Kcomp s1 k) | exists (st, Kcomp s2 k)].
  constructor; econstructor; eauto; econstructor; eauto.
  intros s VC; apply (vcArgWeak VC); auto.
  subst v; constructor; [apply ifteFalseStep; auto | econstructor; eauto];
    econstructor; eauto.
  intros s VC; apply (vcArgWeak VC); auto.
  elf; assumption.

  (* skip *)
  inversion_clear spFr; inversion_clear SF;
    econstructor; constructor; econstructor; eauto; apply Imp; assumption.

  (* Kblock *)
  elf; eauto.

  (* Kcall *)
  elf; auto.

Qed.


Lemma fdefInstantiate : forall P t a aty p fd q,
  fdefOk P (a :: t) aty p fd q -> forall b,
    fdefOk P t aty (p b) fd (q b).
Proof.
  induction t; simpl; intros A aty p fd q Fok b; destruct fd;
    inversion_clear Fok; constructor; simpl in * ; auto.
Qed.


(* Weakening for fdefOk.  Can combine with frame rule. 
   A little hacky, but it works. *)
Theorem fdefOkWeaken : forall P t1 aty p1 fd q1,
  fdefOk P t1 aty p1 fd q1 -> forall t2 p2 q2,
  auxBindFA (fun p2' q2' =>
    auxBindEX (fun p1' q1' =>
      fdefOk P nil aty p1' fd q1' ->
      fdefOk P nil aty p2' fd q2') t1 p1 q1) t2 p2 q2 ->
  fdefOk P t2 aty p2 fd q2.
Proof.
  intros P t1 aty pre1 st post1 Fok t2 pre2 post2 FA.
  destruct st.
  induction t2; simpl in * .

  induction t1; simpl in * .
  auto.
  destruct FA as [b FA].
  hyr Fok (fdefInstantiate Fok b); eauto.

  constructor; simpl.
  2:intros b; assert (X := IHt2 _ _ (FA b)); clear IHt2 FA;
    inversion_clear X; auto.
  clear - Fok; inversion_clear Fok; autof.
Qed.


Lemma expValVarEq : forall bvars ups vf,
  vfEqv bvars ups vf -> forall x,
    inMapListDom x ups -> forall s,
      cvfOf s = vf ->
      expValC s x = Some (mapListFind ups x).
Proof.
  intros; simpl; destruct s; simpl in H1; subst.
  destruct x; eapply vfEqvGetEq; eauto.
Qed.


(* reasoning about statements *)

Lemma cmemOfEq : forall m vf sp,
  cmemOf (cstatei m vf sp) = m.
Proof.
  reflexivity.
Qed.

Lemma cvfOfEq : forall m vf sp,
  cvfOf (cstatei m vf sp) = vf.
Proof.
  reflexivity.
Qed.

Lemma cspOfEq : forall m vf sp,
  cspOf (cstatei m vf sp) = sp.
Proof.
  reflexivity.
Qed.

Lemma cset_memEq : forall m m' vf sp,
  cset_mem (cstatei m vf sp) m' = cstatei m' vf sp.
Proof.
  reflexivity.
Qed.

Lemma cset_vfEq : forall m vf vf' sp,
  cset_vf (cstatei m vf sp) vf' = cstatei m vf' sp.
Proof.
  reflexivity.
Qed.

Lemma cset_spEq : forall m vf sp sp',
  cset_sp (cstatei m vf sp) sp' = cstatei m vf sp'.
Proof.
  reflexivity.
Qed.


Lemma addrOfEq : forall a,
  addrOf (Vptr a) = Some a.
Proof.
  reflexivity.
Qed.


Hint Rewrite bsEq cset_memEq cset_vfEq cset_spEq addrOfEq 
  cmemOfEq cvfOfEq cspOfEq : ProgLog.


Lemma espValEq : forall s,
  expValC s esp = Some (Vptr (cspOf s, 0%word)).
Proof.
  reflexivity.
Qed.


Lemma euopValEq : forall s u e,
  expValC s (euop u e) = 
  match expValC s e with
    | Some v => 
      match u with
        | ldop => loadValC s v
        | notop => Some (valNot v)
      end
    | None => None
  end.
Proof.
  simpl; intros s u e; destruct (expValC s e); auto.
Qed.


Lemma loadValEq : forall i iv A m,
  (i |-> iv ** A) m -> forall s,
    m = (cmemOf s) ->
    loadValC s i = Some iv.
Proof.
  intros i iv A m [m1 m2 J HA _]; clear m.
  unfold loadValC; inversion HA; clear HA; subst.
  intros s EE; destruct s; simpl in EE; simpl; subst.
  apply loadUnionEq; auto; apply sing_memLoadEq.
Qed.


Lemma storeOk : forall x v A m,
  (x |-> v ** A) m -> forall v',
    exists m',
      (forall s, cmemOf s = m -> 
	storeVal s x v' = Some (cset_mem s m')) /\
      (x |-> v' ** A) m'.
Proof.
  intros x v A m Hp v'.
  destruct (storeOkM Hp v') as [m' [E Hp']]; exists m'; constructor; auto.
  unfold storeVal; intros s E'; subst m; simpl; rewrite E; auto.
Qed.
Implicit Arguments storeOk [x v A m].


Lemma beq_var_true : forall x,
  beq_var x x = true.
Proof.
  intros [x]; simpl; rewrite beq_true; auto.
Qed.

Lemma beq_var_false : forall x y,
  x <> y ->
  beq_var x y = false.
Proof.
  intros [x] [y] NE; simpl; rewrite beq_false; auto.
Qed.

Ltac varEvalStep VFE :=
  rewrite (expValVarEq VFE);
    [unfold mapListFind; rewrite beq_var_true;
      repeat (rewrite beq_var_false; [idtac | discriminate]) | 
        simpl; auto; tauto | auto; fail].

Ltac loadEvalStep Hp :=
  rewrite (loadValEq Hp); [idtac | auto; reflexivity]. 

Ltac storeEvalStep Hp v :=
  let m' := fresh "m" in
  let ME' := fresh "ME" in
  let Hp' := fresh "Hp" in
    destruct (storeOk Hp v) as [m' [ME' Hp']];
    rewrite ME'; [autorewrite with ProgLog;
      unfold cmemOf in Hp;
      match type of Hp with
        | (_ ** _) ?m => 
          first [clear m Hp ME'; rename m' into m; rename Hp' into Hp |
            clear Hp ME'; rename Hp' into Hp |
              try clear ME']
      end
      |  auto; reflexivity].


Ltac vfUpdateStep VFE x v :=
  let vf' := fresh "vf" in
  let VFEq := fresh "VFEq" in
  let VFE' := fresh "VFE" in
    destruct (vfUpdateOk VFE x v) as [vf' [VFeq VFE']];
      [simpl; auto; tauto | rewrite VFeq; autorewrite with ProgLog;
	[unfold cvfOf in VFE;
          match type of VFE with
           | vfEqv _ _ ?vf =>
             first [clear vf VFeq VFE; rename vf' into vf; 
                      rename VFE' into VFE |
               clear VFeq VFE; rename VFE' into VFE |
                 try clear VFeq]
         end
        | auto; reflexivity]].


Lemma evalValEq : forall s v,
  expValC s (eval v) = Some v.
Proof.
  auto.
Qed.


Lemma bopValEq : forall s b e1 e2,
  expValC s (ebop b e1 e2) =
  match expValC s e1, expValC s e2 with
    | Some v1, Some v2 => Some (bopVal b v1 v2)
    | _, _ => None
  end.
Proof.
  intros; auto.
Qed.


(* return the position of a binding of x in A *)
   (* I think there is now something like this in SepMatch *)
Ltac contSearch' A x k :=
  match A with
    | sconj (vcont x _) _ => constr:k
    | sconj _ ?A' => contSearch' A' x (S k)
    | vcont x _ => constr:k
    | _ => constr:0
  end.
Ltac contSearch A x := contSearch' A x 1.

(* search Hp for a binding for x and bring it to the front *)
Ltac contLift Hp x :=
  match type of Hp with
    | (?A ** ?B) _ =>
      let k := contSearch (A ** B) x in
        match eval hnf in k with
          | 0 => fail
          | _ => scl k Hp
        end
  end.


Hint Rewrite evalValEq bopValEq espValEq euopValEq : ProgLogExp.



(* deep reading *)

(* inputs: heap predicate, heap location
   output: value at that location in the heap
*)
Inductive readSepVal : mpred -> val -> val -> Prop :=
| base_rsv : forall v v',
  readSepVal (v |-> v') v v'
| sconj1_rsv : forall A B v v',
  readSepVal A v v' ->
  readSepVal (A ** B) v v'
| sconj2_rsv : forall A B v v',
  readSepVal B v v' ->
  readSepVal (A ** B) v v'
| and1_rsv : forall A B v v',
  readSepVal A v v' ->
  readSepVal (A ^^ B) v v'
| and2_rsv : forall A B v v',
  readSepVal B v v' ->
  readSepVal (A ^^ B) v v'.


Lemma readSepValOk' : forall A v v',
  readSepVal A v v' -> forall m,
    A m -> forall s,
      m = cmemOf s ->
      exists p, v = Vptr p /\ loadValC s v = Some v'.
Proof.
  intros A v v' RSV; induction RSV; intros m Hp s E; subst.
  
  (* base case *)
  inversion Hp; clear Hp; subst.
  destruct s; simpl in *; subst.
  exists p; intuition; apply sing_memLoadEq.

  (* sconj1 *)
  inversion Hp; clear Hp; subst.
  destruct s; simpl in *; subst.
  destruct (IHRSV _ H1 (cstatei m1 v0 b)) as [p [E E']]. auto.
  subst; exists p. constructor; auto. clear IHRSV.
  simpl in *; apply loadUnionEq; auto.

  (* sconj2 *)
  scl2 Hp; inversion Hp; clear Hp; subst.
  destruct s; simpl in *; subst.
  destruct (IHRSV _ H1 (cstatei m1 v0 b)) as [p [E E']]. auto.
  subst; exists p. constructor; auto. clear IHRSV.
  simpl in *; apply loadUnionEq; auto.

  (* and1 *)
  destruct Hp as [Hp _].
  apply (IHRSV _ Hp); auto.

  (* and2 *)
  destruct Hp as [_ Hp].
  apply (IHRSV _ Hp); auto.

Qed.


Theorem readSepValOk : forall A v v',
  readSepVal A v v' -> forall m,
    A m -> forall s,
      m = cmemOf s ->
      loadValC s v = Some v'.
Proof.
  intros A v v' RSV m Hp s E.
  destruct (readSepValOk' RSV Hp _ E) as [a [E1 E2]]; auto.
Qed.


Ltac findReadSepVal A0 v :=
  match A0 with
    | (v |-> ?v') => constr:(base_rsv v v')
    | (?A ** ?B) => 
      let H := findReadSepVal A v in
        constr:(sconj1_rsv B H)
    | (?A ** ?B) => 
      let H := findReadSepVal B v in
        constr:(sconj2_rsv A H)
    | (?A ^^ ?B) =>
      let H := findReadSepVal A v in
        constr:(and1_rsv B H)
    | (?A ^^ ?B) =>
      let H := findReadSepVal B v in
        constr:(and1_rsv A H)
  end.

Ltac newSepRead Hp x :=
  match type of Hp with
    | (?A ** ?B) _ =>
      let H := findReadSepVal (A ** B) x in
        rewrite (readSepValOk H Hp); [ | auto; reflexivity]
  end.


(* deep writing *)

Require Import ListAddrSet.

Definition memDom (s:AS.t) : mpred := 
  fun m => (forall a, AS.In a s <-> CM.validPtr m a).

Definition domp (P : AS.t -> Prop) : mpred := 
  lexists s, !(P s) ** memDom s.


(* inputs: heap predicate, value being written to, value being written
   output: new heap predicate 
*)
Inductive writeSepVal : mpred -> val -> val -> mpred -> Prop :=
| base_wsv : forall v v' v'',
  writeSepVal (v |-> v') v v'' (v |-> v'')
| sconj1_wsv : forall A A' B v v',
  writeSepVal A v v' A' ->
  writeSepVal (A ** B) v v' (A' ** B)
| sconj2_wsv : forall A B B' v v',
  writeSepVal B v v' B' ->
  writeSepVal (A ** B) v v' (A ** B')
| and_domp_wsv : forall A A' v v' P,
  writeSepVal A v v' A' ->
  writeSepVal (A ^^ domp P) v v' (A' ^^ domp P).


Theorem memDom_store : forall m m' a v,
  CM.store m a v = Some m' -> forall s,
  memDom s m ->
  memDom s m'.
Proof.
  unfold memDom; intros m m' a v E s X a'.
  rewrite (X a'); clear X. constructor; intros H.
  exact (CM.storePtrI2 E H).
  exact (CM.storePtrI1 E H).
Qed.


Theorem domp_store : forall m m' a v,
  CM.store m a v = Some m' -> forall P,
  domp P m ->
  domp P m'.
Proof.
  unfold domp; intros m m' a v E P Hp.
  sle Hp. exists x; sli.
  apply (memDom_store E); auto.
Qed.


Lemma writeSepValOk' : forall A A' a v,
  writeSepVal A a v A' -> forall m,
    A m -> 
    exists m', exists p,
      CM.store m p v = Some m' /\
      A' m' /\
      a = Vptr p.
Proof.
  intros A A' a v WSV; induction WSV; intros m Hp.
  
  (* base case *)
  destruct (fun X => @storeOkM v v' emp m X v'') as [m' [E Hp']].
  sli. sle Hp'. exists m'.
  inversion Hp. clear Hp; subst.
  exists p; constructor.
  unfold storeValM in E. destruct (CM.store (sing_mem p v')); auto. 
  constructor; auto.

  (* sconj1 *)
  inversion Hp; clear Hp; subst.
  destruct (IHWSV _ H0) as [m' [a [Im' [H0' aE]]]]; clear IHWSV H0 WSV; subst.
  exists (union m' m2); exists a; constructor.
  apply storeUnionEq; auto. constructor. 2:auto.
  constructor; auto. apply dom_disjStore with m1 a v'; auto.

  (* sconj2 *)
  scl2 Hp; inversion Hp; clear Hp; subst.
  destruct (IHWSV _ H0) as [m' [a [Im' [H0' aE]]]]; clear IHWSV H0 WSV; subst.
  exists (union m' m2); exists a; constructor.
  apply storeUnionEq; auto. constructor. 2:auto.
  gscl2; constructor; auto. apply dom_disjStore with m1 a v'; auto.

  (* and *)
  destruct Hp as [Hp1 Hp2].
  destruct (IHWSV _ Hp1) as [m' [p [E1 [E2 E3]]]]; clear IHWSV Hp1 WSV; subst.
  exists m'; exists p; constructor. auto. constructor. 2:auto.
  constructor. auto. eapply domp_store; eauto.

Qed.


Theorem writeSepValOk : forall A A' v v',
  writeSepVal A v v' A' -> forall m,
    A m -> 
    exists m',
      (forall s, cmemOf s = m -> storeVal s v v' = Some (cset_mem s m')) /\
      A' m'.
Proof.
  intros A A' v v' WSV m Hp.
  destruct (writeSepValOk' WSV _ Hp) as [m' [p [E1 [E2 E3]]]]; subst.
  exists m'; constructor; auto.
  intros; subst. unfold storeVal, storeValM. rewrite E1; auto.
Qed.


Ltac findWriteSepVal A0 v v':=
  match A0 with
    | (v |-> ?v0) => constr:(base_wsv v v0 v')
    | (?A ** ?B) => 
      let H := findWriteSepVal A v v' in
        constr:(sconj1_wsv B H)
    | (?A ** ?B) => 
      let H := findWriteSepVal B v v' in
        constr:(sconj2_wsv A H)
    | (?A ^^ domp ?P) =>
      let H := findWriteSepVal A v v' in
        constr:(and_domp_wsv P H)
  end.


Ltac newSepWrite Hp x v :=
  match type of Hp with
    | (?A ** ?B) ?m =>
      let H := findWriteSepVal (A ** B) x v in
      let m' := fresh "m" in
      let ME' := fresh "ME" in 
      let Hp' := fresh "Hp" in
        destruct (writeSepValOk H _ Hp) as [m' [ME' Hp']];
        rewrite ME'; [autorewrite with ProgLog;
          unfold cmemOf in Hp;
            first [clear m Hp ME'; rename m' into m; rename Hp' into Hp |
              clear Hp ME'; rename Hp' into Hp |
                try clear ME']
          | auto; reflexivity]
  end.



Ltac oldSepRead Hp x := contLift Hp x; loadEvalStep Hp.
Ltac oldSepWrite Hp x v := contLift Hp x; storeEvalStep Hp v.


Ltac fancyExpStep :=
  match goal with 
    | VFE : vfEqv _ _ _ |- context [expValC _ (evar _)] => varEvalStep VFE
    | VFE : vfEqv _ _ _ |- someBind (update _ ?x ?v) _ => 
      vfUpdateStep VFE x v; dropVfeDups VFE
    | Hp : (_ ** _) _ |- context [loadValC _ ?x] => newSepRead Hp x
    | Hp : vcont ?x _ _ |- context [loadValC _ ?x] => loadEvalStep Hp
    | Hp : (_ ** _) _ |- someBind (storeVal _ ?x ?v) _ => newSepWrite Hp x v
    | Hp : vcont ?x _ _ |- someBind (storeVal _ ?x ?v) _ => storeEvalStep Hp v
  end.
  (* if we allowed stores or vfile updates inside of expressions, 
     this wouldn't work, because the notion of a head for the VC
     would be more complex *)

Ltac vcSolve0 := autorewrite with ProgLog ProgLogExp; try fancyExpStep.

Ltac vcSolve := repeat (progress vcSolve0).


(* unfolding stmVC one step at a time.  is there a better way? *)

Lemma vassignVC : forall x e,
  stmVC (vassign x e) = 
  (fun _ q _ _ s =>
    letsome v <-- expValC s e;;; letsome vf' <-- VF.update (cvfOf s) x v;;;
    q (cset_vf s vf')).
Proof.
  reflexivity.
Qed.

Lemma stcompVC : forall st st',
  stmVC (stcomp st st') = 
     (fun P q bt g s => stmVC st P (stmVC st' P q bt g) bt g s).
Proof.
  reflexivity.
Qed.

Lemma storeVC : forall e1 e2,
  stmVC (store e1 e2) =
  (fun _ q _ _ s =>
    letsome v1 <-- expValC s e1;;;
    letsome v2 <-- expValC s e2;;;
    letsome s' <-- storeVal s v1 v2;;;
    q s').
Proof.
  reflexivity.
Qed.

Lemma loopVC : forall st,
  stmVC (loop st) = 
  (fun P _ bt g s => exists I : spred, I s /\ 
    forall s', I s' -> stmVC st P I bt g s').
Proof.
  reflexivity.
Qed.

Lemma blockVC : forall st,
  stmVC (block_st st) =
  (fun P q bt g s => stmVC st P (fun _ => False) (addExit bt q) g s).
Proof.
  reflexivity.
Qed.

Lemma exitVC : forall k,
  stmVC (exit k) = (fun _ _ bt _ s => bt k s).
Proof.
  reflexivity.
Qed.

Lemma callVC : forall x fe ee,
  stmVC (call x fe ee) =
      (fun P q _ _ s =>
        letsome v <-- expValC s fe;;;
        letsome f <-- addrOf v;;;
	(exists vv, expValsC s ee vv /\ callOkay vv q x (P f) s)).
Proof.
  reflexivity.
Qed.

Lemma retVC : forall e,
  stmVC (ret e) =
  (fun _ _ _ g s => letsome v <-- expValC s e;;; g v s).
Proof.
  reflexivity.
Qed.

Lemma ifteVC : forall e st1 st2,
  stmVC (ifte e st1 st2) =
  (fun P q bt g s =>
    letsome v <-- expValC s e;;;
    exists qo,
      let q' := match qo with Some qq => qq | None => q end in
	match valBoolCases v with
	  | inleft (left _) => stmVC st1 P q' bt g s
	  | inleft (right _) => stmVC st2 P q' bt g s
	  | inright _ => False
	end /\
        (forall s', q' s' -> q s')).
Proof.
  reflexivity.
Qed.

Lemma skipVC : stmVC skip = (fun _ q _ _ s => q s).
Proof.
  reflexivity.
Qed.


Ltac basicStmVCStep :=
  match goal with
    | |- stmVC (?s;; ?s')%CM _ _ _ _ _ => rewrite (stcompVC s s')
    | |- stmVC (?x <- ?e)%CM _ _ _ _ _ => rewrite (vassignVC x e)
    | |- stmVC ([?e] <- ?e')%CM _ _ _ _ _ => rewrite (storeVC e e')
    | |- stmVC (block_st ?s) _ _ _ _ _ => rewrite (blockVC s)
    | |- stmVC (exit ?k) _ _ _ _ _ => rewrite (exitVC k)
    | |- stmVC (ret ?e) _ _ _ _ _ => rewrite (retVC e)
    | |- stmVC skip _ _ _ _ _ => rewrite skipVC
  end.

Ltac bigStmVCStep :=
  match goal with
    | |- stmVC (ifte ?e ?s ?s') _ _ _ _ _ => 
      rewrite (ifteVC e s s'); vcSolve
    | |- stmVC (loop ?s) _ _ _ _ _ =>
      rewrite (loopVC s)
    | |- stmVC (call ?x ?fe ?ee) _ _ _ _ _ =>
      rewrite (callVC x fe ee); vcSolve
  end.

Ltac vcSteps := repeat (progress (repeat basicStmVCStep; vcSolve));
  try bigStmVCStep.


Ltac fdefBegin := constructor; 
  [reflexivity |
    cbv beta delta [auxBindFA3 auxBindFA tslistToTlist argTysToList] iota].


(* This is potentially dangerous, as it doesn't match only at the head.  
   Should be okay as long as we don't try to automatically apply it. *)
Ltac branchStep0 :=
  match goal with
    | |- context [valBoolCases ?P] =>
      let E := fresh "E" in
        destruct (valBoolCases P) as [[E | E] | E];
          autorewrite with valLogEq in E;
          autorewrite with valEq BlockOf in E;
          autorewrite with valEq2 valEq BlockOf in E;
            try discriminate E;
            try (elf; auto; fail);
          vcSteps
  end.

Ltac branchStep :=
  exists (None (A := spred));
    constructor; [branchStep0 | 
      let s := fresh in let X := fresh in intros s X; apply X].

Tactic Notation "branchStep" "with" constr(x) :=
  exists (Some x);
    constructor;
      [ branchStep0 | let s := fresh "s" in let X := fresh in intros s X].


(* Perform some common operations that we do at call sites.
   Invoke this after the argument list has been instantiated.
     (or use econstructor if you don't want to fill it out right away)
   The code heap type type lookup (something like (P f)) must be reduced to
   a somewhat unfolded specification (looking like (Some (bspeci pre post))).
*)
Ltac callStep :=
  constructor;
    [repeat (constructor; [ | vcSteps; auto]); constructor
      | constructor; autorewrite with ProgLog; unfold auxBindEX, tslistToTlist].