(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import SepLog.
Require Import List.


(* regression testing for the separation logic tactics *)


(* examples that previously caused bugs *)


(* when moving to 8.2, apply started delta reducing, which broke searchMatch *)
Definition HH (a b c d:addr) :=
  a |-> b ** c |-> d.

Lemma weirdness : forall A B a b c d m,
  (HH a b c d ** A) m ->
  mimpl A B ->
  (B ** HH a b c d) m.
Proof.
  intros until m; intros Hp ABimp.
  searchMatch.
Qed.


(* if a subterm simplified to unit, the result wasn't simplified *)
Lemma bangBangBad : forall (A:mpred) m,
  A m ->
  (A ** !(0 = 0) ** !(0 = 0)) m.
Proof.
  intros; ssimpl; auto.
Qed.


(* I was using lexIL and lexIR instead of lexEL and lexER *)
Lemma lexIntroOk : forall A B k x m,
  (A ** k |-> x ** B) m ->
  (((lexists y, A ** k |-> y) ** B)) m.
Proof.
  intros; ssimpl; eauto.  
Qed.



(* I forgot to generate a fresh E, which caused problems here *)
Lemma renaming : forall A B C E m,
  ((A ** B ** C) ** E) m ->
  (A ** B ** C ** E) m.
Proof.
  intros A B C E m Hp.
  ssimpl in Hp; exact Hp.
Qed.


(* This caused a problem at some point.  I think bangE was broken. *)
Lemma bangTest : forall A B C m,
  ((!A ** B) ** C) m ->
  A /\ (B ** C) m.
Proof.
  intros A B C m Hp; ssimpl in Hp; auto.
Qed.


Lemma complexTest : forall p x q s w1 v2 next ppath m,
  (p |-> Vword x ** emp ** top ** top ** (q |-> Vword x ** emp) ** top **
    (top ** s |-> Vword w1 ** emp ** !(w1 = v2) ** top ** next s |-> v2) **
    ppath s ** (foralls x in WS.empty, x |-> Vundef) ** next p |-> x ** 
      ppath x ** next q |-> x ** top ** ppath x) m ->
  (p |-> Vword x ** q |-> Vword x ** s |-> Vword w1 **
    next s |-> v2 ** ppath s ** next p |-> x ** ppath x ** 
      next q |-> x ** ppath x ** top) m.
Proof.
  intros until m; intro Hp.
  ssimpl in Hp.
  assumption.
Qed.


Lemma smatch_cont_match_range_right : forall A B x v w m,
  (A ** x |-> v) m ->
  v = w ->
  mimpl A B ->
  (B ** x |-> w) m.
Proof.
  intros until m; intros Hp Ev Aimp. smatch.
Qed.


(* various cases in search match I didn't cover initially *)

Lemma searchHypForGoalTest : forall m,
  emp m ->
  emp m.
Proof.
  intros m Hp.
  goalSearchLift emp.
  hypSearchLift Hp emp.
  searchHypForGoal Hp.
  searchGoalForHyp Hp.
  searchMatch.
Qed.


(* searchMatch should treat |->- the same way as |-> *)
Lemma searchMatch_any_val_match : forall x v A B B' m,
  (B ** x |-> v ** A) m ->
  mimpl B B' ->
  (A ** x |->- ** B') m.
Proof.
  intros until m; intros Hp Bimp. searchMatch.
Qed.  


(* smatch (and thus searchMatch) shouldn't cancel out top *)
Lemma searchMatch_no_cancel_top : forall A A' m,
  (A ** top) m ->
  mimpl A (A' ** top) ->
  (A' ** top) m.
Proof.
  intros A A' m Hp Aimp.
  searchMatch.  (* shouldn't do anything *)
  srewrite Aimp in Hp. ssimpl in Hp. exact Hp.
Qed.


(* 'unit tests' for various tactics *)


(* reassociator tests *)

Lemma TestAssocG : forall A B C D E m,
  (A ** ((B ** C) ** D) ** E) m ->
  (A ** B ** C ** D ** E) m.
Proof.
  intros A B C D E m Hp.
  assocConjG .[0, .[.[0, 0], 0], 0]%SL.
  exact Hp.
Qed.


Lemma TestAssocG2 : forall A B C D m,
  ((A ** B ** C) ** D) m ->
  (A ** B ** C ** D) m.
Proof.
  intros A B C D m Hp.
  assocConjG .[.[0, 0, 0], 0]%SL.
  exact Hp.
Qed.


Lemma TestAssoc : forall A B C D E m,
  (A ** B ** C ** D ** E) m ->
  (A ** ((B ** C) ** D) ** E) m.
Proof.
  intros A B C D E m Hp.
  assocConj Hp .[0, .[.[0, 0], 0], 0]%SL.
  exact Hp.
Qed.


Lemma TestAssoc2 : forall A B C D m,
  (A ** B ** C ** D) m ->
  ((A ** B ** C) ** D) m.
Proof.
  intros A B C D m Hp.
  assocConj Hp .[.[0, .[0, 0]], 0]%SL.
  exact Hp.
Qed.


(* null test for associator *)
Lemma TestAssoc3 : forall A B C D m,
  (A ** B ** C ** D) m ->
  (A ** B ** C ** D) m.
Proof.
  intros A B C D m Hp.
  assocConj Hp .[0, 0, 0, 0]%SL.
  exact Hp.
Qed.


(* permuter tests *)


Lemma TestPermG : forall A B C D E F G m,
  (G ** F ** E ** D ** C ** A ** B) m ->
  (A ** B ** C ** D ** E ** F ** G) m.
Proof.
  intros A B C D E F G m Hp.
  permConjG (7::6::5::4::3::1::2::nil).
  exact Hp.
Qed.


Lemma TestPerm : forall A B C D E F G m,
  (A ** B ** C ** D ** E ** F ** G) m ->
  (G ** F ** E ** D ** C ** A ** B) m.
Proof.
  intros A B C D E F G m Hp.
  permConj Hp (7::6::5::4::3::1::2::nil).
  exact Hp.
Qed.


(* reassociator + permuter tests *)
Lemma TestAspG : forall A B C D E F G m,
  (G ** F ** ((E ** D) ** C) ** A ** B) m ->
  (A ** B ** C ** D ** E ** F ** G) m.
Proof.
  intros A B C D E F G m Hp.
  srearr .[7, 6, .[.[5, 4], 3], 1, 2]%SL.
  exact Hp.
Qed.


Lemma TestAsp : forall A B C D E F G m,
  (A ** B ** C ** D ** E ** F ** G) m ->
  (G ** F ** ((E ** D) ** C) ** A ** B) m.
Proof.
  intros A B C D E F G m Hp.
  srearr .[7, 6, .[.[5, 4], 3], 1, 2]%SL in Hp.
  exact Hp.
Qed.


(* testing the handling of smatch0's adding in emp *)
Lemma empSmatch0Test1 : forall A B C D m,
  (A ** B ** C ** D) m ->
  (forall m', D m' -> emp m') ->
  (A ** B ** C) m.
Proof.
  intros; smatch0.
Qed.

Lemma empSmatch0Test2 : forall A B C (D:mpred) m,
  (A ** B ** C) m ->
  (forall m', emp m' -> D m') ->
  (A ** B ** C ** D) m.
Proof.
  intros; smatch0.
Qed.



(* tests of scl *)
Lemma sclTest : forall A B C D E m,
  (A ** B ** C ** D ** E) m ->
  (D ** A ** B ** C ** E) m.
Proof.
  intros until m; intros M; scl 4 M; exact M.
Qed.


Lemma sclTestEnd : forall A B C D E m,
  (A ** B ** C ** D ** E) m ->
  (E ** A ** B ** C ** D) m.
Proof.
  intros until m; intros M; scl 5 M; exact M.
Qed.


(* Various tests of advanced cont matching in csm0 and searchMatch *)
Lemma smatch_cont_test0 : forall a va va' m,
  (a |-> va) m ->
  va = va' ->
  (a |-> va') m.
Proof.
  intros until m; intros Hp Ea. smatch.
Qed.

Lemma smatch_cont_test : forall a b c d va va' vb vb' vc vd vd' m,
  (a |-> va ** b |-> vb ** c |-> vc ** d |-> vd) m ->
  va = va' ->
  vb = vb' ->
  vd = vd' ->
  (a |-> va' ** b |-> vb' ** c |->- ** d |-> vd') m.
Proof.
  intros until m; intros Hp Ea Eb Ed. smatch.
Qed.

Lemma searchMatch_cont_test : forall a b c d va va' vb vb' vc vd vd' A D m,
  (a |-> va ** A ** b |-> vb ** c |-> vc ** D ** d |-> vd) m ->
  va = va' ->
  vb = vb' ->
  vd = vd' ->
  (c |->- ** b |-> vb' ** D ** a |-> va' ** d |-> vd' ** A) m.
Proof.
  intros until m; intros Hp Ea Eb Ed. searchMatch.
Qed.



(* tests of searchMatch *)

(* full match *)
Lemma searchMatchTest2 : forall A B C D E F m,
  (D ** C ** A ** B ** F ** E) m ->
  (A ** B ** C ** D ** E ** F) m.
Proof.
  intros; searchMatch.
Qed.


(* partial match *)
Lemma searchMatchTest1 : forall A B C D E F G m,
  (E ** A ** F ** C ** G) m ->
  (forall m, (E ** F ** G) m -> (B ** D) m) ->
  (A ** B ** C ** D) m.
Proof.
  intros; searchMatch.
Qed.



(* setoid rewriting examples *)

Lemma rewriteGoalImpl : forall (A B C D:mpred) m,
  (A ** D ** C) m ->
  mimpl D B ->
  (A ** B ** C) m.
Proof.
  intros A B C D m M E.
  srewrite <- E.
  exact M.
Qed.

Lemma rewriteHypIff : forall (A B C D:mpred) m,
  (A ** B ** C) m ->
  miff B D ->
  (A ** D ** C) m.
Proof.
  intros A B C D m M E.
  srewrite E in M.
  exact M.
Qed.

Lemma srewriteTest : forall (A B C D:mpred) m,
  (A ** B ** C) m ->
  miff D B ->
  (A ** D ** C) m.
Proof.
  intros A B C D m M E.
  srewrite E.
  exact M.
Qed.

Lemma parameterExample : forall (A A':mpred) (B C: nat -> mpred) m,
  (forall n, mimpl (B n) (C n)) -> forall k,
    (A ** B k ** A') m ->
    (A ** C k ** A') m.
Proof.
  intros A A' B C m I k M.
  srewrite <- I.
  exact M.
Qed.