(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export Memory.
Import Memory.M.CM.
Require Export Val.
Require Import List.
Require ListAddrSet.
Require Import Addr.

Set Implicit Arguments.


(*** Separation logic predicates *)

Definition mpred := mem -> Prop.


(**** mpred equivalence and implication *)
Definition miff (A B:mpred) := forall m, A m <-> B m.
Definition mimpl (A B:mpred) := forall m, A m -> B m.


(** empty memory *)
Definition emp : mpred := fun m => m = empty.


(** the trivial predicate *)
Definition top : mpred := fun m => True.


(** separating conjunction *)
Inductive sconj (A B:mpred) : mpred :=
  sconji : forall m1 m2,
    dom_disj m1 m2 ->
    A m1 ->
    B m2 ->
    sconj A B (union m1 m2).

Infix "**" := sconj (at level 60, right associativity).


(** magic wand *)
Definition wand (A B:mpred) : mpred :=
  fun m => forall m', dom_disj m' m -> A m' -> B (union m' m).
Infix "-*" := wand (at level 61, right associativity).


(** non-separating conjunction *)
Definition sand (A B:mpred) : mpred :=
  fun m => A m /\ B m.

Infix "^^" := sand (at level 60, right associativity).


(** intuitionistic truth *)
Definition bang (A:Prop) : mpred :=
  fun m => emp m /\ A.

Notation "! A" := (bang A) (at level 45).


(** existential quantification *)
Definition lex (A:Type) (f:A -> mpred) : mpred :=
  fun s => exists x, f x s.

Notation "'lexists' x , A" := (lex (fun x => A))
  (at level 200, x ident).

Notation "'lexists' x : B , A" := (@lex B (fun x => A))
  (at level 200, x ident).


(** universal quantification *)
Definition lforal (T:Type) (A : T -> mpred) : mpred :=
  fun m => forall x, A x m.

Notation "'lforall' x , A" := (lforal (fun x => A))
  (at level 200, x ident).

Notation "'lforall' x : B , A" := (@lforal B (fun x => A))
  (at level 200, x ident).


(** heap consists entirely of heap location v1 which contains v2,
    which is not Vundef *)
Inductive vcont : val -> val -> mpred :=
  vconti : forall p v,
    vcont (Vptr p) v (sing_mem p v).
Infix "|->" := vcont (at level 59, no associativity).
Notation "x |->-" := (lexists xyz10532, vcont x xyz10532) (at level 59).


(*** Basic properties of the predicates *)

Lemma empI : forall m,
  m = empty ->
  emp m.
Proof.
  unfold emp; auto.
Qed.
Implicit Arguments empI [m].


Lemma empE : forall m,
  emp m ->
  m = empty.
Proof.
  unfold emp; auto.
Qed.
Implicit Arguments empE [m].


Lemma empIR : forall (A:mpred) m,
  A m ->
  (A ** emp) m.
Proof.
  intros A m HA.
  rewrite <- unionEmpty; constructor; auto.
  apply dom_disjEmpty.
  apply empI; auto.
Qed.
Implicit Arguments empIR [A m].


Lemma empER : forall m (A:mpred),
  (A ** emp) m ->
  A m.
Proof.
  intros m a [ma mb JAB HA HB].
  rewrite (empE HB); rewrite unionEmpty; auto.
Qed.
Implicit Arguments empER [m A].


Lemma topI : forall h,
  top h.
Proof.
  unfold top; auto.
Qed.


Lemma sconjComm : forall h A B,
  (A ** B) h ->
  (B ** A) h.
Proof.
  intros s A B [h1 h2 J HA HB].
  rewrite unionComm; apply sconji; auto; apply dom_disjComm; auto.
Qed.
Implicit Arguments sconjComm [h A B].


Lemma sconjAssocL : forall h A B C,
  ((A ** B) ** C) h ->
  (A ** B ** C) h.
Proof.
  intros h A B C [hab hc JABC Hp HC]; destruct Hp as [ha hb JAB HA HB].
  rewrite unionAssoc; constructor; auto.
  apply dom_disjUnionI; auto; apply dom_disjUnionE with hb; auto.
  constructor; auto; rewrite unionComm in JABC;
    apply dom_disjUnionE with ha; auto.
Qed.
Implicit Arguments sconjAssocL [h A B C].


Lemma sconjAssocR : forall h A B C,
  (A ** B ** C) h ->
  ((A ** B) ** C) h.
Proof.
  intros h A B C Hp.
  apply sconjComm; apply sconjAssocL; apply sconjComm; apply sconjAssocL;
    apply sconjComm; auto.
Qed.
Implicit Arguments sconjAssocR [h A B C].


Lemma sconjImpl : forall h A B,
  (A ** B) h -> forall (A':mpred),
    (forall h', A h' -> A' h') -> forall (B':mpred),
    (forall h', B h' -> B' h') ->
    (A' ** B') h.
Proof.
  intros h A B [ha hb JAB HA HB] A' Aimp B' Bimp; apply sconji; auto.
Qed.


Lemma wandI : forall A B,
  mimpl B (A -* (A ** B)).
Proof.
  unfold wand; intros A B m Hp m' D Hp'; constructor; auto.
Qed.


Lemma wandE : forall A B,
  mimpl (A ** (A -* B)) B.
Proof.
  unfold wand; intros A B m Hp; destruct Hp; auto.
Qed.



Lemma sandComm : forall A B m,
  (A ^^ B) m ->
  (B ^^ A) m.
Proof.
  intros A B m [HpA HpB]; constructor; auto.
Qed.


Lemma sandDistr : forall (A B C: mpred) m, 
  ((A ^^ B) ** C) m -> 
  ((A ** C) ^^ (B ** C)) m.
Proof.
  intros A B C m [mab mc J [HpA HpB] HpC]; constructor; constructor; eauto.
Qed.


Lemma bangI : forall h,
  emp h -> forall (P:Prop),
    P ->
    (!P) h.
Proof.
  unfold bang; auto.
Qed.


Lemma bangE1 : forall P h,
  (!P) h ->
  emp h.
Proof.
  unfold bang; intuition.
Qed.


Lemma bangE2 : forall P h,
  (!P) h ->
  P.
Proof.
  unfold bang; intuition.
Qed.


Lemma lexI : forall (A:Type) (f:A -> mpred) h,
  (exists x, f x h) ->
  lex f h.
Proof.
  unfold lex; auto.
Qed.


Lemma lexE : forall (A:Type) (f:A -> mpred) h,
  lex f h ->
  exists x, f x h.
Proof.
  unfold lex; auto.
Qed.


(* I'm not sure what the intro would look like for lforal *)

Lemma lforallE : forall T (t:T) A,
  mimpl (lforal A) (A t).
Proof.
  unfold lforal; intros T t A m Hp; auto.
Qed.



Lemma contI : forall p v,
  (Vptr p |-> v) (sing_mem p v).
Proof.
  intros; constructor.
Qed.


Lemma nullContE : forall v m,
  (null_ptr |-> v) m ->
  False.
Proof.
  inversion 1.
Qed.


Lemma vundefContE : forall v m,
  (Vundef |-> v) m ->
  False.
Proof.
  inversion 1.
Qed.


Lemma vcontDisj : forall v v' v'' h,
  (v |-> v' ** v |-> v'') h -> False.
Proof.
  intros v v' v'' h [h1 h2 JAB HA HB].
  inversion HA; inversion HB; clear HA HB; subst.
  injector H2.
  eapply dom_disjE; eauto; eapply loadPtrI; eauto; apply sing_memLoadEq.
Qed.



(*** derived or more complex props *)

Lemma sconjImplL : forall A B h,
  (A ** B) h -> forall (A':mpred),
    (forall h', A h' -> A' h') ->
    (A' ** B) h.
Proof.
  intros A B h Hp A' Imp; apply (sconjImpl Hp); auto.
Qed.


Lemma sconjImplR : forall A B m,
  (A ** B) m -> forall (B':mpred),
    (forall m, B m -> B' m) ->
    (A ** B') m.
Proof.
  intros A B m Hp A' Imp; apply (sconjImpl Hp); auto.
Qed.


Lemma empIL : forall (A:mpred) m,
  A m ->
  (emp ** A) m.
Proof.
  intros A m Hp; apply sconjComm; apply (empIR Hp).
Qed.
Implicit Arguments empIL [A m].


Lemma empEL : forall m (A:mpred),
  (emp ** A) m ->
  A m.
Proof.
  intros m A Hp; apply sconjComm in Hp; exact (empER Hp).
Qed.


Lemma lexIL : forall (A:Type) (P : A -> mpred) Q m,
  (lex P ** Q) m ->
  (lex (fun x => P x ** Q)) m.
Proof.
  intros A P Q m [ma mb JAB [x HA] HB].
  exists x; constructor; auto.
Qed.


Lemma lexIR : forall (A:Type) (P : A -> mpred) Q h,
  (Q ** lex P) h ->
  (lex (fun x => Q ** P x)) h.
Proof.
  intros A P Q h Hp; hyr Hp (lexIL (sconjComm Hp));
    destruct Hp as [x Hp]; exists x; apply sconjComm; auto.
Qed.


Lemma lexEL : forall (A:Type) (P : A -> mpred) Q m,
  (lex (fun x => P x ** Q)) m ->
  (lex P ** Q) m.
Proof.
  intros A P Q m [x [ma mb JAB HA HB]].
  constructor; auto; exists x; auto.
Qed.
Implicit Arguments lexEL [A P Q m].


Lemma lexER : forall (A:Type) (P : A -> mpred) Q m,
  (lex (fun x => Q ** P x)) m ->
  (Q ** lex P) m.
Proof.
  intros A P Q m [x Hp]; apply sconjComm; apply lexEL;
    exists x; apply sconjComm; auto.
Qed.
Implicit Arguments lexER [A P Q m].


Lemma bangER : forall A P m,
  (A ** !P) m ->
  A m /\ P.
Proof.
  intros A P m [ma mb JAB HA HB].
  assert (X := bangE2 HB); apply bangE1 in HB.
  constructor; auto; clear X.
  apply empER; constructor; auto.
Qed.


Lemma bangEL : forall A P m,
  (!P ** A) m ->
  A m /\ P.
Proof.
  intros A P m Hp; destruct (bangER (sconjComm Hp)); auto.
Qed.


Lemma bangIR : forall (A:mpred) m,
  A m -> forall (P:Prop),
  P ->
  (A ** !P) m.
Proof.
  intros A m Hp P Pok.
  rewrite <- unionEmpty; constructor; auto.
  apply dom_disjEmpty.
  apply bangI; auto; apply empI; auto.
Qed.


Lemma bangIL : forall (A:mpred) m,
  A m -> forall (P:Prop),
  P ->
  (!P ** A) m.
Proof.
  intros A m Hp P Pok; apply sconjComm; apply bangIR; auto.
Qed.


Lemma vcontDisjL : forall v v' v'' A m,
  (v |-> v' ** v |-> v'' ** A) m -> False.
Proof.
  intros v v' v'' m A Hp.
  apply sconjAssocR in Hp; inversion Hp.
  apply (vcontDisj H0).
Qed.





(** Iterated separating conjunction forall_\ast *)


Require Import FSets.


Module FAS (X : FSetEq).
  Import X.
  Import M.
  Import Miff.

  (* compat for 8.2beta3: new finite set doesn't 
     include same simplifications?  *)
  Hint Rewrite FM.remove_iff FM.add_iff FM.union_iff : XFS.
  Tactic Notation "xss" := autorewrite with XFS.
  Tactic Notation "xss" "in" hyp(H) := autorewrite with XFS in H.
  Hint Extern 4 (In _ _) => xss.


  Inductive fasP (P:elt -> mpred) : t -> mpred -> Prop :=
    | e_fasP : 
      fasP P empty emp
    | s_fasP : forall X x Q,
      fasP P X Q ->
      ~In x X ->
      fasP P (add x X) (P x ** Q).

  Hint Constructors fasP sconj.

  Definition fas (P:elt -> mpred) (X:t) : mpred :=
    lexists Q, !(fasP P X Q) ** Q.


  Lemma fasEmptyI : forall m,
    emp m -> forall P,
      fas P empty m.
  Proof.
    unfold fas; intros m Hp P.
    exists emp; apply bangIL; auto.
  Qed.

  Lemma fasI : forall P x X m,
    (P x ** fas P X) m -> 
    ~In x X ->
    (fas P (add x X)) m.
  Proof.
    unfold fas; intros P x X m Hp NI.
    apply lexIR in Hp; destruct Hp as [Q Hp]; exists (P x ** Q).
    hyr Hp (sconjAssocL (sconjComm Hp)).
    destruct (bangEL Hp) as [Hp' fasOk].
    apply bangIL.
    apply sconjComm; auto.
    auto.
  Qed.

  Lemma fasEmptyE : forall P m,
    fas P empty m ->
    emp m.
  Proof.
    unfold fas; intros P m [Q Hp].
    destruct (bangEL Hp) as [Hp' Qok].
    inversion Qok; subst; auto.
    elf; assert (Ix : In x empty).
    rewrite <- H; apply add_1; auto.
    apply (empty_1 Ix).
  Qed.


  Lemma fasInv : forall P X Q,
    fasP P X Q -> forall x,
      In x X ->
      exists Q', fasP P (remove x X) Q' /\
        (forall m, Q m <-> (P x ** Q') m).
  Proof.
    induction 1; intros y Iy.
    elf; apply (empty_1 Iy).

    destruct (X.eq_elt_dec x y).
    subst y; clear Iy.
    exists Q; constructor.
    replace (remove x (add x X)) with X.
    auto.
    apply set_ext; intros a; xss; constructor; intro Ia.
    constructor; auto; intro E; apply elt_eq in E; subst; auto.
    tauto.
    intros m; constructor; auto.
    xss in Iy; destruct Iy as [E | Iy'].
    apply elt_eq in E; elf; auto.
    destruct (IHfasP _ Iy') as [Q' [H' Imp']]; clear IHfasP Iy'.
    exists (P x ** Q'); constructor.
    replace (remove y (add x X)) with (add x (remove y X)).
    constructor; auto.
    intro I; apply remove_3 in I; auto.
    apply set_ext; intros a; constructor; intro Ia.
    xss in Ia; destruct Ia as [E | Ia].
    apply elt_eq in E; subst a; xss; constructor; auto.
    intros E; apply elt_eq in E; auto.
    destruct Ia; auto.
    xss in Ia; destruct Ia as [Ia NE]; xss; tauto.

    intros m; constructor; intro Hp.
    apply sconjComm; apply sconjAssocR; apply (sconjImpl Hp); auto.
    intros m' Hp'; apply sconjComm; destruct (Imp' m') as [Hp'' _]; auto.
    apply (sconjImpl (sconjAssocL (sconjComm Hp))); auto.
    intros m' Hp'; apply sconjComm in Hp';
      destruct (Imp' m') as [_ Hp'']; auto.
  Qed.


  Lemma fasE : forall x Z,
    X.M.In x Z -> forall P,
    mimpl (fas P Z) (P x ** fas P (remove x Z)).
  Proof.
    unfold fas; intros x Z Ix P m [Q Hp].
    destruct (bangEL Hp) as [Hp' fasOk]; clear Hp.
    apply lexER.
    destruct (fasInv fasOk Ix) as [Q' [fasOk' Imp]]; clear fasOk.
    exists Q'; apply sconjComm; apply sconjAssocR; apply bangIL; auto;
      destruct (Imp m); apply sconjComm; auto.
  Qed.


  Lemma fasEqv : forall x Z,
    X.M.In x Z -> forall P,
      miff (fas P Z) (P x ** fas P (remove x Z)).
  Proof.
    intros x Z Ix P m; constructor; intros Hp.
    apply (fasE Ix); auto.
    apply fasI in Hp.
    rewrite (X.set_ext (X.Miff.add_remove _ _ Ix)) in Hp; assumption.
    apply X.M.remove_1; auto.
  Qed.


  Lemma fasImplIn : forall s P m,
    fas P s m -> forall (P': _ -> mpred),
      (forall x m', X.M.In x s -> (P x) m' -> (P' x) m') ->
      fas P' s m.
  Proof.
    unfold fas; intros s P m [Q Hp] P' Imp.
    assert (Hp' : exists Q', fasP P' s Q' /\ Q' m).
    2:destruct Hp' as [Q' [H1 H2]]; exists Q'; apply bangIL; auto.
    destruct (bangEL Hp) as [Hp' fasOk]; clear Hp.
    generalize dependent m; generalize dependent Q.
    induction 1; intros m Hp.
    exists emp; auto.
    destruct Hp as [ma mb JAB HA HB].
    destruct (fun II => IHfasOk II _ HB) as [Q' [FA HB']]; clear IHfasOk HB.
    auto.
    exists (P' x ** Q'); auto 10.
  Qed.


  Lemma fasImpl : forall s P m,
    fas P s m -> forall (P': _ -> mpred),
      (forall x m', (P x) m' -> (P' x) m') ->
      fas P' s m.
  Proof.
    intros s P m Hp P' Imp; eapply fasImplIn; eauto.
  Qed.


  Lemma fasSingletonI : forall (P : elt -> mpred) x m,
    P x m ->
    fas P (singleton x) m.
  Proof.
    intros P x m Hp.
    replace (singleton x) with (add x empty).
    apply fasI.
    hyr Hp (empIR Hp).
    apply (sconjImpl Hp); clear m Hp; intros m Hp; auto.
    apply fasEmptyI; auto.
    intro I; apply (empty_1 I).
    clear; apply set_ext; intros z; constructor; intros Iz.
    destruct (eq_elt_dec x z).
    subst z; apply singleton_2; auto.
    assert (Iz' : In z empty).
    eapply add_3; eauto; intro E; apply H; apply elt_eq; auto.
    elf; apply (empty_1 Iz').
    apply singleton_1 in Iz.
    apply add_1; auto.
  Qed.


  Lemma fasSingletonE : forall (P : elt -> mpred) x m,
    fas P (singleton x) m ->
    P x m.
  Proof.
    intros P x m Hp.
    assert (Ix : In x (singleton x)).
      apply singleton_2; auto.
    hyr Hp (fasE Ix Hp); clear Ix.
    apply empER.
    apply (sconjImpl Hp); clear m Hp; intros m Hp; auto.
    replace (remove x (singleton x)) with empty in Hp.
    apply fasEmptyE with P; auto.
    clear; apply set_ext; intros z; setsimpl; constructor; intros Iz.
    elf; apply (empty_1 Iz).

    destruct (eq_elt_dec x z).

    subst z; elf; assert (~In x (remove x (singleton x))).
    apply remove_1; auto. auto.

    elf; apply remove_3 in Iz; apply singleton_1 in Iz.
    apply H; apply elt_eq; auto.

  Qed.


  Lemma fasDisjMerge : forall P s s',
    X.Disj s s' ->
    mimpl (fas P s ** fas P s') (fas P (X.M.union s s')).
  Proof.
    intros P.
    assert (forall s s', X.Disj s s' -> forall A,
      mimpl (fas P s ** fas P s' ** A) (fas P (X.M.union s s') ** A)).

    induction s using X.Miff.set_induction; intros s' D A m Hp.

    (* base case *)
    assert (s = X.M.empty)
      by (apply set_ext; apply X.Miff.empty_is_empty_1; auto).
    subst s.
    apply empEL.
    rewrite X.empty_union_eqv1.
    apply (sconjImpl Hp); auto.
    clear; intros h Hp; apply (fasEmptyE Hp).

    (* inductive case *)
    assert (s2 = add x s1).
    unfold Add in H0.
    apply set_ext.
    intros z; constructor; intros Iz.
    rewrite H0 in Iz; destruct Iz; [apply add_1 | apply add_2]; auto.
    rewrite H0; destruct (eq_elt_dec x z) as [E | NE].
    subst; auto.
    right; eapply add_3; eauto.
    intro X; apply NE; apply elt_eq; auto.

    clear H0; subst s2.
    assert (((P x ** fas P s1) ** fas P s' ** A) m) as Hp'.
    apply (sconjImpl Hp); clear -H; intros m Hp.
    replace s1 with (remove x (add x s1)).
    apply fasE; auto.
    rewrite (set_ext (Miff.remove_add _ _ H)); auto.
    auto.

    clear Hp; apply sconjAssocL in Hp'.
    assert ((fas P s1 ** fas P s' ** P x ** A) m) as Hp.
    apply sconjComm in Hp'; apply sconjAssocL in Hp'.
    apply (sconjImpl Hp'); clear; intros m Hp; auto.
    apply sconjAssocL in Hp; apply (sconjImpl Hp); clear; intros m Hp; auto.
    apply sconjComm; auto.

    clear Hp'.
    assert (Disj s1 s') as D'.
    apply DisjI; clear - H D; intros y Iy Iy'.
    apply (DisjE D) with y.
    apply add_2; auto.
    auto.
    assert (Hp' := IHs1 _ D' _ _ Hp); clear Hp IHs1.
    apply sconjAssocR in Hp'; apply (sconjImpl Hp'); clear -H D; intros m Hp.
    2:auto.
    replace (union (add x s1) s') with (add x (union s1 s')).
    apply fasI.
    apply sconjComm; assumption.
    intros Ix; apply union_1 in Ix; destruct Ix as [Ix | Ix].
    auto.
    apply (DisjE D) with x; auto.
    clear -H; apply set_ext; intros y; constructor; intros Iy.
    destruct (eq_elt_dec x y) as [E | NE].
    subst; auto.
    assert (Iy' : In y (union s1 s')).
    apply add_3 with x.
    intro E; apply NE; apply elt_eq; auto.
    auto.
    apply union_1 in Iy'; clear Iy; destruct Iy' as [Iy | Iy].
    auto.
    auto.

    apply union_1 in Iy; destruct Iy as [Iy | Iy].
    destruct (eq_elt_dec x y) as [E | NE].
    subst; auto.
    apply add_2; apply union_2; apply add_3 with x.
    intro E; apply NE; apply elt_eq; auto.
    assumption.
    auto.

    intros s s' D m Hp; apply empER; apply H; auto.
    apply sconjAssocL; apply empIR; auto.

  Qed.


  Lemma fasDisjSplit : forall P s s',
    X.Disj s s' ->
    mimpl (fas P (X.M.union s s')) (fas P s ** fas P s').
  Proof.
    intros P.
    induction s using X.Miff.set_induction; intros s' D m Hp.

    (* base case *)
    assert (s = X.M.empty)
      by (apply set_ext; apply X.Miff.empty_is_empty_1; auto); subst s.
    hyr Hp (empIL Hp).
    rewrite X.empty_union_eqv1 in Hp.
    apply (sconjImpl Hp); auto.
    clear; intros h Hp; apply (fasEmptyI Hp).

    (* inductive case *)
    assert (s2 = add x s1).
    unfold Add in H0.
    apply set_ext.
    intros z; constructor; intros Iz.
    rewrite H0 in Iz; destruct Iz; [apply add_1 | apply add_2]; auto.
    rewrite H0; destruct (eq_elt_dec x z) as [E | NE].
    subst; auto.
    right; eapply add_3; eauto.
    intro X; apply NE; apply elt_eq; auto.

    clear H0; subst s2.
    assert (Ix : In x (union (add x s1) s')).
      apply union_2; apply add_1; auto.
    hyr Hp (fasE Ix Hp); clear Ix.
    assert (Hp' : ((P x ** fas P s1) ** fas P s') m).
    apply sconjAssocR; apply (sconjImpl Hp); clear m Hp; intros m Hp. auto.
    apply IHs1; clear IHs1.
    apply DisjI; intros a Ia Ia'; apply (DisjE D) with a; [apply add_2 | ];
      auto.
    assert (SE : union s1 s' = remove x (union (add x s1) s')).
      apply set_ext; intros a; constructor; intros Ia.
      apply union_1 in Ia; destruct Ia; apply remove_2.
      intro E; assert (E' := elt_eq E); clear E; subst a; auto.
      apply union_2; apply add_2; auto.
      intro E; assert (E' := elt_eq E); clear E; subst a.
      apply (DisjE D) with x; auto.
      apply union_3; auto.
      assert (NE := fun X => @remove_1 (union (add x s1) s') x a X Ia).
      apply remove_3 in Ia; apply union_1 in Ia; destruct Ia.
      apply add_3 in H0; auto.
      apply union_3; auto.
      rewrite <- SE in Hp; auto.

      clear Hp.
      apply (sconjImpl Hp'); clear m Hp'; intros m Hp.
      apply fasI; auto.
      auto.

    Qed.


End FAS.


Module Export WFAS := FAS WSE.
Module AFAS := FAS ListAddrSet.ASE.


(* for some reason, Coq isn't smart enough to instantiate Z for fasE *)
Lemma fasE : forall x Z,
  WS.In x Z -> forall P,
    mimpl (fas P Z) (P x ** fas P (WS.remove x Z)).
Proof.
  intros x Z Ix P; apply fasE; auto.
Qed.

Lemma fasaE : forall x Z,
  ListAddrSet.AS.In x Z -> forall P,
    mimpl (AFAS.fas P Z) (P x ** AFAS.fas P (ListAddrSet.AS.remove x Z)).
Proof.
  intros x Z Ix P; apply AFAS.fasE; auto.
Qed.


Notation "'foralls' x 'in' s , A" := (fas (fun x => A) s)
  (at level 200, x ident).

Notation "'forallsa' x 'in' s , A" := (AFAS.fas (fun x => A) s)
  (at level 200, x ident).

Definition stackFrame (sp:block) len :=
  foralls i in zeroWordSet len, (sp + i)%addr |->-.


Lemma stackFrameZeroI : forall sp m,
  emp m ->
  (stackFrame sp 0) m.
Proof.
  unfold stackFrame; intros sp m Hp; setsimpl; eapply fasEmptyI; eauto.
Qed.


Lemma stackFrameZeroE : forall sp m,
  (stackFrame sp 0) m ->
  emp m.
Proof.
  unfold stackFrame; intros sp m Hp; setsimpl; eapply fasEmptyE; eauto.
Qed.




(* only match when we can solve exactly *)
Ltac smatch0 :=
  eauto;
  let m' := fresh "m" in
  let Hp' := fresh "Hp" in
    match goal with
      | Hp : (sconj ?A ?B) ?m |- (sconj ?A ?B) ?m =>
        apply Hp
      | Hp : (sconj ?A ?B) ?m |- (sconj ?A ?C) ?m =>
        apply (sconjImplR Hp); try (clear Hp); try (clear m);
          first [intros m Hp | intros m Hp' | 
                 intros m' Hp | intros m' Hp']; smatch0
      | Hp : (sconj ?A ?B) ?m |- ?A ?m =>
        apply empER; smatch0
      | Hp : ?A ?m |- (sconj ?A ?B) ?m =>
        apply (sconjImplR (empIR Hp)); try (clear Hp); try (clear m);
          first [intros m Hp | intros m Hp' | 
                 intros m' Hp | intros m' Hp']; smatch0
      | Hp : (sconj ?A ?B) ?m |- (sconj ?C ?B) ?m =>
        apply (sconjImplL Hp); try (clear Hp); try (clear m); 
          first [intros m Hp | intros m Hp' | 
                 intros m' Hp | intros m' Hp']; smatch0
    | |- top _ => apply topI
    | |- _ => idtac
  end.


Ltac scar Hp := apply sconjAssocR in Hp.
Ltac gscal := apply sconjAssocL.
