(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import SepCore.
Import Memory.M.CM.
Require Import Addr.


Set Implicit Arguments.


(*** Relating separation logic predicates to the machine *)

Definition storeValM (m:mem) (v v':val) : option mem :=
  match v with
    | Vptr p => 
      match (CM.store m p v') with
	| Some m' => Some m'
	| None => None
      end
    | _ => None
  end.


Lemma storeOkM' : forall x v m,
  (x |-> v) m -> forall v',
    exists m', 
      storeValM m x v' = Some m' /\
      (x |-> v') m'.
Proof.
  unfold storeValM; intros x v m Hp v'; inversion Hp; clear Hp; subst.
  rewrite sing_memStoreEq.
  econstructor; constructor; auto; constructor.
Qed.


Lemma storeOkM : forall x v A m,
  (x |-> v ** A) m -> forall v',
    exists m', 
      storeValM m x v' = Some m' /\
      (x |-> v' ** A) m'.
Proof.
  intros x v A m [ma mb JAB MA MB] v'.
  destruct (storeOkM' MA v') as [ma' [E MA']]; clear MA.
  destruct x; try (destruct p); simpl in E; try (discriminate E).
  assert (ME : store ma a v' = Some ma').
  destruct (store ma a v'); [auto | discriminate E].
  simpl; exists (union ma' mb); constructor.
  rewrite (storeUnionEq ME); auto.
  constructor; auto.
  apply (dom_disjStore ME); auto.
Qed.
Implicit Arguments storeOkM [x v A m].


Lemma fasBlockI : forall b len vv,
  (foralls x in zeroWordSet len, 
    (lexists v, Vptr (b, x) |-> v)) (block_mem b len vv).
Proof.
  intros b len; induction len using wordInd4; intros vv.

  (* base cases *)
  setsimpl; rewrite block_memEmptyEq.
  apply fasEmptyI; apply empI; auto.
  apply wordInj; auto.
  
  rewrite zeroWordSetEmptyEq.
  rewrite block_memEmptyEq.
  apply fasEmptyI; apply empI; auto.
  apply wordInj; auto.
  apply wordInj; auto.

  rewrite zeroWordSetEmptyEq.
  rewrite block_memEmptyEq.
  apply fasEmptyI; apply empI; auto.
  apply wordInj; auto.
  apply wordInj; auto.

  rewrite zeroWordSetEmptyEq.
  rewrite block_memEmptyEq.
  apply fasEmptyI; apply empI; auto.
  apply wordInj; auto.
  apply wordInj; auto.

  (* inductive case *)
  rewrite zeroWordSetSplitEnd; auto.
  apply fasI.
  rewrite block_memSplit; auto.
  constructor; auto.
  apply dom_disjBlock; auto.
  exists (List.hd Vundef vv); constructor.
  intro Iw.
  setsimpl in Iw; wunfold; omega.
Qed.


Lemma fasBlockE : forall (b:block) len m,
  (foralls x in zeroWordSet len, (lexists v, Vptr (b + x)%addr |-> v)) m ->
  exists vv, m = block_mem b len vv.
Proof.
  induction len using wordInd4; intros m Hp.

  (* base cases *)
  Ltac baseCase Hp :=
    rewrite zeroWordSetEmptyEq in Hp;
      [hyr Hp (fasEmptyE Hp); exists (List.nil (A := val));
        rewrite block_memEmptyEq; auto; apply wordInj; auto |
          apply wordInj; auto].
  baseCase Hp.
  baseCase Hp.
  baseCase Hp.
  baseCase Hp.

  (* inductive case *)
  rewrite zeroWordSetSplitEnd in Hp; auto.
  rename len into w.
  assert (I : WS.In (wordAlign w) (WS.add (wordAlign w) (zeroWordSet w))).
  setsimpl; auto.
  assert (X := fasE I); hyr Hp (X _ _ Hp); cbv beta in Hp; clear I X.
  inversion_clear Hp.
  destruct H1 as [v Hp].
  replace (WS.remove (wordAlign w) (WS.add (wordAlign w) (zeroWordSet w))) with
    (zeroWordSet w) in H2.
  hyr IHlen (IHlen _ H2); clear H2; destruct IHlen as [vv E]; subst m2.
  exists (List.cons v vv); rewrite block_memSplit; auto; simpl.
  inversion_clear Hp.
  replace (0 + wordAlign w)%word with (wordAlign w) by ring; auto.
  apply WSE.set_ext; intros x.
  assert (NI : ~WS.In (wordAlign w) (zeroWordSet w)).
  setsimpl; intros [_ LT]; wunfold; omega.
  assert (X := (WSE.set_ext (WSE.Miff.remove_add NI))).
  pattern (zeroWordSet w) at 1; rewrite <- X; tauto.
Qed.


Lemma addFrameOk : forall (A:mpred) m,
  A m -> forall b i,
  ~CM.validBlock m b -> forall vv,
  (A ** stackFrame b i) (union m (block_mem b i vv)).
Proof.
  intros A m Aok b i NI.
  constructor; auto.
  intros p' V V'.
  destruct (block_memPtrE V') as [E _]; subst.
  destruct V; auto.
  unfold stackFrame.
  unfold Addr.addrAdd; simpl.
  eapply fasImpl. apply fasBlockI.
  clear. cbv beta. intros x m Hp.
  replace (0 + x)%word with x. auto.
  unfold WSE.M.elt in * ; ring.
Qed.


Lemma dropFrameOk : forall b i (A:mpred) m,
  (stackFrame b i ** A) m ->
  exists vv, exists m',
    m = union (block_mem b i vv) m' /\ dom_disj m' (block_mem b i vv) /\
    A m'.
Proof.
  unfold stackFrame.
  intros b i A m Hp; destruct Hp as [m1 m2 J Ha Hb].
  destruct (fasBlockE _ Ha) as [vv E]; exists vv; subst.
  exists m2.
  constructor; auto; constructor; auto; apply dom_disjComm; auto.
Qed.


