(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import SepCore.
Require Import SepRearr.
Require Import SepSimpl.
Require Import SepSetoid.
Require Import List.

Require Import Addr.

Close Scope addr_scope.
Close Scope word_scope.




Set Implicit Arguments.


(* matching hypotheses with goals *)


Lemma cont_imp_eq : forall a v m,
  (a |-> v) m -> forall v',
    v = v' ->
    (a |-> v') m.
Proof.
  intros a v m Hp v' E; subst v'; assumption.
Qed.
Implicit Arguments cont_imp_eq [a v m].


Ltac cont_match_range Hp :=
  try (apply Hp); apply (cont_imp_eq Hp); eauto 1.

Ltac cont_match_dom Hp x y :=
  autorewrite with AddrSimpl in Hp |- *; try (apply Hp);
    try (replace y with x; [apply Hp | addrEqSolve]).


(* like smatch0, but smarter about matching |-> *)
Ltac smatch :=
  eauto;
  let m' := fresh "m" in
  let Hp' := fresh "Hp" in
    match goal with
      | Hp : (sconj ?A ?B) ?m |- (sconj ?A ?B) ?m =>
        simple apply Hp
      | Hp : (sconj ?A ?B) ?m |- (sconj ?A ?C) ?m =>
        simple apply (sconjImplR Hp); try (clear Hp); try (clear m);
          first [intros m Hp | intros m Hp' | 
                 intros m' Hp | intros m' Hp']; smatch
      | Hp : (sconj ?A ?B) ?m |- ?A ?m =>
        simple apply empER; smatch
      | Hp : ?A ?m |- (sconj ?A ?B) ?m =>
        simple apply (sconjImplR (empIR Hp)); try (clear Hp); try (clear m);
          first [intros m Hp | intros m Hp' | 
                 intros m' Hp | intros m' Hp']; smatch
      | |- top _ => simple apply topI

      (* cases for cont *)
      | Hp : (?x |-> ?v) ?m |- (?x |-> ?w) ?m => 
        cont_match_range Hp
      | Hp : (?x |-> ?v) ?m |- (?x |->-) ?m => 
        exists v; apply Hp
      | Hp : (sconj (?x |-> ?v) ?B) ?m |- (sconj (?x |-> ?w) ?C) ?m =>
        simple apply (sconjImpl Hp); clear Hp; try (clear m);
          first [intros m Hp | intros m' Hp]; 
            [cont_match_range Hp | smatch]
      | Hp : (sconj ?B (?x |-> ?v)) ?m |- (sconj ?C (?x |-> ?w)) ?m =>
        simple apply (sconjImpl Hp); clear Hp; try (clear m);
          first [intros m Hp | intros m' Hp]; 
            [smatch | cont_match_range Hp]
      | Hp : (sconj (?x |-> ?v) ?B) ?m |- (sconj (?x |->-) ?C) ?m =>
        simple apply (sconjImpl Hp); clear Hp; try (clear m);
          first [intros m Hp | intros m' Hp]; 
            [exists v; apply Hp | smatch]
      | Hp : (sconj ?A (?x |-> ?v)) ?m |- (sconj ?C (?x |->-)) ?m =>
        simple apply (sconjImpl Hp); clear Hp; try (clear m);
          first [intros m Hp | intros m' Hp]; [smatch | exists v; apply Hp]

      (* we don't want to cancel out top, which will be on the right *)
      | Hp : (sconj ?A top) ?m |- (sconj ?C top) ?m => idtac

      | Hp : (sconj ?A ?B) ?m |- (sconj ?C ?B) ?m =>
        simple apply (sconjImplL Hp); try (clear Hp); try (clear m); 
          first [intros m Hp | intros m Hp' | 
                 intros m' Hp | intros m' Hp']; smatch

    | |- _ => idtac
  end.


Ltac ssplit :=
  eauto; (* make something simpler? *)
    let m' := fresh "m" in
      match goal with
        | Hp : (sconj ?A ?B) ?m |- (sconj ?A ?B) ?m => simple apply Hp
              (* shouldn't this be covered by eauto? *)
        | Hp : (sconj ?A ?B) ?m |- (sconj ?A ?C) ?m =>
          simple apply (sconjImplR Hp); clear Hp; try (clear m);
            first [intros m Hp | intros m' Hp]; ssplit
        | Hp : (sconj ?A ?B) ?m |- (sconj ?C ?B) ?m =>
          simple apply (sconjImplL Hp); clear Hp; try (clear m);
            first [intros m Hp | intros m' Hp]; ssplit
        | Hp : (sconj _ _) ?m |- (sconj _ _) ?m =>
          simple apply (sconjImpl Hp); clear Hp; try (clear m);
            first [intros m Hp | intros m' Hp]; ssplit

        | |- top _ => simple apply topI

        | Hp : (?x |->-) ?m |- (?y |->-) ?m =>
          cont_match_dom Hp x y
        | Hp : (?x |-> ?v) ?m |- (?x |->-) ?m =>
          exists v; apply Hp
        | Hp : (?x |-> ?v) ?m |- (?y |->-) ?m =>
          exists v; cont_match_dom Hp x y
        | Hp : (?x |-> ?v) ?m |- (?y |-> ?v) ?m =>
          cont_match_dom Hp x y
        | Hp : (?x |-> ?v) ?m |- (?x |-> ?w) ?m =>
          cont_match_range Hp
        | Hp : (?x |-> ?v) ?m |- (?y |-> ?w) ?m =>
          autorewrite with AddrSimpl in Hp |- *; try (apply Hp);
            (replace y with x; [cont_match_range Hp | addrEqSolve])

        | |- _ => idtac
      end.




(* searchMatch

   1) Overview

   searchMatch simplifies goals involving separating conjunctions by 
   eliminating common parts with a hypothesis.  This works even if
   the common parts are not in the same order.

   I think it is quadratic in the number of conjuncts.
  
   Caveat: It may not work correctly in the presence of duplicates, but
   I think it is unlikely duplicates will arise.

   2) Implementation

   There are three phases to this tactic:
     1. move things in the hyp that are in the goal to the front of the hyp
     2. move things in the goal that are in the hyp to the front of the goal
     3. eliminate things common to the start of the hyp and goal, using
        an existing tactic, smatch

   Phase 1 and 2 are basically the same, and the order they are done in 
   shouldn't matter.

   A. Phase 1 (searchHypForGoal)

   To do phase 1, we consider each part of the goal, going from right-to-left.
   The order is important!  For example, in searchExample above, we consider
   F, then E, then D, etc.

   When we are considering, say, B, we determine which position D occurs in
   in the goal, by recursively examining the type of the hypothesis H.  This
   search will determine that B occurs in the 4th position of H.  Then we use
   the scl tactic above to lift B to the start of H.
  
   The end result of phase 1 is that all of the parts of the hypothesis that
   are in the goal are moved to the front.  Because we proceed from right
   to left when going through the goal, the order of the elements moved to
   the front matches the order they appear in the goal.

   In the example above, after phase 1 is run, the H and the goal are 
   the same.  But what if some of the things in the goal are not in the hyp?

   B. Phase 2 (searchGoalForHyp)

   Phase 2 is the same as phase 1, except that it looks for things from the hyp
   in the goal and moves them, instead of the other way around.  This is needed
   in the case where there are things in the goal that are not in the hyp.

   Consider this example: 
     hyp:  (E ** A ** F ** C ** G) m
     goal: (A ** B ** C ** D) m.
   
   After we run phase 1, we have:
     hyp:  (A ** C ** E ** F ** G) m
     goal: (A ** B ** C ** D) m

   The search moved C near the beginning, where it can be matched, but
   B (which does not occur in the hyp) is in the way, if we try to match
   leading things.

   By running phase 2, we pull all of the things that were matched in the goal
   to the front of the goal.  Because the order is preserved, the end result
   will be that all things that match will be in the front of hyp and goal,
   and in the same order.
   
   B. Phase 3

   Phase 3 applies the tactic smatch that will match up common prefixes of the
   goal and hypothesis.  For example,
     hyp:  (A ** B ** C ** E ** D) m
     goal: (A ** B ** D ** B ** C) m

   will become:
     hyp:  (C ** E ** D) m'
     goal: (D ** B ** C) m'

   (it does a little more than that, but that's all we need from it here)

*)

(* This is the basic searching tactic.

   It searches the conjuncts of A for B, and returns the position that
   B appears in, if any.  If B isn't present, 0 is returned.

   A should be in the usual right-normal form *)
Ltac sconjBasicSearch' A B k :=
  match A with
    | sconj B _ => constr:k
    | sconj _ ?A' => sconjBasicSearch' A' B (S k)
    | B => constr:k
    | _ => constr:0%nat
  end.
Ltac sconjBasicSearch A B := sconjBasicSearch' A B 1%nat.

(* like the above, but search for a vcont that binds x *)
Ltac sconjContSearch' A x k :=
  match A with
    | sconj (x |-> _) _ => constr:k
    | sconj (x |->-) _ => constr:k
    | sconj _ ?A' => sconjContSearch' A' x (S k)
    | x |-> _ => constr:k
    | x |->- => constr:k
    | _ => constr:0%nat
  end.
Ltac sconjContSearch A x := sconjContSearch' A x 1%nat.

Ltac sconjSearch A B := (*sconjBasicSearch A B.*)
  match B with
    | ?x |-> _ => sconjContSearch A x
    | ?x |->- => sconjContSearch A x
    | _ => sconjBasicSearch A B
  end.

(* Phase 1 *)

(* search H for B, and lift B to the front, if found. *)
Ltac hypSearchLift H B := 
  match type of H with
    | sconj ?A1 ?A2 _ => 
      let k := sconjSearch (sconj A1 A2) B in
        scl k H
    | _ => idtac
  end.

(* right to left, lift any part of A in H to the front of H *)
Ltac searchHypForGoal' H A :=
  match A with
    | sconj ?B ?A' => searchHypForGoal' H A'; hypSearchLift H B
    | top => idtac
    | ?B => hypSearchLift H B
  end.

(* right to left, lift any part of the goal in H to the front of H *)
Ltac searchHypForGoal H :=
  match goal with
    | |- sconj ?A ?B _ => searchHypForGoal' H (sconj A B)
    | |- ?A _ => searchHypForGoal' H A
    | |- _ => idtac
        (* dealing with coq pattern matching not recognizing currying *)
  end.


(* Phase 2 *)

(* search goal for B, and lift B to the front, if found. *)
Ltac goalSearchLift B := 
  match goal with
    | |- sconj ?A1 ?A2 _ => 
      let k := sconjSearch (sconj A1 A2) B in
        gscl k
    | |- _ => idtac
  end.

(* right-to-left, lift any part of A in goal to front of goal *)
Ltac searchGoalForHyp' A :=
  match A with
    | sconj ?B ?A' => searchGoalForHyp' A'; goalSearchLift B
    | top => idtac
    | ?B => goalSearchLift B
  end.

(* right-to-left, lift any part of type of H in goal to front of goal *)
Ltac searchGoalForHyp H :=
  match type of H with
    | sconj ?A ?B _ => searchGoalForHyp' (sconj A B)
    | ?A _ => searchGoalForHyp' A
    | _ => idtac  (* dealing with coq pattern matching not 
                     recognizing currying *)
  end.


(* match common parts of hypothesis and goal, even if out of order *)
Ltac oldSearchMatch :=
  match goal with
    | H : sconj _ _ ?m |- sconj _ _ ?m =>
      try (progress (searchHypForGoal H; smatch; searchGoalForHyp H; smatch); 
        oldSearchMatch); try simple apply topI
    | _ => smatch
  end.


Inductive matchElem : Type :=
  | contEl : nat -> val -> matchElem
  | varEl : nat -> matchElem
  | baseEl : mpred -> matchElem.

Inductive matchList : Type :=
  | empMl : matchList
  | topMl : matchList
  | consMl : matchElem -> matchList -> matchList.

Fixpoint envFind (A:Type) (g:list A) (def:A) (k:nat) {struct k} : A :=
  match g, k with
    | x::_, 0%nat => x
    | _::g', S k' => envFind g' def k'
    | _, _ => def
  end.

Definition unElem (g:list mpred) (gv:list val) (e:matchElem) : mpred :=
  match e with
    | contEl x v => envFind gv Vundef x |-> v
    | varEl x => envFind g emp x
    | baseEl A => A
  end.

Fixpoint unMatchList' (g:list mpred) (gv:list val) (m:matchList) {struct m}
  : mpred :=
  match m with
    | empMl => emp
    | topMl => top
    | consMl e m' => unElem g gv e ** unMatchList' g gv m'
  end.

Fixpoint unMatchList (g:list mpred) (gv:list val) (m:matchList) {struct m} 
  : mpred :=
  match m with
    | empMl => emp
    | topMl => top
    | consMl e empMl => unElem g gv e
    | consMl e m' => unElem g gv e ** unMatchList g gv m'
  end.

Ltac listFind A l k :=
  match eval hnf in l with
    | A :: _ => k
    | _ :: ?l' => listFind A l' (S k)
    | nil => false
  end.

Ltac toMatchList l A :=
  match A with
    | emp => constr:(empMl, l)
    | top => constr:(topMl, l)
    | sconj ?B ?C =>
      let x := listFind B l 0%nat in
        match x with 
          | false => 
            let r := toMatchList (l ++ B :: nil) C in
              match r with
                | (?C', ?l') => constr:(consMl (varEl (length l)) C', l')
              end
          | ?k => 
            let r := toMatchList l C in
              match r with
                | (?C', ?l') => constr:(consMl (varEl k) C', l')
              end
        end
    | ?B =>
      let x := listFind B l 0%nat in
        match x with 
          | false => constr:(consMl (varEl (length l)) empMl, l ++ B ::nil)
          | ?k => constr:(consMl (varEl k) empMl, l)
        end
  end.


Fixpoint sconjVarSub (k:nat) (ml:matchList) {struct ml} :=
  match ml with
    | empMl => None
    | topMl => None
    | consMl (varEl k') ml' =>
      if beq_nat k k' then
        Some ml'
      else
        match sconjVarSub k ml' with
          | Some ml'' => Some (consMl (varEl k') ml'')
          | None => None
        end
    | consMl e ml' => 
      match sconjVarSub k ml' with
        | Some ml'' => Some (consMl e ml'')
        | None => None
      end
  end.


Fixpoint contVarSub (k:nat) (v:val) (ml:matchList) {struct ml} :=
  match ml with
    | empMl => None
    | topMl => None
    | consMl (contEl k' v') ml' =>
      if beq_nat k k' then
        Some (consMl (baseEl (!(v = v'))) ml')
      else
        match contVarSub k v ml' with
          | Some ml'' => Some (consMl (contEl k' v') ml'')
          | None => None
        end
    | consMl e ml' =>
      match contVarSub k v ml' with
        | Some ml'' => Some (consMl e ml'')
        | None => None
      end
  end.


Definition elSub (e:matchElem) (m:matchList) : option matchList :=
  match e with
    | contEl k v => contVarSub k v m
    | varEl k => sconjVarSub k m
    | baseEl _ => None
  end.


Fixpoint mlSub (l1 l2:matchList) {struct l1} : prod matchList matchList :=
  match l1 with
    | (empMl | topMl) => (l1, l2)
    | consMl e l1' =>
      match elSub e l2 with
        | Some l2' => mlSub l1' l2'
        | None =>
          let (l1'', l2'') := mlSub l1' l2 in
            (consMl e l1'', l2'')
      end
  end.




Lemma sconjVarSubOk : forall k l,
  match sconjVarSub k l with
    | Some l' => 
      (forall g g', mimpl (unMatchList' g g' (consMl (varEl k) l'))
        (unMatchList' g g' l))
    | None => True
  end.
Proof.
  induction l as [| | e l]; simpl; auto.
  destruct e.

  destruct (sconjVarSub k l) as [l' | ].
  2:trivial.
  intros g g'; rewrite <- IHl; clear IHl; intros m Hp; simpl in * ; gscl2; auto.

  compare k n; intro NE.
  subst n; rewrite beq_true; simpl in * ; intros g g' h Hp; auto.
  rewrite (beq_false NE); clear NE.
  destruct (sconjVarSub k l) as [l' | ].
  intros g g' m Hp; srewrite <- IHl; clear IHl; simpl in * ; gscl2; auto.

  trivial.

  destruct (sconjVarSub k l) as [l' | ].
  rename m into A; simpl in * ; intros g g' m Hp; srewrite <- IHl; clear IHl.
  gscl2; auto.
  
  trivial.

Qed.


Lemma elSubOk : forall e l,
  match elSub e l with
    | Some l' => 
      (forall g g', mimpl (unMatchList' g g' (consMl e l'))
        (unMatchList' g g' l))
    | None => True
  end.
Proof.
  induction l; destruct e; simpl; auto.

  simpl in IHl.
  destruct (contVarSub n v l) as [l' | ].
  destruct m.
  compare n n0; intros NE.

  subst n0; rewrite beq_true; simpl in * ; intros g g' m Hp.
  sle Hp; subst; auto.
  rewrite (beq_false NE); simpl in * ; intros g g' m Hp.
  srewrite <- IHl; gscl2; auto.

  simpl in * ; intros g g' m Hp; srewrite <- IHl; gscl2; auto.

  simpl in * ; intros g g' x Hp; srewrite <- IHl; gscl2; auto.
  
  clear IHl; destruct m.
  compare n n0; intros NE.

  subst n0; rewrite beq_true.
  simpl; intros g g' m Hp; sle Hp; subst; auto.
  rewrite (beq_false NE); auto.
  auto.
  auto.

  simpl in IHl.
  destruct m; simpl.
  destruct (sconjVarSub n l).
  simpl in * ; intros g g' x Hp.
  srewrite <- IHl; gscl2; auto.
  trivial.
  
  compare n n0; intros NE.
  subst n0; rewrite beq_true.
  intros g g' m Hp; auto.
  rewrite (beq_false NE).
  destruct (sconjVarSub n l).
  simpl in * ; intros g g' x Hp.
  srewrite <- IHl; gscl2; auto.
  trivial.
  destruct (sconjVarSub n l).
  simpl in * ; intros g g' m' Hp; srewrite <- IHl; gscl2; auto.
  trivial.

Qed.


Lemma unMatchListEqv : forall l g g',
  miff (unMatchList g g' l) (unMatchList' g g' l).
Proof.
  induction l as [| | e l]; simpl; intros g g' m; constructor; intros Hp; auto.
  destruct l; srewrite <- IHl; simpl; sli. 
  destruct l; simpl in * ; auto.
  sle Hp; auto.
  srewrite <- IHl in Hp; auto.
Qed.


Lemma mlSubOk' : forall l1 g g' A m,
  (unMatchList' g g' l1 ** A) m -> forall l2,
  (let (l1', l2') := mlSub l1 l2 in
    (mimpl (unMatchList' g g' l1' ** A) (unMatchList' g g' l2'))) ->
  (unMatchList' g g' l2) m.
Proof.
  induction l1 as [| | e l]; simpl; intros g g' A m Hp l2 I.
  srewrite <- I; auto.
  srewrite <- I; auto.

  assert (Y := elSubOk e l2); simpl in Y.
  destruct (elSub e l2) as [l2' | ].

  (* case: found the variable *)
  srewrite <- Y; clear Y; smatch0.
  sle Hp; smatch0.

  (* case: didn't find the variable *)
  clear Y; sle Hp.
  hyr IHl (fun x y z => IHl g g' x y z l2).
  destruct (mlSub l l2) as [l1' l2']; simpl in * .
  apply (IHl (unElem g g' e ** A)); clear IHl.
  gscl2; auto.
  rewrite <- I; clear I; clear m Hp; intros m Hp.
  sli; gscl2; auto.
Qed.


Lemma mlSubOk : forall g g' l1 l2 m,
  (unMatchList g g' l1) m ->
  (let (l1', l2') := mlSub l1 l2 in
    (mimpl (unMatchList g g' l1') (unMatchList g g' l2'))) ->
  (unMatchList g g' l2) m.
Proof.
  intros g g' l1 l2 m Hp I.
  srewrite unMatchListEqv.
  apply mlSubOk' with l1 emp.
  sli; srewrite <- unMatchListEqv; auto.
  destruct (mlSub l1 l2).
  clear m Hp; intros m Hp; sle Hp.
  srewrite <- unMatchListEqv; srewrite <- I.
  srewrite unMatchListEqv; auto.
Qed.


(* This doesn't seem to be faster or produce smaller proof terms than 
   the regular search match, and the regular search match is nice enough
   that I think I won't use the new one. *)
Ltac rsearchMatch :=
  match goal with
    | Hp : sconj ?A ?B ?m |- sconj ?A' ?B' ?m =>
      let res := toMatchList (nil (A := mpred)) (sconj A B) in
        match res with
        | (?lh, ?g) => 
          let res' := toMatchList g (sconj A' B') in
            match res' with
              | (?lg, ?g') =>
                let X := fresh in
                let Y := fresh in
                  assert (X := mlSubOk g');
                  cbv beta delta [app] iota in X;
                  assert (Y := X lh lg); clear X;
                  cbv beta delta [app mlSub sconjVarSub envFind
                    unMatchList beq_nat length] iota in Y;
                  simple apply (Y _ Hp); clear Y;
                  let m' := fresh "m" in
                  let Hp' := fresh "Hp" in
                  clear Hp; intros m' Hp'; rename Hp' into Hp;
                  try (clear m; rename m' into m);
                  try simple apply topI; eauto
                end
          end
    | |- _ => smatch; auto
  end.


Ltac searchMatch := oldSearchMatch.




(* Solve goals when Hp is of the form x |-> _ ** x |-> _ ** ... 
   No matter where in Hp the conts occur
 *)

Ltac contContSolver' Hp k A :=
  match A with
    | (?x |-> _ ** ?A') =>
      let k' := sconjContSearch A' x in
        match eval compute in k' with
          | 0%nat => fail
          | _ => scl k Hp; scl (k + k')%nat Hp; elf; 
            first [simple apply (vcontDisjL Hp) | simple apply (vcontDisj Hp)]
        end
    | (_ ** ?A') => contContSolver' Hp (S k) A'
  end.

Ltac contContSolver Hp :=
  match type of Hp with
    | (?A ** ?B) _ => contContSolver' Hp 1%nat (A ** B)
  end.

