(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import SepCore.
Require Import SepSimpl.
Require Import Setoid.


(* using setoids to do rewriting in separation logic predicates *)


(* mpred equivalence is a relation *)

Lemma miffRefl : reflexive _ miff.
Proof.
  intros A m; tauto.
Qed.

Lemma miffSymm : symmetric _ miff.
Proof.
  intros A B E m; hyr E (E m); tauto.
Qed.

Lemma miffTrans : transitive _ miff.
Proof.
  intros A B C E1 E2 m; hyr E1 (E1 m); hyr E2 (E2 m); tauto.
Qed.


Add Relation mpred miff
  reflexivity proved by miffRefl
  symmetry proved by miffSymm
  transitivity proved by miffTrans
  as miffRel.


(* mpred implication is a non-symmetric relation *)

Lemma mimplRefl : reflexive _ mimpl.
Proof.
  intros A m; tauto.
Qed.

Lemma mimplTrans : transitive _ mimpl.
Proof.
  intros A B C E1 E2 m; hyr E1 (E1 m); hyr E2 (E2 m); tauto.
Qed.

Add Relation mpred mimpl
  reflexivity proved by mimplRefl
  transitivity proved by mimplTrans
  as mimplRel.


(* equivalence and implication for fas and lex bodies. 
   I'm not sure how useful it is. *)

Definition amiff (T:Type) (A B:T -> mpred) := forall x m, A x m <-> B x m.

Lemma amiffRefl : forall T, reflexive _ (amiff T).
Proof.
  intros T A m; tauto.
Qed.

Lemma amiffSymm : forall T, symmetric _ (amiff T).
Proof.
  intros T A B E x m; hyr E (E x m); tauto.
Qed.

Lemma amiffTrans : forall T, transitive _ (amiff T).
Proof.
  intros T A B C E1 E2 x m; hyr E1 (E1 x m); hyr E2 (E2 x m); tauto.
Qed.

Add Parametric Relation (T:Type) : (T -> mpred) (amiff T)
  reflexivity proved by (amiffRefl T)
  symmetry proved by (amiffSymm T)
  transitivity proved by (amiffTrans T)
  as amiffRel.


Definition amimpl (T:Type) (A B:T -> mpred) := forall x m, A x m -> B x m.

Lemma amimplRefl : forall T, reflexive _ (amimpl T).
Proof.
  intros T A x m; tauto.
Qed.

Lemma amimplTrans : forall T, transitive _ (amimpl T).
Proof.
  intros T A B C E1 E2 x m; hyr E1 (E1 x m); hyr E2 (E2 x m); tauto.
Qed.

Add Parametric Relation (T:Type) : (T -> mpred) (amimpl T)
  reflexivity proved by (amimplRefl T)
  transitivity proved by (amimplTrans T)
  as amimplRel.


(* mpOk makes things in a nice form so that setoids can use them *)
Definition mpOk : mpred -> CM.mem -> Prop := fun A m => A m.

Add Morphism mpOk
  with signature mimpl ++> @eq _ ==> impl
  as mpOkIffMorph.
Proof.
  unfold mimpl, mpOk, impl; auto.
Qed.

Add Morphism mpOk
  with signature miff ==> @eq _ ==> iff
  as mpOkImpMorph.
Proof.
  unfold mpOk; intros A A' EA m; hyr EA (EA m); tauto.
Qed.


(* morphisms for the connectives *)

Add Morphism sconj
  with signature miff ==> miff ==> miff
  as sconjIffMorph.
Proof.
  intros A A' EA B B' EB m; constructor; intro M;
    apply (sconjImpl M); intros; hyr EA (EA h'); hyr EB (EB h'); tauto.
Qed.

Add Morphism sconj
  with signature mimpl ++> mimpl ++> mimpl
  as sconjImpMorph.
Proof.
  intros A A' EA B B' EB m M;
    apply (sconjImpl M); intros; hyr EA (EA h'); hyr EB (EB h'); tauto.
Qed.

Add Morphism bang
  with signature iff ==> miff
  as bangIffMorph.
Proof.
  intros A B E m; constructor; intro M; slei M; auto.
  rewrite <- E; auto.
  rewrite E; auto.
Qed.

Add Morphism bang
  with signature impl ++> mimpl
  as bangImpMorph.
Proof.
  intros A B E m M; slei M; auto.
Qed.


Add Parametric Morphism T : (@lex T)
  with signature (amiff T) ==> miff
  as lexMorph.
Proof.
  unfold amiff; intros A B E m; constructor; intro M;
    sle M; hyr E (E x m); sli; instantiate (1 := x); tauto.
Qed.

Add Parametric Morphism T : (@lex T)
  with signature (amimpl T) ++> mimpl
  as lexImpMorph.
Proof.
  unfold amiff; intros A B E m M;
    sle M; hyr E (E x m); sli; instantiate (1 := x); tauto.
Qed.

Add Morphism fas
  with signature @amiff _ ==> @eq _ ==> miff
  as fasMorph.
Proof.
  intros P P' E s m; constructor; intro M; apply (fasImpl M); intros.
  rewrite <- (E x); auto.
  rewrite (E x); auto.
Qed.

Add Morphism fas
  with signature @amimpl _ ++> @eq _ ==> mimpl
  as fasImplMorph.
Proof.
  intros P P' E s m M; apply (fasImpl M); auto.
Qed.


Add Morphism wand
  with signature miff ==> miff ==> miff
  as wandIffMorph.
Proof.
  intros A A' Aiff B B' Biff m.
  constructor; intros Hp.

  intros m' D Hp'; hyr Hp (Hp m' D).
  assert (mpOk B' (union m' m)). 2:auto.
  rewrite <- Biff.
  apply Hp.
  assert (mpOk A' m') by auto.
  rewrite <- Aiff in H; auto.

  intros m' D Hp'; hyr Hp (Hp m' D).
  assert (mpOk B (union m' m)). 2:auto.
  rewrite Biff.
  apply Hp.
  assert (mpOk A m') by auto.
  rewrite Aiff in H; auto.

Qed.


Add Morphism wand
  with signature mimpl --> mimpl ++> mimpl
  as wandImpMorph.
Proof.
  intros A A' AI B B' BI m Hp.
  intros m' D Hp'; apply BI.
  apply Hp; auto.
Qed.


(* To use setoid rewriting, the type being rewritten must be in the
   form mpOk A m, but it is annoying to use this everywhere, so
   instead we will define new rewriting tactics that translate to and
   from this. *)

Lemma mpOkEq : forall A m,
  A m = mpOk A m.
Proof.
  auto.
Qed.


Add Morphism mimpl
  with signature miff ==> miff ==> iff
  as miffMimplMorph.
Proof.
  intros A A' IA B B' IB.
  constructor; intro II; intros m Hp.
  rewrite mpOkEq; rewrite <- IB; unfold mpOk.
  apply II.
  rewrite mpOkEq; rewrite IA; unfold mpOk; auto.
  rewrite mpOkEq; rewrite IB; unfold mpOk.
  apply II.
  rewrite mpOkEq; rewrite <- IA; unfold mpOk; auto.
Qed.


(* have to try for second rewrite as T can generate multiple subgoals,
   and we only will be able to rewrite for one of them *)

Ltac mpUnMp T :=
  rewrite mpOkEq; T; try (rewrite <- mpOkEq).

Ltac mpUnMpIn H T :=
  rewrite mpOkEq in H; T; try (rewrite <- mpOkEq in H).

Tactic Notation "srewrite" constr(H) := mpUnMp ltac : (rewrite H).
Tactic Notation "srewrite" "->" constr(H) := srewrite H.
Tactic Notation "srewrite" "<-" constr(H) := mpUnMp ltac : (rewrite <- H).
Tactic Notation "srewrite" constr(E) "in" hyp(H) := mpUnMpIn H ltac : (rewrite E in H).
Tactic Notation "srewrite" "->" constr(E) "in" hyp(H) := srewrite E in H.
Tactic Notation "srewrite" "<-" constr(E) "in" hyp(H) := mpUnMpIn H ltac : (rewrite <- E in H).


(* autorewriting for mpred equivalence 

Databases:

There are two databases uses, SLogI and SLogE.

The rules in SLogI should break things apart going left to right,
while the rules in SLogE should break things apart going right to
left.

*)


Tactic Notation "sarewriteE" "in" constr(H) :=
  mpUnMpIn H ltac : (autorewrite with SLogI in H) ; sle H.
Tactic Notation "sarewriteI" "in" constr(H) :=
  mpUnMpIn H ltac : (autorewrite with SLogE in H).

Tactic Notation "sarewriteI" :=
  mpUnMp ltac : (autorewrite with SLogI) ; sli.
Tactic Notation "sarewriteE" :=
  mpUnMp ltac : (autorewrite with SLogE).