(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Import SepCore.
Require Import List.


Set Implicit Arguments.


(* Simplification of separation logic hypotheses and goals *)



(* support lemmas *)

Lemma sconjLift2Ok : forall A B C h,
  (A ** B ** C) h ->
    (B ** A ** C) h.
Proof.
  intros A B C m Hp.
  apply sconjComm; apply sconjAssocR.
  apply (sconjImpl Hp); auto; clear m Hp; intros m Hp; apply sconjComm; auto.
Qed.


Lemma sconjI : forall (A B:mpred) m m1 m2,
  A m1 ->
  B m2 ->
  dom_disj m1 m2 ->
    m = union m1 m2 ->
    (A ** B) m.
Proof.
  intros A B m m1 m2 HA HB D E; subst m; constructor; auto.
Qed.
Implicit Arguments sconjI [A B m m1 m2].


Lemma sconjE : forall A B m,
  (A ** B) m ->
  exists m1, exists m2,
    dom_disj m1 m2 /\
    m = union m1 m2 /\
    A m1 /\
    B m2.
Proof.
  intros A B m [m1 m2 J HA HB]; exists m1; exists m2; intuition.
Qed.


Lemma sandI : forall (A B:mpred) m,
  A m ->
  B m ->
  (A ^^ B) m.
Proof.
  intros; constructor; auto.
Qed.
Implicit Arguments sandI [A B m].


Lemma sandE : forall A B m,
  (A ^^ B) m ->
  A m /\ B m.
Proof.
  intros A B m Hp; destruct Hp; auto.
Qed.


(* Support for "legacy" tactics *)


Lemma fasEmptyIL : forall (A:mpred) m,
  A m -> forall P,
    (fas P WS.empty ** A) m.
Proof.
  intros A m Hp P.
  hyr Hp (empIL Hp); apply (sconjImpl Hp); auto.
  intros m' Hp'; apply fasEmptyI; auto.
Qed.
Implicit Arguments fasEmptyIL [A m].


Lemma fasEmptyIR : forall (A:mpred) m,
  A m -> forall P,
    (A ** fas P WS.empty) m.
Proof.
  intros A m Hp P; apply sconjComm; apply (fasEmptyIL Hp).
Qed.
Implicit Arguments fasEmptyIR [A m].


Lemma fasEmptyEL : forall P (A:mpred) m,
  (fas P WS.empty ** A) m ->
  A m.
Proof.
  intros P A m Hp.
  apply empEL; apply (sconjImpl Hp); intros; auto; apply (fasEmptyE H).
Qed.


Lemma fasEmptyER : forall P (A:mpred) m,
  (A ** fas P WS.empty) m ->
  A m.
Proof.
  intros P A m Hp; eapply fasEmptyEL; apply sconjComm; eauto.
Qed.


Lemma topTopI : forall m,
  top m ->
  (top ** top) m.
Proof.
  intros m Hp; hyr Hp (empIL Hp); apply (sconjImpl Hp); unfold top; auto.
Qed.


Lemma topTopIL : forall A m,
  (top ** A) m ->
  (top ** top ** A) m.
Proof.
  intros A m Hp; hyr Hp (empIL Hp); apply (sconjImpl Hp); unfold top; auto.
Qed.


Lemma topTopEL : forall A m,
  (top ** top ** A) m ->
  (top ** A) m.
Proof.
  intros A m Hp; apply sconjAssocR in Hp; apply (sconjImpl Hp);
    unfold top; auto.
Qed.


Lemma stackFrameZeroIL : forall (A:mpred) m,
  A m -> forall sp,
  (stackFrame sp 0%word ** A) m.
Proof.
  unfold stackFrame; intros A m Hp sp; setsimpl; apply fasEmptyIL; auto.
Qed.


Lemma stackFrameZeroIR : forall (A:mpred) m,
  A m -> forall sp,
  (A ** stackFrame sp 0%word) m.
Proof.
  intros A m Hp sp; apply sconjComm; apply stackFrameZeroIL; auto.
Qed.


Lemma stackFrameZeroEL : forall sp A m,
  (stackFrame sp 0%word ** A) m ->
  A m.
Proof.
  unfold stackFrame; setsimpl; intros sp A m Hp; apply (fasEmptyEL Hp).
Qed.


Lemma stackFrameZeroER : forall A sp m,
  (A ** stackFrame sp 0%word) m ->
  A m.
Proof.
  intros A m sp Hp; eapply stackFrameZeroEL; apply sconjComm; eauto.
Qed.


(* legacy simplification tactics *)

Ltac oldSli' :=
  match goal with
    | |- top _ => fail
    | |- emp _ => assumption
    | |- bang _ _ => apply bangI
    | |- lex _ _ => apply lexI; econstructor; try oldSli'
    | |- sconj (fas _ WS.empty) _ _ => apply fasEmptyIL; try oldSli'
    | |- sconj _ (fas _ WS.empty) _ => apply fasEmptyIR; try oldSli'
    | |- sconj (stackFrame _ 0%word) _ _ => apply stackFrameZeroIL; try oldSli'
    | |- sconj _ (stackFrame _ 0%word) _ => apply stackFrameZeroIR; try oldSli'
    | |- sconj _ emp _ => apply empIR; try oldSli'
    | |- sconj emp _ _ => apply empIL; try oldSli'
    | |- sconj (bang _) _ _ => apply bangIL; [try oldSli' | auto]
    | |- sconj _ (bang _) _ => apply bangIR; [try oldSli' | auto]
    | |- sconj (lex _) _ _ => apply lexEL; try oldSli'
    | |- sconj _ (lex _) _ => apply lexER; try oldSli'
    | |- sconj (sconj _ _) _ _ => apply sconjAssocR; try oldSli'
    | |- sconj top ?A _ => 
      match A with
	| top => apply topTopI; try oldSli'
	| sconj ?B _ =>
	  match B with
	    | top => apply topTopIL; try oldSli'
	    | _ => apply sconjLift2Ok; try oldSli'
	  end
	| _ => apply sconjComm; try oldSli'
      end
    | |- sconj _ _ _ => sconjSubIntro
  end
with sconjSubIntro :=
  let s := fresh "m" in 
  let H := fresh "Hp" in
    first
    [eapply sconjImpl; 
      [ |
	intros s H; oldSli'; try (apply H); try (clear s H) |
	intros s H; try oldSli'; try (apply H); try (clear s H)] |
      eapply sconjImpl;
	[ |
	  intros s H; try (apply H); try (clear s H) |
	  intros s H; oldSli'; try (apply H); try (clear s H)]];
    try (apply empIR); try (apply empIL).

Ltac oldSli := first [oldSli'; try apply topI | apply topI].


Ltac oldSle Hp :=
  let Hp' := fresh "Hp" in
    (match type of Hp with
       | sconj (sconj _ _) _ _ => assert (Hp' := sconjAssocL Hp)
       | sconj emp _ _ => assert (Hp' := empEL Hp)
       | sconj _ emp _ => assert (Hp' := empER Hp)
       | sconj (fas _ WS.empty) _ _ => assert (Hp' := fasEmptyEL Hp)
       | sconj _ (fas _ WS.empty) _ => assert (Hp' := fasEmptyER Hp)
       | sconj (stackFrame _ 0%word) _ _ => assert (Hp' := stackFrameZeroEL Hp)
       | sconj _ (stackFrame _ 0%word) _ => assert (Hp' := stackFrameZeroER Hp)
       | sconj top ?A ?s => 
	 match A with 
	   | top => assert (Hp' := topI s)
	   | sconj ?B _ =>
	     match B with
 	       | top => assert (Hp' := topTopEL Hp)
	       | _ => assert (Hp' := sconjLift2Ok Hp)
	     end
	   | _ => assert (Hp' := sconjComm Hp)
	 end
       | lex _ _ => 
	 let x := fresh "x" in
           destruct (lexE Hp) as [x Hp']
       | bang _ _ =>
	 let HA := fresh in
	   (assert (HA := bangE2 Hp);
	    assert (Hp' := bangE1 Hp))
       | sconj _ _ ?m => sconjSubSimpl Hp Hp' m
     end;
    first [(clear Hp; rename Hp' into Hp; try (oldSle Hp)) |
      try (oldSle Hp')])
with sconjSubSimpl Hp Hp' m :=
  let s1 := fresh "m" in
  let s2 := fresh "m" in
  let J := fresh "J" in
  let E := fresh "E" in
  let Hp1 := fresh "Hp" in
  let Hp2 := fresh "Hp" in
    (destruct (sconjE Hp) as [s1 [s2 [J [E [Hp1 Hp2]]]]];
      ((oldSle Hp1; try (oldSle Hp2)) ||
	(oldSle Hp2));
      assert (Hp' := sconjI Hp1 Hp2 J E);
      clear s1 s2 J E Hp1 Hp2).


(* inductively defined version of separation logic predicate *)
Inductive stree : Type :=
  | empStree : stree
  | faseStree : (word -> mpred) -> stree
  | stackeStree : block -> stree
  | topStree : stree
  | bangStree : Prop -> stree
  | contuStree : val -> stree
  | contzStree : val -> stree
  | baseStree : mpred -> stree
  | scTree : stree -> stree -> stree
  | sandTree : stree -> stree -> stree.


(* convert from an stree back to an mpred *)
Fixpoint unstree (t:stree) : mpred :=
  match t with
    | empStree => emp
    | faseStree P => fas P WS.empty
    | stackeStree b => stackFrame b 0%word
    | topStree => top
    | bangStree A => !A
    | contuStree v => vcont Vundef v
    | contzStree v => vcont 0%word v
    | baseStree A => A
    | scTree t1 t2 => unstree t1 ** unstree t2
    | sandTree t1 t2 => unstree t1 ^^ unstree t2
  end.


(* tactic to convert sep predicate to an stree *)
Ltac toStree A :=
  match A with
    | emp => constr:empStree
    | fas ?P WS.empty => constr:(faseStree P)
    | stackFrame ?b (nat_to_word 0) => constr:(stackeStree b)
    | top => constr:topStree
    | bang ?A => constr:(bangStree A)
    | vcont Vundef ?v => constr:(contuStree v)
    | vcont null_ptr ?v => constr:(contzStree v)
    | ?A ** ?B => 
      let t1 := toStree A in
      let t2 := toStree B in
        constr:(scTree t1 t2)
    | ?A ^^ ?B =>
      let t1 := toStree A in
      let t2 := toStree B in
        constr:(sandTree t1 t2)
    | _ => constr:(baseStree A)
  end.


Inductive stree2 : Type :=
  | empty_st2 : stree2
  | star_st2 : mpred -> stree2 -> stree2
  | and_star_st2 : stree2 -> stree2 -> stree2 -> stree2.


(* return value for streeSimpl':
   stree2 is the predicates
   bool is true iff the predicate contains top at least once
   option Prop
      Some A if predicate contains !A (possibly split up)
      None if predicate contains no bangs
*)
Inductive streeSimplRet : Type :=
  | ssRet : stree2 -> bool -> option Prop -> streeSimplRet.


(* add top to the end of the stree2 *)
Fixpoint addTop (l:stree2) : stree2 :=
  match l with
    | empty_st2 => star_st2 top empty_st2
    | star_st2 A l' => star_st2 A (addTop l')
    | and_star_st2 l1 l2 l' => and_star_st2 l1 l2 (addTop l')
  end.

(* flatten the stree, while performing some basic simplifications *)
(* the list mpred argument is a continuation (?) *)
(* return None if we can prove False from the stree *)
Fixpoint streeSimpl' (t:stree) (l:stree2) : option streeSimplRet :=
  match t with
    | (empStree | faseStree _ | stackeStree _) => Some (ssRet l false None)
    | topStree => Some (ssRet l true None)
    | bangStree P => Some (ssRet l false (Some P))
    | baseStree A => Some (ssRet (star_st2 A l) false None)
    | contuStree v => None
    | contzStree v => None
    | scTree t1 t2 => 
      match streeSimpl' t2 l with
        | Some (ssRet l' b OP) =>
          match streeSimpl' t1 l' with
            | Some (ssRet l'' b' OP') => 
              Some (ssRet l'' (if b then true else b')
                (match OP, OP' with
                   | Some P, Some P' => Some (P /\ P')
                   | Some _, None => OP
                   | None, Some _ => OP'
                   | None, None => None
                 end))
            | None => None
          end
        | None => None
      end
    | sandTree t1 t2 => 
      match streeSimpl' t1 empty_st2 with
        | Some (ssRet l1 b1 OP1) => 
          match streeSimpl' t2 empty_st2 with
            | Some (ssRet l2 b2 OP2) =>
              Some (ssRet (and_star_st2
                (if b1 then addTop l1 else l1) (if b2 then addTop l2 else l2) l)
                false
                (match OP1, OP2 with
                   | Some P1, Some P2 => Some (P1 /\ P2)
                   | Some _, None => OP1
                   | None, Some _ => OP2
                   | None, None => None
                 end))
            | None => None
          end
        | None => None
      end
  end.


Definition streeSimpl t := 
  match streeSimpl' t empty_st2 with
    | Some (ssRet l b OP) => Some (if b then (addTop l, OP) else (l, OP))
    | None => None
  end.


(* convert an stree2 into an mpred *)
Fixpoint unstree2 (t:stree2) : mpred :=
  match t with
    | empty_st2 => emp
    | star_st2 A t' => A ** unstree2 t'
    | and_star_st2 A B t' => (unstree2 A ^^ unstree2 B) ** unstree2 t'
  end.


(* equivalent to unstree2, but produces a nicer result at a cost of 
   being harder to reason about *)
Fixpoint unstree2' (t:stree2) : mpred :=
  match t with
    | empty_st2 => emp
    | star_st2 A empty_st2 => A
    | star_st2 A t' => A ** unstree2' t'
    | and_star_st2 A B empty_st2 => unstree2' A ^^ unstree2' B
    | and_star_st2 A B t' => (unstree2' A ^^ unstree2' B) ** unstree2' t'
  end.


(* unstree2 and unstree2' are equivalent *)

Lemma unstree2E : forall t m,
  unstree2 t m ->
  unstree2' t m.
Proof.
  induction t as [ | A t | t1 t2 t]; simpl; intros m Hp. auto.

  destruct t; smatch0.

  assert (Hp' : ((unstree2' t1 ^^ unstree2' t)** unstree2' t3) m).
    apply (sconjImpl Hp); clear m Hp; intros m Hp. 2:auto.
    destruct Hp; constructor; auto.
  clear IHt1 IHt2 Hp; rename Hp' into Hp.
  destruct t3. exact (empER Hp). auto. auto.

Qed.


Lemma unstree2I : forall t m,
  unstree2' t m ->
  unstree2 t m.
Proof.
  induction t as [ | A t | t1 t2 t]; simpl; intros m Hp. auto.

  destruct t; smatch0.

  destruct t3.
  simpl; apply empIR; destruct Hp; constructor; auto.
  apply (sconjImpl Hp); clear m Hp; intros m Hp.
  destruct Hp; constructor; auto. auto.
  apply (sconjImpl Hp); clear m Hp; intros m Hp.
  destruct Hp; constructor; auto. auto.

Qed.


(* addTop equivalence *)

Lemma addTopI : forall t m,
  (unstree2 t ** top) m ->
  unstree2 (addTop t) m.
Proof.
  induction t; simpl; intros m0 Hp.
  apply sconjComm; auto.
  hyr Hp (sconjAssocL Hp); smatch0.
  hyr Hp (sconjAssocL Hp); smatch0.
Qed.


Lemma addTopE : forall l m,
  unstree2 (addTop l) m ->
  (unstree2 l ** top) m.
Proof.
  induction l; simpl; intros m0 Hp.
  apply sconjComm; auto.
  apply sconjAssocR; smatch0.
  apply sconjAssocR; smatch0.
Qed.


(* stree simplification soundness introduction *)

Lemma streeSimpl'I : forall t l A m,
  (unstree t ** unstree2 l ** A) m ->
  match streeSimpl' t l with
    | Some (ssRet l' b OP) =>
      (unstree2 l' ** (if b then top else emp) ** A) m /\
      match OP with
        | Some P => P
        | None => True
      end
    | None => False
  end.
Proof.
  induction t; simpl; intros l A m0 Hp; try (constructor; [ | trivial; fail]).
  hyr Hp (sconjLift2Ok Hp); auto.
  hyr Hp (sconjLift2Ok Hp); smatch0; exact (fasEmptyE Hp).
  hyr Hp (sconjLift2Ok Hp); smatch0; exact (stackFrameZeroE Hp).
  hyr Hp (sconjLift2Ok Hp); auto.
  destruct (bangEL Hp) as [Hp' Pok]; clear Hp.
  constructor; [smatch0 | auto]; apply empIL; auto.
  destruct Hp; inversion H0.
  destruct Hp; inversion H0.
  oldSli; auto.

  rename t1 into t; rename t2 into t'; rename m0 into m.
  assert (Hp' : (unstree t' ** unstree2 l ** unstree t ** A) m).
  hyr Hp (sconjLift2Ok (sconjAssocL Hp)); smatch0.
  hyr Hp (sconjLift2Ok Hp); auto.
  clear Hp; rename Hp' into Hp.
  
  hyr Hp (IHt2 _ _ _ Hp); clear IHt2.
  destruct (streeSimpl' t' l). 2:autof.
  destruct s; destruct Hp as [Hp X].
  rename s into l0.
  assert (Hp' : (unstree t ** unstree2 l0 ** (if b then top else emp) ** A) m).
    apply sconjComm; apply sconjAssocR; smatch0; apply sconjAssocR; smatch0.
    apply sconjComm; auto.
  clear Hp; rename Hp' into Hp.
  hyr Hp (IHt1 _ _ _ Hp); clear IHt1.
  destruct (streeSimpl' t l0). 2:autof.
  destruct s; destruct Hp as [Hp Y].
  constructor. smatch0. destruct b; destruct b0; try oldSli.
  apply sconjComm.
  hyr Hp (sconjAssocR Hp); apply (sconjImpl Hp); clear m Hp; unfold top; auto.
  apply sconjComm; exact (empEL Hp).
  apply sconjComm; exact (empEL (sconjLift2Ok Hp)).
  exact (empEL (empEL Hp)).
  destruct o; destruct o0; auto.

  rename t1 into t; rename t2 into t'; rename m0 into m.
  destruct Hp as [m1 m2 Disj [Hp1 Hp1'] Hp2].
  hyr IHt1 (IHt1 empty_st2 emp m1 (sconjAssocL (empIR (empIR Hp1)))).
  hyr IHt2 (IHt2 empty_st2 emp m1 (sconjAssocL (empIR (empIR Hp1')))).
  clear Hp1 Hp1'.
  destruct (streeSimpl' t empty_st2). 2:autof. destruct s. 
  destruct (streeSimpl' t' empty_st2). 2:autof. destruct s0.
  destruct IHt1 as [Hp1 H1]; destruct IHt2 as [Hp1' H1'].
  constructor. 
  clear H1 H1'. simpl.
  apply sconjAssocR in Hp1; apply empER in Hp1.
  apply sconjAssocR in Hp1'; apply empER in Hp1'.
  apply sconjAssocR; constructor; auto. constructor.
  destruct b. apply addTopI; auto. apply empER; auto.
  destruct b0. apply addTopI; auto. apply empER; auto.
  clear - Hp2. smatch0. apply empIL; auto.
  clear - H1 H1'. destruct o; destruct o0; auto.

Qed.


Lemma streeSimplI' : forall t m,
  unstree t m ->
  match streeSimpl t with
    | Some (l, OP) => 
      unstree2' l m /\ (match OP with Some P => P | None => True end)
    | None => False
  end.
Proof.
  unfold streeSimpl; intros t m Hp.
  hyr Hp (streeSimpl'I _ empty_st2 (sconjAssocL (empIR (empIR Hp)))).
  destruct (streeSimpl' t empty_st2). 2:autof.
  destruct s; destruct Hp as [Hp H].
  destruct b; constructor; auto.
  apply unstree2E; apply addTopI; smatch0.
  apply unstree2E; exact (empER (empEL (sconjLift2Ok Hp))).
Qed.


Lemma streeSimplI : forall t m,
  unstree t m ->
  match streeSimpl t with
    | Some (l, (Some P)) => unstree2' l m /\ P
    | Some (l, None) => unstree2' l m
    | None => False
  end.
Proof.
  intros t m M; hyr M (streeSimplI' _ _ M).
  destruct (streeSimpl t) as [[l [P | ]] | ]; auto.
  destruct M; auto.
Qed.



(* stree simplification elimination *)

Lemma streeSimpl'E : forall t l A m,
  match streeSimpl' t l with
    | Some (ssRet l' b OP) =>
      (unstree2 l' ** (if b then top else emp) ** A) m /\
      match OP with
        | Some P => P
        | None => True
      end
    | None => False
  end ->
  (unstree t ** unstree2 l ** A) m.
Proof.
  induction t; simpl; intros l A m0 Hp; try (constructor; [ | trivial; fail]);
    try (destruct Hp as [Hp H]).
  hyr Hp (sconjLift2Ok Hp); auto.
  hyr Hp (sconjLift2Ok Hp); smatch0; apply fasEmptyI; auto.
  hyr Hp (sconjLift2Ok Hp); smatch0; exact (stackFrameZeroI _ Hp).
  hyr Hp (sconjLift2Ok Hp); auto.
  apply bangIL; [smatch0 | auto]; apply empEL; auto.
  elf; auto.
  elf; auto.
  hyr Hp (sconjAssocL Hp); smatch0; apply empEL; auto.

  rename t2 into t'; rename t1 into t; rename m0 into m.
  cut ((unstree t' ** unstree2 l ** unstree t ** A) m).
  intro Hp'.
  apply sconjAssocR; apply sconjComm; apply sconjAssocR; smatch0.
  apply sconjAssocR; smatch0; apply sconjComm; smatch0.

  apply IHt2; clear IHt2.
  destruct (streeSimpl' t' l).
  2:auto.
  destruct s.
  rename s into l0.
  hyr IHt1 (IHt1 l0); destruct (streeSimpl' t l0).
  destruct s; destruct Hp as [Hp H].
  constructor.
  cut ((unstree t ** unstree2 l0 ** (if b then top else emp) ** A) m).
  intros Hp'.
  hyr Hp' (sconjComm Hp').
  repeat (hyr Hp' (sconjAssocL Hp'); smatch0); apply sconjComm; auto.
  apply IHt1; clear IHt1.
  constructor.
  smatch0; apply sconjAssocL; apply (sconjImpl Hp); clear m Hp; intros m Hp.
  destruct b; destruct b0.
  smatch0.
  apply empIL; auto.
  apply empIR; auto.
  smatch0.
  auto.
  destruct o; destruct o0; auto.
  destruct H; auto.
  clear IHt1 Hp; destruct o; destruct o0; auto.
  destruct H; auto.
  
  elf; auto.

  rename t2 into t'; rename t1 into t; rename m0 into m.
  hyr IHt2 (IHt2 empty_st2 emp). hyr IHt1 (IHt1 empty_st2 emp).
  destruct (streeSimpl' t empty_st2). destruct s.
  destruct (streeSimpl' t' empty_st2). destruct s0.
  destruct Hp as [Hp HP]. simpl in IHt2; simpl in IHt1.
  simpl in Hp; apply sconjAssocL in Hp; apply (sconjImpl Hp); clear m Hp; intros m Hp.
  destruct Hp as [Hp1 Hp2]; constructor; apply empER; apply empER; apply sconjAssocR.
  apply IHt1; clear Hp2 IHt1 IHt2. constructor. oldSli. destruct b.
  apply addTopE. auto. apply empIR; auto. destruct o. destruct o0. tauto. auto. auto.
  apply IHt2; clear Hp1 IHt1 IHt2. constructor. oldSli. destruct b0.
  apply addTopE. auto. apply empIR; auto. destruct o. destruct o0. tauto. trivial. 
  destruct o0; auto.
  
  smatch0. apply empEL; auto.

  tauto.

  tauto.

Qed.


Lemma streeSimplE' : forall t m,
  match streeSimpl t with
    | Some (l, OP) => 
      unstree2' l m /\ (match OP with Some P => P | None => True end)
    | None => False
  end ->
  unstree t m.
Proof.
  unfold streeSimpl; intros t m Simp.
  assert (Simp' := streeSimpl'E t empty_st2 emp m).
  destruct (streeSimpl' t empty_st2).
  2:elf; auto.
  destruct s.
  destruct b.
  destruct Simp as [Hp H].
  cut ((unstree t ** emp ** emp) m).
  intros; apply empER; smatch0.
  apply Simp'; clear Simp'; constructor; auto.
  apply sconjAssocL; apply empIR; apply addTopE; auto.
  apply unstree2I; auto.

  destruct Simp as [Hp H].
  cut ((unstree t ** emp ** emp) m).
  intros; apply empER; smatch0.
  apply Simp'; clear Simp'; constructor; auto.
  apply sconjAssocL; apply empIR; apply empIR; apply unstree2I; auto.
Qed.


Lemma streeSimplE : forall t m,
  match streeSimpl t with
    | Some (l, (Some P)) => unstree2' l m /\ P
    | Some (l, None) => unstree2' l m
    | None => False
  end ->
  unstree t m.
Proof.
  intros t m X; apply streeSimplE'.
  destruct (streeSimpl t) as [[l [P | ]] | ]; auto.
Qed.


(* the simplification tactics *)


(* simplify a hypothesis *)

Ltac rsle' origH A :=
  let t := toStree A in
  let extraH := fresh in (* first because it might not be cleared later *)
  let Himp := fresh in
  let newH := fresh in
    assert (Himp := streeSimplI t);
    (* reduce t early as possible for speed *)
    cbv beta delta [unstree streeSimpl streeSimpl' addTop unstree2'] iota in Himp;
    assert (newH := Himp _ origH); clear Himp origH; rename newH into origH;
    try (elf; exact origH);
    match type of origH with
      | _ /\ (_ /\ _) => destruct origH as [origH extraH];
        decompose [and] extraH; clear extraH
      | _ /\ _ => destruct origH as [origH extraH]
      | _ => idtac
    end.

Ltac rsle origH :=
  match type of origH with
    | ?A ?B ?C ?D ?E ?F _ => rsle' origH (A B C D E F)
    | ?A ?B ?C ?D ?E _ => rsle' origH (A B C D E)
    | ?A ?B ?C ?D _ => rsle' origH (A B C D)
    | ?A ?B ?C _ => rsle' origH (A B C)
    | ?A ?B _ => rsle' origH (A B)
    | ?A _ => rsle' origH A
  end.


(* simplify the goal *)

Ltac conjSplitter :=
  match goal with
    | |- _ /\ _ => constructor; conjSplitter
    | |- _ => auto
  end.

(* we have to pass in m to work around a weird unification error *)
Ltac rsli' A m :=
  let t := toStree A in
  assert (Himp := streeSimplE t);
  cbv beta delta [unstree streeSimpl streeSimpl' addTop unstree2'] iota in Himp;
  apply (Himp m); clear Himp;
  match goal with 
    | |- _ /\ _ => constructor; 
      [(try apply topI); try assumption | conjSplitter]
    | |- _ => (try apply topI); try assumption
  end.

Ltac rsli :=
  match goal with
    | |- ?A ?B ?C ?D ?E ?F ?m => rsli' (A B C D E F) m
    | |- ?A ?B ?C ?D ?E ?m => rsli' (A B C D E) m
    | |- ?A ?B ?C ?D ?m => rsli' (A B C D) m
    | |- ?A ?B ?C ?m => rsli' (A B C) m
    | |- ?A ?B ?m => rsli' (A B) m
    | |- ?A ?m => rsli' A m
  end.

(* same as above, but conjSplitter doesn't try to solve anything *)
Ltac conjSplitterD :=
  match goal with
    | |- _ /\ _ => constructor; conjSplitterD
    | |- _ => idtac
  end.

(* we have to pass in m to work around a weird unification error *)
Ltac rsliD' A m :=
  let t := toStree A in
  assert (Himp := streeSimplE t);
  cbv beta delta [unstree streeSimpl streeSimpl' addTop unstree2'] iota in Himp;
  apply (Himp m); clear Himp;
  match goal with 
    | |- _ /\ _ => constructor; 
      [(try apply topI); try assumption | conjSplitterD]
    | |- _ => (try apply topI); try assumption
  end.

Ltac rsliD :=
  match goal with
    | |- ?A ?B ?C ?D ?E ?F ?m => rsliD' (A B C D E F) m
    | |- ?A ?B ?C ?D ?E ?m => rsliD' (A B C D E) m
    | |- ?A ?B ?C ?D ?m => rsliD' (A B C D) m
    | |- ?A ?B ?C ?m => rsliD' (A B C) m
    | |- ?A ?B ?m => rsliD' (A B) m
    | |- ?A ?m => rsliD' A m
  end.



(* check to see if a type contains lex *)
Ltac typeHasLex' A :=
  match A with
    | lex _ => true
    | ?B ** ?C =>
      let t := typeHasLex' B in
      match t with
        | true => true
        | false => let t' := typeHasLex' C in t'
      end
(*    | ?B ^^ ?C =>
      let t := typeHasLex' B in
      match t with
        | true => true
        | false => let t' := typeHasLex' C in t'
      end*)
    | _ => false
  end.

Ltac typeHasLex origH :=
  match type of origH with
    | ?A ?B ?C ?D ?E ?F _ => typeHasLex' (A B C D E F)
    | ?A ?B ?C ?D ?E _ => typeHasLex' (A B C D E)
    | ?A ?B ?C ?D _ => typeHasLex' (A B C D)
    | ?A ?B ?C _ => typeHasLex' (A B C)
    | ?A ?B _ => typeHasLex' (A B)
    | ?A _ => typeHasLex' A
  end.

Ltac goalHasLex :=
  match goal with
    | |- ?A ?B ?C ?D ?E ?F _ => typeHasLex' (A B C D E F)
    | |- ?A ?B ?C ?D ?E _ => typeHasLex' (A B C D E)
    | |- ?A ?B ?C ?D _ => typeHasLex' (A B C D)
    | |- ?A ?B ?C _ => typeHasLex' (A B C)
    | |- ?A ?B _ => typeHasLex' (A B)
    | |- ?A _ => typeHasLex' A
  end.


(* remove all lexists from a separation logic hypothesis *)
Ltac lexRemoveHyp Hp :=
  let Hp' := fresh "Hp" in
    match type of Hp with
      | lex _ _ => 
	let x := fresh "x" in
          destruct (lexE Hp) as [x Hp']; try lexRemoveHyp Hp'
      | (_ ** _) _ => 
        let s1 := fresh "m" in
        let s2 := fresh "m" in
        let J := fresh "J" in
        let E := fresh "E" in
        let Hp1 := fresh "Hp" in
        let Hp2 := fresh "Hp" in
          destruct (sconjE Hp) as [s1 [s2 [J [E [Hp1 Hp2]]]]];
          ((lexRemoveHyp Hp1; try (lexRemoveHyp Hp2)) || lexRemoveHyp Hp2);
          assert (Hp' := sconjI Hp1 Hp2 J E);
          clear s1 s2 J E Hp1 Hp2
(*      | (_ ^^ _) _ => 
        let Hp1 := fresh "Hp" in
        let Hp2 := fresh "Hp" in
          destruct (sandE Hp) as [Hp1 Hp2];
          ((lexRemoveHyp Hp1; try (lexRemoveHyp Hp2)) || lexRemoveHyp Hp2);
          assert (Hp' := sandI Hp1 Hp2);
          clear Hp1 Hp2*)
    end;
    clear Hp; rename Hp' into Hp.

(* do a lex remove only if there is something to remove *)
Ltac tryLexRemoveHyp Hp :=
  let hasL := typeHasLex Hp in
    match hasL with
      | true => lexRemoveHyp Hp
      | false => idtac
    end.


Ltac lexRemoveGoal :=
  match goal with
    | |- lex _ _ => apply lexI; econstructor; try lexRemoveGoal
    | |- sconj _ _ _ =>
      let m := fresh "m" in
      let H := fresh "Hp" in
        first
          [eapply sconjImpl; 
            [ | intros m H; lexRemoveGoal; apply H |
	        intros m H; try lexRemoveGoal; apply H ] |
           eapply sconjImpl;
	    [ | intros m H; apply H |
	        intros m H; lexRemoveGoal; apply H]]
    (* also need case for ^^, if i add that back in *)
  end.

Ltac tryLexRemoveGoal :=
  let hasL := goalHasLex in
    match hasL with
      | true => lexRemoveGoal
      | false => idtac
    end.


Ltac sle Hp := tryLexRemoveHyp Hp; rsle Hp.
(*
On my computer, SepRearr takes:
  1.6 sec with rsle Hp
  1.6 sec with tryLexRemoveHyp Hp
  2.4 sec with try (lexRemoveHyp Hp); rsle Hp
  8.2 sec with oldSle
*)


Ltac sli := tryLexRemoveGoal; rsli. 
Ltac slid := tryLexRemoveGoal; rsliD.
Ltac slei Hp := sle Hp; sli.

Ltac ssimpl := sli.
Ltac ssimpld := slid.
Tactic Notation "ssimpl" "in" hyp(H) := sle H.
Tactic Notation "ssimplg" "in" hyp(H) := slei H.

(*
SepRearr takes: 
  ~1.45 sec with rsli  (no noticeable change for using tryLexRemoveGoal
  1.5-6 sec with oldSli
*)


