(******************************************************************

   Separation logic toolkit

   Andrew McCreight
   HASP Project
   Portland State University

   http://hasp.cs.pdx.edu/

   Copyright 2009 Portland State University

   This file is part of the separation logic toolkit.

   The separation logic toolkit is free software: you can redistribute
   it and/or modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation, either
   version 3 of the License, or (at your option) any later version.

   The separation logic toolkit is distributed in the hope that it
   will be useful, but WITHOUT ANY WARRANTY; without even the implied
   warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with the separation logic toolkit.  If not, see
   <http://www.gnu.org/licenses/>.

*******************************************************************)


Require Export Word.
Require Export Memory.
Require Export Bool.
Require Import Setoid.
Require Import Addr.


Set Implicit Arguments.


Coercion Vptr : addr >-> val.


Inductive bop : Set :=
  | plusop | subop
  | multop
  | ltop | gtop | lteop | gteop
  | eqop | neqop.


Inductive blop : Set := andop | orop.

Definition null_ptr : val := Vword 0%word.


Inductive trueVal : val -> Prop :=
  | wordTrueVal : forall w,
    w <> 0%word ->
    trueVal (Vword w)
  | ptrTrueVal : forall a,
    trueVal (Vptr a).


Lemma valBoolCases : forall v,
  {trueVal v} + {v = Vword 0} + {v = Vundef}.
Proof.
  destruct v.
  right; auto.
  destruct (eq_word_dec w 0).
  subst w; left; right; reflexivity.
  left; left; constructor; assumption.
  left; left; constructor.
Qed.


Definition valNot (v:val) : val :=
  match v with
    | Vword w => Vword (bool_to_word (negb (word_to_bool w)))
    | _ => Vundef
  end.

(* Even this isn't very good because it requires that both expressions
   successfully evaluate.
   I guess now I see why this isn't a regular binop in real CMinor... *)
Definition valAnd (v v':val) : val :=
  match valBoolCases v with
    | inleft (left _) =>
      match valBoolCases v' with
        | inleft (left _) => Vword true
        | inleft (right _) => Vword false
        | inright _ => Vundef
      end
    | inleft (right _) => Vword false
    | inright _ => Vundef
  end.

Definition valOr (v v':val) : val :=
  match valBoolCases v with
    | inleft (left _) => Vword true
    | inleft (right _) =>
      match valBoolCases v' with
        | inleft (left _) => Vword true
        | inleft (right _) => Vword false
        | inright _ => Vundef
      end
    | inright _ => Vundef
  end.


Definition bopWordVal (b:bop) (w1 w2:word) : word :=
  match b with
    | plusop => (w1 + w2)%word
    | subop => (w1 - w2)%word
    | multop => (w1 * w2)%word
    | ltop => if lt_word_dec w1 w2 then true else false
    | gtop => if lt_word_dec w2 w1 then true else false
    | lteop => if lte_word_dec w1 w2 then true else false
    | gteop => if lte_word_dec w2 w1 then true else false
    | eqop => if eq_word_dec w1 w2 then true else false
    | neqop => if eq_word_dec w1 w2 then false else true
  end.

(*
    | andop => andb (word_to_bool w1) (word_to_bool w2)
    | orop => orb (word_to_bool w1) (word_to_bool w2)
  end.
*)


Definition eqVal (v v':val) : val :=
  match v, v' with
    | Vword w, Vword w' => Vword (bopWordVal eqop w w')
    | Vptr (b, k), Vptr (b', k') =>
      Vword (bool_to_word (if eq_block_dec b b' then 
                         if eq_word_dec k k' then true else false
                         else false))
      (* I'm pretty sure this case is wrong, at least for the equality op.
         If the base addresses are unequal, equality testing should fail, not
         produce false. *)
    | Vptr (b1, k1), Vword k2 | Vword k2, Vptr (b1, k1) =>
      if eq_word_dec 0 k2 then
        Vword 0
      else
        Vundef
    | _, _ => Vundef
  end.


Definition bopVal (b:bop) (v1 v2:val) : val :=
  match b, v1, v2 with
(*    | andop, _, _ => valAnd v1 v2
    | orop, _, _ => valOr v1 v2*)
    | _, Vword w1, Vword w2 => Vword (bopWordVal b w1 w2)
    | eqop, _, _ => eqVal v1 v2
    | neqop, _, _ => valNot (eqVal v1 v2)
    | (subop | ltop | gtop | lteop | gteop), Vptr (b1, k1), Vptr (b2, k2) =>
      if eq_block_dec b1 b2 then
        Vword (bopWordVal b k1 k2)
      else
        Vundef
    | (plusop | subop), Vptr (b1, k1), Vword k2 => Vptr (b1, bopWordVal b k1 k2)
    | plusop, Vword k1, Vptr (b1, k2) =>
      Vptr (b1, bopWordVal b k1 k2)
    | _, _, _ => Vundef
  end.


(* value syntax *)
Delimit Scope val_scope with val.
Notation "!" := valNot (at level 45) : val_scope.
Notation "v + w" := (bopVal plusop v w) : val_scope.
Notation "v - w" := (bopVal subop v w) : val_scope.
Notation "v * w" := (bopVal multop v w) : val_scope.
Notation "v < w" := (bopVal ltop v w) : val_scope.
Notation "v > w" := (bopVal gtop v w) : val_scope.
Notation "v <= w" := (bopVal lteop v w) : val_scope.
Notation "v >= w" := (bopVal gteop v w) : val_scope.
Notation "v == w" := (bopVal eqop v w)
  (at level 74, left associativity) : val_scope.
Notation "v != w" := (bopVal neqop v w) (at level 74) : val_scope.
(*Notation "v &&& w" := (bopVal andop v w) (at level 74) : val_scope.
Notation "v ||| w" := (bopVal orop v w) (at level 74) : val_scope.*)

Coercion Vword : word >-> val.


Definition addrOf (v:val) : option addr :=
  match v with
    | Vptr a => Some a
    | _ => None
  end.


Lemma valAddrAddEq : forall (a:addr) (k:word),
  (a + k)%val = a + k.
Proof.
  intros [b k] k'; auto.
Qed.

Lemma valWordMultEq : forall (x y:word),
  (x * y)%val = x * y.
Proof.
  intros x y; reflexivity.
Qed.


Lemma addrOfAddr : forall (a:addr),
  addrOf a = Some a.
Proof.
  destruct a; auto.
Qed.

Hint Rewrite addrOfAddr : AddrSimpl.
  (* addrSubEq causes problems where there are existential vars *)
  (* maybe it will work if we make the coercion explicit? *)


Lemma addrSubEq : forall (a:addr) (k:word),
  (a - k)%val = ((a - k)%addr:val).
Proof.
  intros [a k] k'; reflexivity.
Qed.

Lemma addrAddEq : forall (a:addr) (k:word),
  (a + k)%val = ((a + k)%addr:val).
Proof.
  destruct a; auto.
Qed.


Lemma zeroNotTrue :
  trueVal (null_ptr) <-> False.
Proof.
  constructor; intro H.
  inversion H; auto.
  elf; auto.
Qed.

Lemma undefNotTrue :
  trueVal Vundef <-> False.
Proof.
  constructor; intro H.
  inversion H; auto.
  elf; auto.
Qed.

Hint Rewrite zeroNotTrue undefNotTrue : valEq.

Ltac vsimpl := autorewrite with valEq in * |- * ; 
  try (elf; auto; fail).


Ltac addrSplit :=
  match goal with
    | x : addr |- _ => destruct x
    | x : prod block word |- _ => destruct x
  end.

Ltac valSplit :=
  repeat addrSplit;
  match goal with
    | H : context [eq_word_dec ?x ?y] |- _ => 
      let E := fresh "E" in
        let NE := fresh "NE" in
          destruct (eq_word_dec x y) as [E | NE]; [subst |]
    | H : context [eq_block_dec ?x ?y] |- _ => 
      let E := fresh "E" in
        let NE := fresh "NE" in
          destruct (eq_block_dec x y) as [E | NE]; [subst |]
    | H : context [lt_word_dec ?x ?y] |- _ =>
      let E := fresh "E" in
        let NE := fresh "NE" in
          destruct (lt_word_dec x y) as [E | NE]
    | H : context [lte_word_dec ?x ?y] |- _ =>
      let E := fresh "E" in
        let NE := fresh "NE" in
          destruct (lte_word_dec x y) as [E | NE]
  end.

Ltac simplValSplit :=
  repeat addrSplit;
    vsimpl; simpl in *; try (valSplit; vsimpl); auto.

(* 8.2b3 possible danger from changed (?) rewrite vs setoid_rewrite *)
Lemma trueValEq : forall x y,
  impl (trueVal (x == y)%val) (x = y).
Proof.
  intros x y E.
  destruct x; simplValSplit; destruct y; simplValSplit; simplValSplit.
Qed.


Ltac discValSplit :=
  repeat addrSplit;
    try discriminate; simpl in *; unfold word_to_bool in *; simpl in * ;
      try valSplit; try discriminate.

(* 8.2b3 possible danger from changed (?) rewrite vs setoid_rewrite *)
Lemma nonzeroValNeq : forall x y,
  impl ((x != y)%val = 0%word) (x = y).
Proof.
  intros x y E.
  destruct x; discValSplit; destruct y; discValSplit; auto.
  elf; auto.
  simpl in E; compute in E; discValSplit; elf; auto.
  clear E; discValSplit.
  discValSplit; auto; elf; auto.
  elf; discValSplit; auto.
Qed.
  
Lemma nonzeroValEq : forall v w,
  impl ((v == w)%val = 0%word) (v <> w).
Proof.
  intros v w E.
  destruct v; discValSplit; destruct w; discValSplit.
  contradict NE; auto.
  injector NE; auto.
  contradict NE; injector NE; auto.
  elf; auto.
  discValSplit; contradict NE; injector NE; auto.
  contradict NE; injector NE; auto.
Qed.

Lemma trueValNeq : forall v w,
  impl (trueVal (v != w)%val) (v <> w).
Proof.
  intros v w E.
  apply nonzeroValEq.
  simpl in * ; destruct v; simpl in * ; vsimpl.
  destruct w; simplValSplit.
  destruct w; simplValSplit; simplValSplit.
  elf; compute in E.
  valSplit.
  discriminate.
  inversion E; auto.
Qed.


Lemma eqNullPtrVundef : forall v,
  ((v == null_ptr)%val = Vundef) <-> (v = Vundef).
Proof.
  destruct v; simpl; constructor; intros E; auto.
  valSplit; discriminate.
  discriminate.
  destruct a; destruct (eq_word_dec 0%word 0%word). discriminate.
  elf; auto.
  discriminate.
Qed.


Lemma neqNullPtrVundef : forall v,
  ((v != null_ptr)%val = Vundef) <-> (v = Vundef).
Proof.
  destruct v; simpl; constructor; intros E; auto.
  valSplit; discriminate. discriminate.
  destruct a; destruct (eq_word_dec 0%word 0%word).
  discriminate.
  elf; auto.
  discriminate.
Qed.


Lemma trueBopWordValImp : forall b w w',
  impl (trueVal (bopVal b (Vword w) (Vword w')))
  (match b with
     | plusop => (w + w' <> 0)%word
     | subop => (w - w' <> 0)%word
     | multop => (w * w' <> 0)%word
     | ltop => (w < w')%word
     | gtop => (w' < w)%word
     | lteop => (w <= w')%word
     | gteop => (w' <= w)%word
     | eqop => w = w'
     | neqop => w <> w'
(*     | andop => True
     | orop => True*)
   end).
Proof.
  intros b w w' H; destruct b; simpl in H; try (valSplit; auto; vsimpl).
  inversion H; auto.
  inversion H; auto.
  inversion H; auto.
Qed.

Lemma injVW : forall w w',
  Vword w = Vword w' ->
  w = w'.
Proof.
  injection 1; auto.
Qed.


Lemma zeroWordBopVal : forall b w w',
  impl (bopVal b (Vword w) (Vword w') = 0%word)
  (match b with
     | plusop => (w + w' = 0)%word
     | subop => (w - w' = 0)%word
     | multop => (w * w' = 0)%word
     | ltop => (w' <= w)%word
     | gtop => (w <= w')%word
     | lteop => (w' < w)%word
     | gteop => (w < w')%word
     | eqop => w <> w'
     | neqop => w = w'
   end).
Proof.
  intros b w w' H; destruct b; simpl in H;
    try (discValSplit; clear H; destruct w; destruct w'; wunfold; omega).
  hyr H (injVW H); auto.
  hyr H (injVW H); auto.
  hyr H (injVW H); auto.
  discValSplit; auto.
  discValSplit; auto.
Qed.


Lemma vundefWordBopVal : forall b w w',
  impl (bopVal b (Vword w) (Vword w') = Vundef)
  (match b with
     | plusop => False
     | subop => False
     | multop => False
     | ltop => False
     | gtop => False
     | lteop => False
     | gteop => False
     | eqop => False
     | neqop => False
   end).
Proof.
  intros b w w' H; destruct b; auto; discriminate.
Qed.


(* This isn't actually used, as I think this kind of test
   is a dead end.  Can't get anything useful out of the false case. *)
Lemma trueValPtrLt : forall x (k:word),
  trueVal (x < x + 4%word * k)%val ->
  k <> 0%word.
Proof.
  intros x k T E; subst; destruct x; simpl in T.
  inversion T.
  replace (w + 4 * 0)%word with w in T by ring.
  destruct (lt_word_dec w w).
  hyr w0 (wordLtInj w0); omega.
  inversion T; auto.
  destruct a; destruct (eq_block_dec b b).
  2:auto.
  replace (w + 4 * 0)%word with w in T by ring.
  destruct (lt_word_dec w w).
  hyr w0 (wordLtInj w0); omega.
  inversion T; auto.
Qed.


Lemma trueBopAddrAddrValImp : forall b a a',
  impl (trueVal (bopVal b (Vptr a) (Vptr a')))
  (match b with
     | plusop => False
     | subop => blockOf a = blockOf a' /\ 
       trueVal (offsetOf a - offsetOf a')%word
     | multop => False
     | ltop => (a < a')%addr
     | gtop => (a' < a)%addr
     | lteop => (a <= a')%addr
     | gteop => (a' <= a)%addr
     | eqop => a = a'
     | neqop => a <> a'
   end).
Proof.
  intros b [a x] [a' x'] H; destruct b; try trivial; simpl in H; vsimpl;
    repeat (valSplit; auto; vsimpl; unfold word_to_bool in *; try constructor);
      try (intros X; injector X; auto).
  discriminate E.
Qed.


Ltac valBoolCasesSplit :=
  match goal with
    | H : context [valBoolCases ?x] |- _ =>
      let T := fresh "T" in
      let F := fresh "F" in
      let U := fresh "U" in
        destruct (valBoolCases x) as [[T | F] | U]
  end.


Lemma zeroBopAddrAddrValImp : forall b a a',
  impl (bopVal b (Vptr a) (Vptr a') = 0%word)
  (match b with
     | plusop => False
     | subop => blockOf a = blockOf a' /\ 
       (offsetOf a - offsetOf a' = 0)%word
     | multop => False
     | ltop => (a' <= a)%addr
     | gtop => (a <= a')%addr
     | lteop => (a' < a)%addr
     | gteop => (a < a')%addr
     | eqop => a <> a'
     | neqop => a = a'
   end).
Proof.
  intros b [a x] [a' x'] H; destruct b; simpl in H;
    try discriminate H;
      repeat (valSplit; auto; vsimpl; unfold word_to_bool in *;
        try constructor; auto);
      try discriminate H; try (womega; fail);
        try (intros X; injector X; auto).

  simpl; apply injVW in H; auto.
Qed.
  

Lemma vundefBopAddrAddrValImp : forall b a a',
  impl (bopVal b (Vptr a) (Vptr a') = Vundef)
  (match b with
     | (plusop | multop) => True
     | (eqop | neqop) => False
     | (subop | ltop | lteop | gtop | gteop) => 
       blockOf a <> blockOf a'
   end).
Proof.
  intros b [a x] [a' x'] H; destruct b; simpl in H; trivial;
    try (valSplit; try discriminate); auto.
Qed.



(* At some point, might want the rules for word-ptr and ptr-word,
   though I think that it is unlikely that those will be used. *)

(*
Lemma trueValAndEq : forall x y,
  impl (trueVal (x &&& y)%val) (trueVal x /\ trueVal y).
Proof.
  simpl; intros x y Hp; unfold valAnd in Hp.
  destruct (valBoolCases x) as [[T | F] | U].
  constructor; auto.
  destruct (valBoolCases y) as [[T' | F] | U]; auto.
  simpl in Hp; inversion Hp; elf; auto.
  inversion Hp.
  destruct (valBoolCases y) as [[T | F'] | U];
    inversion Hp; elf; auto.
  inversion Hp.
Qed.


Lemma trueValOrEq : forall v v',
  impl (trueVal (v ||| v')%val) (trueVal v \/ trueVal v').
Proof.
  simpl; unfold valOr; intros x y Hp.
  destruct (valBoolCases x) as [[T | F] | U]; auto.
  right; clear x F; destruct (valBoolCases y) as [[T | F] | U]; auto.
  inversion Hp.
  elf; auto.
  inversion Hp.
  inversion Hp.
Qed.


Lemma zeroValAndEq : forall x y,
  impl ((x &&& y)%val = 0%word) (x = 0%word \/ y = 0%word).
Proof.
  simpl; unfold valAnd; intros x y Hp.
  destruct (valBoolCases x) as [[T | F] | U].
  destruct (valBoolCases y) as [[T' | F] | U]; auto; discriminate.
  auto.
  discriminate.
Qed.

Lemma zeroValOrEq : forall x y,
  impl ((x ||| y)%val = 0%word) (x = 0 /\ y = 0)%word.
Proof.
  simpl; unfold valOr; intros x y E.
  destruct (valBoolCases x) as [[T | F] | U].
  destruct (valBoolCases y) as [[T' | F] | U]; discriminate.
  destruct (valBoolCases y) as [[T | F'] | U]; try discriminate.
  auto.
  discriminate.
Qed.


Lemma undefValAndEq : forall x y,
  impl ((x &&& y)%val = Vundef) (x = Vundef \/ y = Vundef).
Proof.
  simpl; unfold valAnd; intros x y E.
  destruct (valBoolCases x) as [[_ | _] | U]; auto;
    destruct (valBoolCases y) as [[_ | _] | U]; auto; discriminate.
Qed.


Lemma undefValOrEq : forall x y,
  impl ((x ||| y)%val = Vundef) (x = Vundef \/ y = Vundef).
Proof.
  simpl; unfold valOr; intros x y E.
  destruct (valBoolCases x) as [[_ | _] | U]; auto;
    destruct (valBoolCases y) as [[_ | _] | U]; auto; discriminate.
Qed.
*)

Lemma trueValBoolEq : forall w,
  trueVal (Vword w) ->
  word_to_bool w = true.
Proof.
  destruct w; inversion 1; subst; unfold word_to_bool; 
    destruct (eq_word_dec (wordi wok)); auto.
Qed.


Lemma falseValBoolEq :
  word_to_bool 0%word = false.
Proof.
  unfold word_to_bool; destruct (eq_word_dec 0%word); auto; elf; auto.
Qed.


(* tests on booleans *)
Lemma bool_to_wordTrueEqv : forall b,
  trueVal (Vword (bool_to_word b)) <->
  b = true.
Proof.
  intros b; constructor; intro H.
  apply trueValBoolEq in H.
  rewrite bool_to_word_to_bool in H; auto.
  subst; simpl; constructor; discriminate.
Qed.


Lemma bool_to_wordFalseEqv : forall b,
  (bool_to_word b = 0%word) <-> (b = false).
Proof.
  intros b; constructor.
  destruct b; simpl; intros E; [discriminate | auto].
  intros; subst; reflexivity.
Qed.


Lemma vwordInjEqv : forall x y,
  (Vword x = Vword y) <-> (x = y).
Proof.
  intros x y; constructor; intros E; [injector E | subst]; auto.
Qed.


Hint Rewrite zeroNotTrue undefNotTrue : valEq.
Hint Rewrite trueBopWordValImp trueValEq trueValNeq undefNotTrue
  zeroNotTrue zeroWordBopVal nonzeroValEq nonzeroValNeq 
  neqNullPtrVundef eqNullPtrVundef : valEq.
Hint Rewrite vwordInjEqv bool_to_wordTrueEqv bool_to_wordFalseEqv 
  trueBopAddrAddrValImp zeroBopAddrAddrValImp vundefBopAddrAddrValImp : valEq.
Hint Rewrite (*trueValAndEq trueValOrEq zeroValAndEq zeroValOrEq*)
  (*undefValAndEq undefValOrEq*) vundefWordBopVal : valLogEq.


Lemma bopValWWeq : forall b w w',
  bopVal b (Vword w) (Vword w') = Vword (bopWordVal b w w').
Proof.
  intros b w w'; destruct b; auto; simpl.
Qed.
Hint Rewrite bopValWWeq : valEq2.


Lemma eq_val_dec : forall (v v':val),
  {v = v'} + {v <> v'}.
Proof.
  destruct v; destruct v'; auto; try (right; discriminate).
  destruct (eq_word_dec w w0).
  subst; auto.
  right; intro X; injector X; auto.
  destruct a; destruct a0; destruct (eq_block_dec b b0).
  subst; destruct (eq_word_dec w w0).
  subst; auto.
  right; intro X; injector X; auto.
  right; intro X; injector X; auto.
Qed.


Lemma valAddAssoc : forall x y z,
  (x + (y + z) = x + y + z)%val.
Proof.
  destruct x; try (destruct a); auto;
    destruct y; try (destruct a); auto;
      destruct z; try (destruct a); auto; simpl; f_equal; [ring | | | ];
        f_equal; ring.
Qed.


Lemma VwordNeqInj : forall w w',
  (Vword w <> Vword w') <-> (w <> w').
Proof.
  intros w w'; constructor; intros NE E.
  subst w'; auto.
  injector E; auto.
Qed.

Lemma bopWordValPlusEq : forall w1 w2,
  bopWordVal plusop w1 w2 = (w1 + w2)%word.
Proof.
  reflexivity.
Qed.

Lemma bopWordValSubEq : forall w1 w2,
  bopWordVal subop w1 w2 = (w1 - w2)%word.
Proof.
  reflexivity.
Qed.

Lemma bopWordValMultEq : forall w1 w2,
  bopWordVal multop w1 w2 = (w1 * w2)%word.
Proof.
  reflexivity.
Qed.

Hint Rewrite VwordNeqInj 
  bopWordValPlusEq bopWordValSubEq bopWordValMultEq : valEq.

Lemma addrWordAddEq : forall (a:addr) x,
  (a + Vword x)%val = (a + x)%addr.
Proof.
  intros [a x] y; reflexivity.
Qed.
Hint Rewrite addrWordAddEq : valEq.


Lemma ptrSubEq : forall x y z,
  (Vptr (x + y)%addr - Vptr (x + z)%addr)%val = Vword (y - z)%word.
Proof.
  clear.  intros [b x] y z; simpl.
  destruct (eq_block_dec b b).
  f_equal; ring.
  elf; auto.
Qed.
Hint Rewrite ptrSubEq : valEq.


Lemma zeroBopWordValImp : forall b w w',
  impl (bopWordVal b w w' = 0%word)
  (match b with
     | plusop => (w + w' = 0)%word
     | subop => (w - w' = 0)%word
     | multop => (w * w' = 0)%word
     | ltop => (w' <= w)%word
     | gtop => (w <= w')%word
     | lteop => (w' < w)%word
     | gteop => (w < w')%word
     | eqop => w <> w'
     | neqop => w = w'
   end).
Proof.
  intros b w w' H; destruct b; simpl in H; auto; try (apply injVW in H; auto);
    try (valSplit; auto; vsimpl); try discriminate H.
  womega.
  womega.
Qed.
Hint Rewrite zeroBopWordValImp : valEq.


Lemma bangTrueIff : forall e,
  (trueVal (valNot e)) <-> (e = 0%word).
Proof.
  clear. intros e; constructor; intros T.
  inversion T; clear T; subst.

  (* case: word *)
  destruct e; simpl in H; try discriminate H.
  injector H.
  unfold word_to_bool in H0; valSplit; auto; elf; auto.

  (* case: ptr *)
  destruct e; simpl in H0; try discriminate H0.

  subst e; simpl; unfold word_to_bool.
  destruct (eq_word_dec 0 0).
  simpl; constructor; discriminate.
  elf; auto.
Qed.


Lemma bangZeroImp : forall e,
  impl (valNot e = 0%word) (trueVal e).
Proof.
  clear. intros e T; destruct e; try discriminate T.
  simpl in T; unfold word_to_bool in T.
  valSplit.
  discriminate T.
  constructor; auto.
Qed.

Lemma bangUndefImp : forall e,
  impl (valNot e = Vundef) (e = Vundef \/ addrOf e <> None).
Proof.
  clear. intros e T; destruct e.  auto. discriminate T. right; discriminate.
Qed.

Hint Rewrite bangTrueIff bangZeroImp bangUndefImp : valEq.


Definition valToBool v : val :=
  match valBoolCases v with
    | inleft (left _) => true
    | inleft (right _) => false
    | inright _ => Vundef
  end.


(* additional address simplification rules *)

Lemma wordValMultEq : forall (w w':word),
  (w * w')%val = (w * w')%word.
Proof.
  reflexivity.
Qed.

Lemma addrWordValPlusEq : forall (a:addr) (w:word),
  (a + w)%val = (a + w)%addr.
Proof.
  destruct a; reflexivity.
Qed.

Lemma addrWordValSubEq : forall (a:addr) (w:word),
  (a - w)%val = (a - w)%addr.
Proof.
  destruct a; reflexivity.
Qed.

Hint Rewrite wordValMultEq addrWordValSubEq addrWordValPlusEq : AddrSimpl.
